/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  G_file.cc
    եɤ߽
*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream.h>
#include "GrayImage.h"
#include "etc.h"
#include "com.h"

/***********************************************************/
/********** class GrayImage(public Image) methods **********/
/***********************************************************/

/**********************/
/******** load ********/
/**********************/

// եǤΥ
// SIZE_OF_FLOAT != sizeof(float) ưʤ
#define SIZE_OF_FLOAT  4
#define SIZE_OF_DOUBLE 8

enum color_mode { cm_gray, cm_rgb };

static int load_from_file(GrayImage* img[], char* filename, color_mode cm);
static int load_vios_obj(GrayImage* img[], FILE* fp, color_mode cm);
static int load_pnm(GrayImage* img[], FILE* fp, color_mode cm);

int
GrayImage::load(char* filename)
{
    GrayImage* img[1];
    img[0] = this;
    return load_from_file(img, filename, cm_gray);
}

int
GrayImage::cload(char* filename, GrayImage* rgb_img_array[])
{
    return load_from_file(rgb_img_array, filename, cm_rgb);
}

static int
load_from_file(GrayImage* img[], char* filename, color_mode cm)
{
    FILE* fp;
    int   st;

    if ((fp = fopen(filename, "rb")) == NULL) {
	fprintf(stderr, "load_from_file: file `%s' open error\n", filename);
	perror(filename);
	return FAIL;
    }

    // եȽ
    int c1;
    c1 = getc(fp);
    if (c1 != EOF) ungetc(c1, fp);
    if (c1 == '!') {
	// VIOS  obj ե
	st = load_vios_obj(img, fp, cm);
    } else if (c1 == 'P') {
	// pnm ե
	st = load_pnm(img, fp, cm);
    } else {
	st = FAIL;
    }
    fclose(fp);

    return st;
}

// ---- VIOS ȼ version ----

static int read_int(int& ret, FILE* fp);
static int read_float(float& ret, FILE* fp);
static int read_double(double& ret, FILE* fp);
static int read_complex(complex& ret, FILE* fp);

typedef struct {
    int  width;
    int  height;
    int  pixel_type; /* int, real, ... */
    char name[64];
    int  hot_x;
    int  hot_y;
    int  image_type; /* gray, rgb */
} image_header_t;

static int load_vios_obj_header(FILE* fp, image_header_t* h);
static int load_vios_obj_image(GrayImage* img[], color_mode cm, FILE*);

static int
load_vios_obj(GrayImage* img[], FILE* fp, color_mode cm)
{
    image_header_t h;
    int v, values;

    if (cm == cm_gray) {
	values = 1;
    } else if (cm == cm_rgb) {
	values = 3;
    } else {
	fprintf(stderr, "load_vios_obj: bad color mode (%d)\n", (int)cm);
	return FAIL;
    }

    if (load_vios_obj_header(fp, &h) != OKAY) {
	fprintf(stderr, "load_vios_obj: load header error\n");
	return FAIL;
    }

    if ((cm == cm_gray && h.image_type != 1) ||
	(cm == cm_rgb  && h.image_type != 2))
    {
	fprintf(stderr, "load_vios_obj: error: image type mismatch\n");
	return FAIL;
    }
    for (v = 0; v < values; v++) {
	img[v]->setparam(h.width, h.height, h.hot_x, h.hot_y,
			 (pixel_type)h.pixel_type, 0/*cache*/);
	img[v]->setorgsize(h.width, h.height);
    }

    if (load_vios_obj_image(img, cm, fp) != OKAY) {
	fprintf(stderr, "load_vios_obj: obj load error\n");
	return FAIL;
    }

    return OKAY;
}

static int
load_vios_obj_header(FILE* fp, image_header_t* h)
{
    char magic[4];

    // magic Υå
    if (fread(magic, 1, 4, fp) != 4)
	return FAIL;
    if (strncmp(magic, "!OBJ", 4) != 0)
	return FAIL;

    // إåɤ߹
    if (read_int(h->width, fp) != OKAY) return FAIL;
    if (read_int(h->height, fp) != OKAY) return FAIL;
    if (read_int(h->pixel_type, fp) != OKAY) return FAIL;
    if (fread(h->name, 1, 64, fp) != 64) return FAIL;
    if (read_int(h->hot_x, fp) != OKAY) return FAIL;
    if (read_int(h->hot_y, fp) != OKAY) return FAIL;
    if (read_int(h->image_type,  fp) != OKAY) return FAIL;

    return OKAY;
}

static int
load_vios_obj_image(GrayImage* img[], color_mode cm, FILE* fp)
{
    int 	x, y;
    int 	v, values;
    int 	ival;
    float	fval;
    complex	cval;

    if (cm == cm_gray) {
	values = 1;
    } else if (cm == cm_rgb) {
	values = 3;
    } else {
	fprintf(stderr, "load_vios_obj_image: bad color mode (%d)\n", (int)cm);
	return FAIL;
    }

    for (y = 0; y < img[0]->get_height(); y++) {
	for (x = 0; x < img[0]->get_width(); x++) {
	    for (v = 0; v < values; v++) {
		void* image = img[v]->get_image();
		if (img[v]->get_type() == pt_int) {
		    read_int(ival, fp);
		    ((int*)image)[x+y*img[v]->get_width()] = ival;
		} else if (img[v]->get_type() == pt_real) {
		    read_float(fval, fp);
		    ((float*)image)[x+y*img[v]->get_width()] = fval;
		} else if (img[v]->get_type() == pt_complex) {
		    read_complex(cval, fp);
		    ((complex*)image)[x+y*img[v]->get_width()] = cval;
		} else {
		    fprintf(stderr, "load: unknown type %d\n",
			    (int)img[v]->get_type());
		    return FAIL;
		}
	    }
	}
    }

    if (feof(fp) || ferror(fp)) {
	return FAIL;
    } else {
	return OKAY;
    }
}

// ޥ endian ˰¸ʤɤ߹

int
read_int(int& ret, FILE* fp)
{
    int c1, c2, c3, c4;

    c1 = getc(fp);
    c2 = getc(fp);
    c3 = getc(fp);
    c4 = getc(fp);
    if (c4 == EOF) return FAIL;
    ret = (c1 << 24) | (c2 << 16) | (c3 << 8) | c4;
    return OKAY;
}

int
read_float(float& ret, FILE* fp)
{
    int  c, i;
    int  endian = 0x01;
    unsigned char value[SIZE_OF_FLOAT];

    for (i = 0; i < SIZE_OF_FLOAT; i++) {
	if ((c = getc(fp)) == EOF) return FAIL;
	if ((*(unsigned char*)&endian) == 0x01) {
	    // little endian
	    value[(SIZE_OF_FLOAT-1)-i] = (unsigned char)c;
	} else {
	    // big endian
	    value[i] = (unsigned char)c;
	}
    }

    ret = *(float*)(&value[0]);
    return OKAY;
}

int
read_double(double& ret, FILE* fp)
{
    int  c, i;
    int  endian = 0x01;
    unsigned char value[SIZE_OF_DOUBLE];

    for (i = 0; i < SIZE_OF_DOUBLE; i++) {
	if ((c = getc(fp)) == EOF) return FAIL;
	if ((*(unsigned char*)&endian) == 0x01) {
	    // little endian
	    value[(SIZE_OF_DOUBLE-1)-i] = (unsigned char)c;
	} else {
	    // big endian
	    value[i] = (unsigned char)c;
	}
    }

    ret = *(double*)(&value[0]);
    return OKAY;
}

int
read_complex(complex& ret, FILE* fp)
{
    double re, im;

    if (read_double(re, fp) != OKAY) return FAIL;
    if (read_double(im, fp) != OKAY) return FAIL;

    complex c(re, im);
    ret = c;
    return OKAY;
}

// -------- pnm version --------

static int read_number(int *ret, FILE *fp);

static int
load_pnm(GrayImage* img[], FILE* fp, color_mode cm)
{
    int c1, c2;
    color_mode file_cm;
    int plain_or_raw;
    int width, height;
    int max_value;
    int v, values; // gray: 1, color: 3

    // magic Υå
    c1 = getc(fp); if (c1 != 'P') return FAIL;
    c2 = getc(fp);
    if (c2 == '2') {
	file_cm = cm_gray;
	plain_or_raw = 0;  // plain
    } else if (c2 == '3') {
	file_cm = cm_rgb;
	plain_or_raw = 0;  // plain;
    } else if (c2 == '5') {
	file_cm = cm_gray;
	plain_or_raw = 1;  // raw
    } else if (c2 == '6') {
	file_cm = cm_rgb;
	plain_or_raw = 1;  // raw
    } else {
	fprintf(stderr, "load_pnm: not pgm or ppm file\n");
	return FAIL;
    }

    if (file_cm != cm) {
	fprintf(stderr, "load_pnm: color mode mismatch\n");
	return FAIL;
    }
    
    // إåɤ߹
    if (read_number(&width,     fp) != 0) return FAIL;
    if (read_number(&height,    fp) != 0) return FAIL;
    if (read_number(&max_value, fp) != 0) return FAIL;

    if (plain_or_raw == 1) {
	// raw ΤȤ max_value μˤʤ⤷ʤ
	// ڡɤФƤ
	int c;
	c = getc(fp);
	if (!isspace(c))
	    ungetc(c, fp);
    }

    // 
    if (file_cm == cm_gray) {
	values = 1; // gray
    } else {
	values = 3; // r g b
    }

    for (v = 0; v < values; v++) {
	img[v]->setparam(width, height, 0/*hx*/, 0/*hy*/, pt_int, 0/*cache*/);
	img[v]->setorgsize(width, height);
    }

    // ɤ߹
    int* image[3];
    for (v = 0; v < values; v++) {
	image[v] = (int*)img[v]->get_image();
    }
    int  x, y;
    int  value;
    int  ret;

    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		if (plain_or_raw == 0) {
		    // plain
		    ret = read_number(&value, fp);
		} else {
		    // raw
		    value = getc(fp);
		    ret = (value == EOF);
		}
		if (ret) {
		    fprintf(stderr, "read error\n");
		    return FAIL;
		}
		image[v][x+y*width] = value;
	    }
	}
    }

    return OKAY;
}

// pnm file ɤ߹Ѵؿ

// ȤɤФ1ʸ֤
static int
get(FILE* fp)
{
    int c;

    for (;;) {
	c = getc(fp);
	if (c != '#')
	    break;
	while ((c = getc(fp)) != '\n' && c != EOF)
	    ;
    }
    return c;
}

static void
skip_space(FILE* fp)
{
    int c;

    while (c = get(fp), isspace(c))
	;
    if (c != EOF)
	ungetc(c, fp);
}

// ɤ߹֤
// ɤ᤿ 0 ɤʤä 1 ֤
static int
read_number(int* ret, FILE* fp)
{
    int c;
    int num;

    skip_space(fp);
    c = get(fp);
    ungetc(c, fp);
    if (isdigit(c) || c == '-') {
	if (fscanf(fp, "%d", &num) == 1) {
	    *ret = num;
	    return 0;
	} else {
	    return 1;
	}
    } else {
	return 1;
    }
}

/**********************/
/******** save ********/
/**********************/

typedef int (*save_func_t)(void* buf, int size, void* arg);

static int save_func(save_func_t func, void* arg,
		     GrayImage* img[], int mode, color_mode cm);
static int save_in_fp(FILE* fp, GrayImage* img[], int mode, color_mode cm);

static int store_header_func(save_func_t func, void* arg,
			     GrayImage* img[], color_mode cm);
static int store_image_func(save_func_t func, void* arg,
			    GrayImage* img[], color_mode cm);

static int write_int_func(int value, save_func_t func, void* arg);
static int write_float_func(float value, save_func_t func, void* arg);
static int write_double_func(double value, save_func_t func, void* arg);
static int write_complex_func(complex value, save_func_t func, void* arg);

static int save_pnm(save_func_t func, void* arg, GrayImage* img[],
		    color_mode cm);

// ǻδؿͳǥ֤᥽å

int
GrayImage::save_through_func(int (*outfunc)(void* buf, int size, void* arg),
			     void* arg, int mode)
{
    GrayImage* img[1];
    img[0] = this;
    return save_func(outfunc, arg, img, mode, cm_gray);
}

int
GrayImage::csave_through_func(int (*outfunc)(void* buf, int size, void* arg),
			     void* arg, int mode, GrayImage* rgb_img_array[])
{
    return save_func(outfunc, arg, rgb_img_array, mode, cm_rgb);
}

// ⡼ɤ˱¸ؿƤӽФ
static int
save_func(save_func_t func, void* arg, GrayImage* img[],
	  int mode, color_mode cm)
{
    if (mode == 0) { // vios obj mode
	if (store_header_func(func, arg, img, cm) == FAIL)
	    return FAIL;
	if (store_image_func(func, arg, img, cm) == FAIL)
	    return FAIL;
	return OKAY;
    } else if (mode == 1) { // pnm mode
	return save_pnm(func, arg, img, cm);
    } else {
	fprintf(stderr, "save_func: bad mode (%d)\n", mode);
	return FAIL;
    }
}

// ե˥֤᥽å
// ؿͳǥ֤᥽åɤȤ

static int
save_in_fp_func(void* buf, int size, void* arg)
{
    if (fwrite(buf, 1, size, (FILE*)arg) == (size_t)size) {
	return size;
    } else {
	return -1;
    }
}

int
GrayImage::save(char* filename, int mode)
{
    GrayImage* img[1]; img[0] = this;
    FILE* fp;
    if ((fp = fopen(filename, "wb")) == NULL) {
	fprintf(stderr, "GrayImage::save: cannot open file `%s'\n",
		filename);
	return FAIL;
    }
    if (save_in_fp(fp, img, mode, cm_gray) == FAIL) {
	fprintf(stderr, "GrayImage::save: error in save\n");
	fclose(fp);
	return FAIL;
    }
    fclose(fp);
    return OKAY;
}

int
GrayImage::csave(char* filename, int mode, GrayImage* rgb_img_array[])
{
    FILE* fp;
    if ((fp = fopen(filename, "wb")) == NULL) {
	fprintf(stderr, "GrayImage::csave: cannot open file `%s'\n",
		filename);
	return FAIL;
    }
    if (save_in_fp(fp, rgb_img_array, mode, cm_rgb) == FAIL) {
	fprintf(stderr, "GrayImage::csave: error in save\n");
	fclose(fp);
	return FAIL;
    }
    fclose(fp);
    return OKAY;
}

static int
save_in_fp(FILE* fp, GrayImage* img[], int mode, color_mode cm)
{
    return save_func(save_in_fp_func, fp, img, mode, cm);
}

// ؿͳǥ֤βؿ

static int
store_header_func(save_func_t func, void* arg, GrayImage* img[], color_mode cm)
{
    int img_class;

    if (cm == cm_gray) {
	img_class = 1; // gray
    } else if (cm == cm_rgb) {
	img_class = 2; // rgb
    } else {
	fprintf(stderr, "store_header_func: vios csave cmode error\n");
	return FAIL;
    }

    if ((*func)("!OBJ", 4, arg) < 0) return FAIL;
    if (write_int_func(img[0]->get_width(), func, arg) == FAIL) return FAIL;
    if (write_int_func(img[0]->get_height(), func, arg) == FAIL) return FAIL;
    if (write_int_func((int)img[0]->get_type(), func, arg)== FAIL) return FAIL;

    char buf[64];
    memset(buf, '\0', 64);
    if ((*func)(buf, 64, arg) < 0) return FAIL;

    if (write_int_func(img[0]->get_hot_x(), func, arg) == FAIL) return FAIL;
    if (write_int_func(img[0]->get_hot_y(), func, arg) == FAIL) return FAIL;
    if (write_int_func(img_class, func, arg) == FAIL) return FAIL;

    return OKAY;
}

static int
store_image_func(save_func_t func, void* arg, GrayImage* img[], color_mode cm)
{
    int x, y;
    int st;
    void* image;
    int v, values;

    if (cm == cm_gray) {
	values = 1;
    } else if (cm == cm_rgb) {
	values = 3;
    } else {
	fprintf(stderr, "store_header_func: vios csave cmode error\n");
	return FAIL;
    }

    for (y = 0; y < img[0]->get_height(); y++) {
	for (x = 0; x < img[0]->get_width(); x++) {
	    for (v = 0; v < values; v++) {
		image = img[v]->get_image();
		if (img[v]->get_type() == pt_int) {
		    st = write_int_func(((int*)image)[x+y*img[v]->get_width()],
					func, arg);
		} else if (img[v]->get_type() == pt_real) {
		    st = write_float_func(((float*)image)
					  [x+y*img[v]->get_width()],func, arg);
		} else if (img[v]->get_type() == pt_complex) {
		    st =write_complex_func(((complex*)image)
					   [x+y*img[v]->get_width()],func,arg);
		} else {
		    fprintf(stderr, "save: unknown type %d\n",
			    img[v]->get_type());
		    return FAIL;
		}
		if (st == FAIL) return FAIL;
	    }
	}
    }
    return OKAY;
}

// ޥ endian ˰¸ʤ񤭹

// ޥ endian Ĵ٤ big endian ǽ񤭽Ф
static int
write_with_big_endian(void* buf, int size, save_func_t func, void* arg)
{
    int c = 0x01;
    void* p;
    unsigned char tmp[16];
    if ((*(unsigned char*)&c) == 0x01) {
	// little endian
	unsigned char* q = (unsigned char*)buf;
	int i;
	for (i = 0; i < size; i++) {
	    tmp[i] = q[size - i -1];
	}
	p = tmp;
    } else {
	// big endian
	p = buf;
    }
    if ((*func)(p, size, arg) < 0) return FAIL;
    return OKAY;
}

static int
write_int_func(int value, save_func_t func, void* arg)
{
    unsigned char v[4];
    v[0] = (value>>24) & 0xff;
    v[1] = (value>>16) & 0xff;
    v[2] = (value>> 8) & 0xff;
    v[3] = (value    ) & 0xff;
    if ((*func)(v, 4, arg) < 0) return FAIL;
    return OKAY;
}

static int
write_float_func(float value, save_func_t func, void* arg)
{
    if (sizeof(float) != SIZE_OF_FLOAT) {
	fprintf(stderr, "write_float: sizeof(float) != %d\n", SIZE_OF_FLOAT);
	return FAIL;
    }
    return write_with_big_endian(&value, SIZE_OF_FLOAT, func, arg);
}

static int
write_double_func(double value, save_func_t func, void* arg)
{
    if (sizeof(double) != SIZE_OF_DOUBLE) {
	fprintf(stderr, "write_double: sizeof(double) != %d\n",SIZE_OF_DOUBLE);
	return FAIL;
    }
    return write_with_big_endian(&value, SIZE_OF_DOUBLE, func, arg);
}

static int
write_complex_func(complex value, save_func_t func, void* arg)
{
    if (write_double_func(real(value), func, arg) == FAIL) return FAIL;
    if (write_double_func(imag(value), func, arg) == FAIL) return FAIL;
    return OKAY;
}

// -------- pnm version --------

static int
save_pnm(save_func_t func, void* arg, GrayImage* img[], color_mode cm)
{
    int x, y;
    int max_value;
    int *image[3];
    char buf[256];
    int value;
    int v, values; // gray: 1, rgb: 3

    if (cm == cm_gray) {
	values = 1;
    } else {
	values = 3;
    }
    for (v = 0; v < values; v++) {
	if (img[v]->get_type() != pt_int) {
	    fprintf(stderr, "save_pnm: error: only int image\n");
	    return FAIL;
	}
	image[v] = (int*)img[v]->get_image();
    }

    // max_value
    max_value = 0;
    for (v = 0; v < values; v++) {
	for (y = 0; y < img[v]->get_height(); y++) {
	    for (x = 0; x < img[v]->get_width(); x++) {
		value = image[v][x + y*img[v]->get_width()];
		if (value > max_value) max_value = value;
	    }
	}
    }

    // output
    sprintf(buf, "P%c\n%d %d\n%d\n", cm == cm_gray ? '2' : '3',
	    img[0]->get_width(), img[0]->get_height(), max_value);
    if ((*func)(buf, strlen(buf), arg) < 0) return FAIL;


    int nl_mode = img[0]->get_width() > 8; // Ԥ뤫

    for (y = 0; y < img[0]->get_height(); y++) {
	for (x = 0; x < img[0]->get_width(); x++) {
	    for (v = 0; v < values; v++) {
		value = image[v][x + y*img[v]->get_width()];
		if (nl_mode) {
		    sprintf(buf, "%d\n", value);
		} else {
		    sprintf(buf, "%3d ", value);
		}
		if ((*func)(buf, strlen(buf), arg) < 0) return FAIL;
	    }
	}
	if (!nl_mode) {
	    sprintf(buf, "\n");
	    if ((*func)(buf, strlen(buf), arg) < 0) return FAIL;
	}
    }

    return OKAY;
}
