/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  GrayImage.cc
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "GrayImage.h"
#include "protocol.h"
#include "msgqueue.h"
#include "vbuf.h"
#include "waitlist.h"

extern msgqueue send_q;
waitlist image_boundary_list;

/***********************************************************/
/********** class GrayImage(public Image) methods **********/
/***********************************************************/

//  (ѥγ)
void
GrayImage::setsize(int w, int h, pixel_type ty, int ca)
{
    width  = w;
    height = h;
    type   = ty;
    cache  = ca;

    int memsize = get_image_size();

    if (image != NULL) {
	free(image);
    }
    image = malloc(memsize);
    if (image == NULL) {
	fprintf(stderr,
	"GrayImage::setsize: fatal error, cannot allocate memory for image\n");
	fprintf(stderr, "GrayImage::setsize: size = %d(byte)\n", memsize);
	fprintf(stderr, "GrayImage::setsize: w:%d, h:%d, ca:%d\n", w, h, ca);
	abort();
    }
}

// -------- ͤꡦ --------

// ---- кɸ ----

// ͤ

// ưѴĤ
complex
GrayImage::getpixel(int x, int y, pixel_type& t)
{
    t = type;
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::getpixel: out of range\n");
	return complex(0);
    }
    if      (type == pt_int)	 return complex((    (int*)image)[i]);
    else if (type == pt_real)	 return complex((  (float*)image)[i]);
    else if (type == pt_complex) return complex(((complex*)image)[i]);
    return complex(0);
}

// ηΤ
float
GrayImage::getpixelr(int x, int y)
{
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::getpixelr: out of range\n");
	return 0;
    }
    return ((float*)image)[i];
}

int
GrayImage::getpixeli(int x, int y)
{
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::getpixeli: out of range\n");
	return 0;
    }
    return ((int*)image)[i];
}

// ͤ

// ưѴĤ
void
GrayImage::setpixel(int x, int y, complex val)
{
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::setpixel: out of range\n");
	return;
    }
    if      (type == pt_int)	(    (int*)image)[i] = (int)real(val);
    else if (type == pt_real)	(  (float*)image)[i] = (float)real(val);
    else    			((complex*)image)[i] = val;
}

// ηΤ
void
GrayImage::setpixel(int x, int y, float val)
{
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::setpixel: out of range\n");
	return;
    }
    ((float*)image)[i] = val;
}

void
GrayImage::setpixel(int x, int y, int val)
{
    int i = xy_to_i(x, y);
    if (range_check(i) == FAIL) {
	fprintf(stderr, "GrayImage::setpixel: out of range\n");
	return;
    }
    ((int*)image)[i] = val;
}

// ---- кɸ ----

// ͤ

// ưѴĤ
complex
GrayImage::agetpixel(int x, int y, pixel_type& t)
{
    return getpixel(x - hot_x, y - hot_y, t);
}

// ηΤ
float
GrayImage::agetpixelr(int x, int y)
{
    return getpixelr(x - hot_x, y - hot_y);
}

int
GrayImage::agetpixeli(int x, int y)
{
    return getpixeli(x - hot_x, y - hot_y);
}

// ͤ򥻥å

// ưѴĤ
void
GrayImage::asetpixel(int x, int y, complex val)
{
    setpixel(x - hot_x, y - hot_y, val);
}

// ηΤ
void
GrayImage::asetpixel(int x, int y, float val)
{
    setpixel(x - hot_x, y - hot_y, val);
}

void
GrayImage::asetpixel(int x, int y, int val)
{
    setpixel(x - hot_x, y - hot_y, val);
}

// 

int
GrayImage::get_image_info(image_info& i)
{
    i.width      = width;
    i.height     = height;
    i.org_width  = org_width;
    i.org_height = org_height;
    i.pixel_type = (int)type;
    i.hot_x      = hot_x;
    i.hot_y      = hot_y;
    i.image_type = (int)it_gray;
    i.image_id   = image_id;
    i.div_id     = div_id;
    i.cache      = cache;
    return OKAY;
}

// boundary check դкɸɤ߽Ф
// 桼⥸塼⤫ƤФ
int
GrayImage::readint(int job_id, int image_no, boundary_type b,
		   int x, int y, int def)
{
    int xmin;
    int xmax;
    int ymin;
    int ymax;

    if (b == bd_nocache) {
	xmin = 0;
	xmax = width - 1;
	ymin = 0;
	ymax = height - 1;
    } else {
	xmin = -cache;
	xmax = width + cache*2 - 1;
	ymin = -cache;
	ymax = height + cache*2 - 1;
    }

    if (x < xmin || x > xmax ||	y < ymin || y > ymax) {
	if (b == bd_get || b == bd_nocache) {
	    // OM ׵Ф
	    ipu_ipu msg;
	    msg.to_om = new ipu_om;
	    msg.to_om->type = p_io_image_request;
	    msg.img_info = new image_info;
	    msg.img_info->width = 1;
	    msg.img_info->height = 1;
	    msg.img_info->hot_x = x + hot_x;
	    msg.img_info->hot_y = y + hot_y;
	    msg.img_info->cache = 0;
	    msg.img_info->image_id = image_id;
	    msg.img_info->div_id = -2; // Ȥʤ
	    msg.img_info->job_id = job_id;
	    msg.img_info->image_no = image_no;
	    send_q.putmsg(msg);

	    // ֻԤ
	    int* val;
	    int  ret;
	    val = (int*)image_boundary_list.wait(image_id, job_id);
	    ret = *val;
	    delete[] val;
	    return ret;
	} else if (b == bd_complete) {
	    if (x < xmin) x = xmin;
	    if (x > xmax) x = xmax;
	    if (y < ymin) y = ymin;
	    if (y > ymax) y = ymax;
	    return ((int*)image)[(y + cache) * (width + cache*2) + (x +cache)];
	} else if (b == bd_const) {
	    return def;
	} else {
	    fprintf(stderr, "GrayImage::readint: bad boundary %d\n", b);
	    return 0;
	}
    } else {
	return ((int*)image)[(y + cache) * (width + cache*2) + (x + cache)];
    }
}

float
GrayImage::readfloat(int job_id, int image_no, boundary_type b,
		     int x, int y, float def)
{
    (void)job_id; (void)image_no; (void)b; (void)x; (void)y; (void)def;
    fprintf(stderr, "GrayImage::readfloat: not implemented\n");
    return 0;
}

complex
GrayImage::readcomplex(int job_id, int image_no, boundary_type b,
		       int x, int y, complex def)
{
    (void)job_id; (void)image_no; (void)b; (void)x; (void)y; (void)def;
    fprintf(stderr, "GrayImage::readcomplex: not implemented\n");
    return 0;
}

// ɽ (ºݤɽ VPE Ԥ)

extern msgqueue send_q;
static int display_out_func(void* buf, int size, void* arg);

int
GrayImage::display()
{
    ipu_ipu msg;
    vbuf buf;

    if (save_through_func(display_out_func, &buf, 1) != OKAY) {
	fprintf(stderr, "GrayImage::display: save to memory failed\n");
	return FAIL;
    }

    int size = buf.getsize();
    msg.to_om = new ipu_om;
    msg.to_om->type   = p_io_image;
    msg.to_om->result = OKAY;
    msg.to_om->size = size;
    msg.buf = new char[size];
    memcpy(msg.buf, buf.getbuf(), size);

    send_q.putmsg(msg);

    return OKAY;
}

int
GrayImage::cdisplay(GrayImage* rgb_image[])
{
    ipu_ipu msg;
    vbuf buf;

    if (csave_through_func(display_out_func, &buf, 1, rgb_image) != OKAY) {
	fprintf(stderr, "GrayImage::cdisplay: save to memory failed\n");
	return FAIL;
    }

    int size = buf.getsize();
    msg.to_om = new ipu_om;
    msg.to_om->type   = p_io_image;
    msg.to_om->result = OKAY;
    msg.to_om->size = size;
    msg.buf = new char[size];
    memcpy(msg.buf, buf.getbuf(), size);

    send_q.putmsg(msg);

    return OKAY;
}

static int
display_out_func(void* buf, int size, void* arg)
{
    vbuf* bufp = (vbuf*)arg;
    if (bufp->append(buf, size) < 0) {
	return -1;
    } else {
	return size;
    }
}
