/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  Image.h
*********************************************************************/

#ifndef IMAGE_H
#define IMAGE_H

#include <stdio.h>
#include <string.h>
#include "comp.h"

#include "def.h"
#include "etc.h"
#include "protocol.h"

//##############################
//   typeʥ֥Ȥη
//##############################

enum image_type {
    it_null, it_gray, it_binary, it_rgb
};

enum pixel_type {
    pt_null, pt_int, pt_real, pt_complex
};

enum boundary_type {
    bd_null,
    bd_nocheck,  // 󥰥åȳåʤ
    bd_get,      // äƤ
    bd_complete, // 䴰
    bd_const,    // 
    bd_nocache   // äƤ (cache () äƤȤʤ)
};


//##############################
//        饹ΰ
//##############################
class Image;                //ܥ饹
    class GrayImage;        //public Image : ǻø

//##############################
//        ƥ饹
//##############################
class Image
{
  private:
  protected:
    int 	image_id;	//  ID
    int 	div_id; 	//  ID Ǥʬ ID
    int 	width;		// 
    int 	height;		// ⤵
    int 	org_width;	// ʬ
    int 	org_height;	// ʬι⤵
    int 	hot_x, hot_y;	// ۥåȥݥ
    pixel_type	type;		// ǡΥ (pt_int, pt_real, ...)
    int 	cache;		// ղǤ

    virtual image_type g_class() { return it_null; };

    char* pixel_type2string(pixel_type t) {
	switch (t) {
	  case pt_null:    return "pt_null";
	  case pt_int:     return "pt_int";
	  case pt_real:    return "pt_real";
	  case pt_complex: return "pt_complex";
	  default:         return "unknown";
	}
    };

  public:
    /***** 󥹥ȥ饯 *****/
    Image(int i_id, int d_id) {
	image_id   = i_id;
	div_id     = d_id;
	width      = 0;
	height     = 0;
	org_width  = 0;
	org_height = 0;
	hot_x      = 0;
	hot_y      = 0;
	type       = pt_null;
	cache      = 0;
    };

    /***** ǥȥ饯 *****/
    virtual ~Image() { };

    /***** ᥽å ***********/

    void	set_id(int i_id, int d_id)
	{ image_id = i_id; div_id = d_id; };
    virtual void setsize(int w, int h, pixel_type ty, int ca)
	{ width = w; height = h; type = ty; cache = ca; };
    void	sethot(int h_x, int h_y)
	{ hot_x = h_x; hot_y = h_y; };
    void	setparam(int w, int h, int h_x, int h_y,
				 pixel_type ty, int ca)
	{ setsize(w, h, ty, ca); sethot(h_x, h_y); };
    void	setorgsize(int ow, int oh)
	{ org_width = ow; org_height = oh; };

    // °᥽å
    int 	get_image_id() { return image_id; };
    int 	get_div_id() { return div_id; };
    int 	get_width() { return width; };
    int 	get_height() { return height; };
    int 	get_org_width() { return org_width; };
    int 	get_org_height() { return org_height; };
    int 	get_hot_x() { return hot_x; };
    int 	get_hot_y() { return hot_y; };
    pixel_type	get_type() { return type; };
    int 	get_cache() { return cache; };

    virtual int get_image_info(image_info& i)
	{ (void)i; return FAIL; };

    int 	get_one_width()	{		// 1Ǥ礭
	switch (type) {
	  case pt_int:     return sizeof(int);
	  case pt_real:    return sizeof(float);
	  case pt_complex: return sizeof(complex);
	  default:
	    // error
	    fprintf(stderr, "Image::get_one_width: bad type %d\n", (int)type);
	    return 1;
	}
    };

    int 	get_image_size() {		// 礭ϤɤΤ餤
	return (width + cache*2) * (height + cache*2) * get_one_width();
    };

    int 	have_id(int i_id, int d_id) {	//  ID ɤ
	if (image_id == i_id && div_id == d_id) return OKAY;
	else					return FAIL;
    };

    virtual Image& operator = (Image& a) {
	image_id   = a.image_id;
	div_id     = a.div_id;
	width      = a.width;
	height     = a.height;
	org_width  = a.org_width;
	org_height = a.org_height;
	hot_x      = a.hot_x;
	hot_y      = a.hot_y;
	type       = a.type;
	cache      = a.cache;
	return *this;
    };

    // 
    virtual void* get_image() { return NULL; };

    // ͤ
    virtual complex agetpixel(int x, int y, pixel_type& t)
	{ (void)x; (void)y; t = pt_null; return complex(0); };
    virtual float   agetpixelr(int x, int y)
	{ (void)x; (void)y; return 0; };
    virtual int     agetpixeli(int x, int y)
	{ (void)x; (void)y; return 0; };

    virtual complex getpixel(int x, int y, pixel_type& t)
	{ (void)x; (void)y; t = pt_null; return complex(0); };
    virtual float   getpixelr(int x, int y)
	{ (void)x; (void)y; return 0; };
    virtual int     getpixeli(int x, int y)
	{ (void)x; (void)y; return 0; };

    // ͤ
    virtual void asetpixel(int x, int y, complex c)
	{ (void)x; (void)y; (void)c; };
    virtual void asetpixel(int x, int y, float c)
	{ (void)x; (void)y; (void)c; };
    virtual void asetpixel(int x, int y, int c)
	{ (void)x; (void)y; (void)c; };

    virtual void setpixel(int x, int y, complex c)
	{ (void)x; (void)y; (void)c; };
    virtual void setpixel(int x, int y, float c)
	{ (void)x; (void)y; (void)c; };
    virtual void setpixel(int x, int y, int c)
	{ (void)x; (void)y; (void)c; };

    // boundary check դ
    virtual int readint(int job_id, int image_no, boundary_type b,
			int x, int y, int def = 0)
	{ (void)job_id; (void) image_no; (void)b; (void)x; (void)y;
	  (void)def; return 0; };
    virtual float readfloat(int job_id, int image_no, boundary_type b,
			    int x, int y, float def = 0)
	{ (void)job_id; (void) image_no; (void)b; (void)x; (void)y;
	  (void)def; return 0; };
    virtual complex readcomplex(int job_id, int image_no, boundary_type b,
				int x, int y, complex def = 0)
	{ (void)job_id; (void) image_no; (void)b; (void)x; (void)y;
	  (void)def; return 0; };

    // ɸѴ (x, y 1Υǥå)
    int xy_to_i(int x, int y) {
	return (y + cache) * (width + 2 * cache) + (x + cache);
    };
    int range_check(int x, int y) {
	if (x >= -cache && x < width  + cache &&
	    y >= -cache && y < height + cache)
	{
	    return OKAY;
	} else {
	    return FAIL;
	}
    }
    int range_check(int i) {
	if (i >= 0 && i < (width + 2 * cache) * (height + 2 * cache)) {
	    return OKAY;
	} else {
	    return FAIL;
	}
    }

    // ɽ
    virtual int display() { return FAIL; };
    virtual int cdisplay(GrayImage* rgb_image_array[])
	{ (void)rgb_image_array; return FAIL; };

    // ʬ䡦
    virtual int	split(int xn, int xunit, int yn, int yunit, int w, Image* a[])
	{ (void)xn;(void)xunit;(void)yn;(void)yunit;(void)w;(void)a;
	  return FAIL; };
    virtual int	join(int n, Image* a[])
	{ (void)n;(void)a; return FAIL; };

    // ɤ߹
    virtual int load(char* filename) // 쥤
	{ (void)filename; return FAIL; }
    virtual int cload(char* filename, GrayImage* rgb_img_array[]) // 顼
	{ (void)filename; (void)rgb_img_array; return FAIL; }

    // ¸
    virtual int save(char* filename, int mode) // mode 0: vios , 1: pnm
	{ (void)filename; (void)mode; return FAIL; };
    virtual int csave(char* filename, int mode, GrayImage* rgb_img_array[])
	{ (void)filename; (void)mode; (void)rgb_img_array; return FAIL; };

    // debug 
    void print() {
	fprintf(stdout, "********************************\n");
	fprintf(stdout, "image_id: %d, div_id: %d\n", image_id, div_id);
	fprintf(stdout, "     width = %d, height = %d, cache = %d\n",
		width, height, cache);
	fprintf(stdout, "     org_width = %d, org_height = %d\n",
		org_width, org_height);
	fprintf(stdout, "     HotPoint(%d, %d) PixelType: %s\n",
		hot_x, hot_y, pixel_type2string(type));
	fprintf(stdout, "********************************\n");
    };
    virtual void print_image() {
    };
};

#endif /* IMAGE_H */
