/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  array.h
    IPU-p & C++ ˤ⥸塼ǻȤ (Хѿ)
*********************************************************************/

#ifndef ARRAY_H
#define ARRAY_H

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include "def.h"
#include "comp.h"
#include "protocol.h"

enum array_type {
    at_int, at_real, at_complex
};

enum unify_type {
    uf_null, uf_max, uf_min, uf_dmax, uf_dmin, uf_add
};

class array {
  private:
    int 		array_id;	// ꤹ ID (OM )
    int			sub_id;		//  ID Ǥ ID
    array_type		type;		// Ǥη
    int 		size;		// ǿ (礭)
    void*		data;		// μ
    int 		job_id;
    int 		n_subs;         // ԡο (ɬ)
    pthread_mutex_t	user_lock;	// 桼ץफΥå

    int get_new_array_id(); // OM  ID 餦

    void set_param(array_type t, int s) {
	type = t;
	size = s;
	delete[] data;
	switch (type) {
	  case at_int:
	    data = new int[size];
	    memset(data, 0, size*sizeof(int));
	    break;
	  case at_real:
	    data = new float[size];
	    memset(data, 0, size*sizeof(float));
	    break;
	  case at_complex:
	    data = new complex[size];
	    memset(data, 0, size*sizeof(complex));
	    break;
	  default:
	    fprintf(stderr, "array::array: unknown type %d\n", type);
	    data = NULL;
	    return;
	}
    };

  public:

    array(array_type t, int s);

    ~array() {
	if (data != NULL) {
	    delete[] data;
	}
    };

    // ⥸塼⤫ƤФ륰Хѿ
    array(int job_id, int no, array_type t, int s);

    // OM  ID äƥץ饤١ѿ˥å
    int set_new_array_id() {
	array_id = get_new_array_id();
	return 0;
    };

    // Ǥη (¿ʤ) ֤
    array_type get_type() {
	return type;
    }

    // 礭 (Ǥο) ֤
    int get_size() {
	return size;
    }

    // 1ĤǤ礭֤
    int get_width() {
	switch (type) {
	  case at_int:     return sizeof(int);
	  case at_real:    return sizeof(float);
	  case at_complex: return sizeof(complex);
	  default:
	    fprintf(stderr, "array::get_width: unknown type %d\n", type);
	    return 1;
	}
    }

    // ID ֤
    int get_array_id() {
	return array_id;
    };
    int get_sub_id() {
	return sub_id;
    };

    // ID 򥻥å
    void set_array_id(int id) {
	array_id = id;
    };
    void set_sub_id(int id) {
	sub_id = id;
    };

    int get_sub_num() {
	return n_subs;
    }
    void set_sub_num(int n) {
	n_subs = n;
    }

    // ƬǤؤΥݥ󥿤֤
    void* get_ptr() {
	return data;
    };

    // ͤ򥻥å
    int set_value(int i, int ival) {
	if (i < 0 || i >= size) return FAIL;
	((int*)data)[i] = ival;
	return OKAY;
    };
    int set_value(int i, float rval) {
	if (i < 0 || i >= size) return FAIL;
	((float*)data)[i] = rval;
	return OKAY;
    };
    int set_value(int i, complex cval) {
	if (i < 0 || i >= size) return FAIL;
	((complex*)data)[i] = cval;
	return OKAY;
    };

    // ͤ򥲥å
    int get_value(int i, int& ival) {
	if (i < 0 || i >= size) return FAIL;
	ival = ((int*)data)[i];
	return OKAY;
    };
    int get_value(int i, float& rval) {
	if (i < 0 || i >= size) return FAIL;
	rval = ((float*)data)[i];
	return OKAY;
    };
    int get_value(int i, complex& cval) {
	if (i < 0 || i >= size) return FAIL;
	cval = ((complex*)data)[i];
	return OKAY;
    };

    // ¾桢
    int lock();
    int unlock();
    int read();              // 
    int read(int i);         // Ǥ
    int read(int i, int n);  //  i  n 
    int write();             // 
    int write(int i);        // Ǥ
    int write(int i, int n); //  i  n 
    int unify(unify_type);

    int lock_prog();
    int unlock_prog();
    int unify_prog(void** arrs, unify_type ut);
};

// array δ
class ArrayList {
  private:
    ArrayList* next;
    array*     val;

  public:
    ArrayList() { next = NULL; };
    ArrayList(ArrayList* n, array* v) { next = n; val = v; };
    ~ArrayList() { delete next; delete val; };

    // ꥹȤ ID  array õ
    array* search(int array_id, int sub_id) {
	ArrayList* p;
	for (p = next; p != NULL; p = p->next) {
	    if (p->val->get_array_id() == array_id &&
		p->val->get_sub_id()   == sub_id)
	    {
		return p->val;
	    }
	}
	return NULL;
    };

    // ꥹȤɲ
    int append(array* a) {
	if (search(a->get_array_id(), a->get_sub_id()) != NULL) {
	    // ǤϿƤ
	    return FAIL;
	}
	next = new ArrayList(next, a);
	return OKAY;
    };

    // ꥹȤ
    // array ΤΤ
    int del(int array_id) {
	ArrayList* p;
	ArrayList* prev;
	for (p = next, prev = this; p != NULL; prev = p, p = p->next) {
	    if (p->val->get_array_id() == array_id) {
		prev->next = p->next;
		p->next = NULL;
		delete p;
		return OKAY;
	    }
	}
	// Ĥʤä
	return FAIL;
    }
};

extern ArrayList AL;

int array_read(array_id_info* a_info, array_data_info* ad_info, int ipu_no);
int array_write(array_id_info* a_info, array_data_info* ad_info, void* buf);
int array_lock(array_id_info* a_info, int ipu_no);
int array_unlock(array_id_info* a_info, int ipu_no);
int array_lock_done(array_id_info* a_info);
int array_unify(array_id_info* a_info, void* buf);
int array_unlock_done(array_id_info* a_info);

#endif /* ARRAY_H */
