/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  calc.cc
*********************************************************************/

#include "comp.h"
#include "calc.h"
#include "stack.h"
#include "table.h"
#include "ObjList.h"
#include "def.h"

int
f_add(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    c = b + a;

    if      (t1 == nt_complex || t2 == nt_complex) ST.push(c);
    else if (t1 == nt_real    || t2 == nt_real   ) ST.push((float)real(c));
    else					   ST.push((int)real(c));
    return OKAY;
}

int
f_sub(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    c = b - a;

    if      (t1 == nt_complex || t2 == nt_complex) ST.push(c);
    else if (t1 == nt_real    || t2 == nt_real   ) ST.push((float)real(c));
    else					   ST.push((int)real(c));
    return OKAY;
}

int
f_mul(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    c = b * a;

    if      (t1 == nt_complex || t2 == nt_complex) ST.push(c);
    else if (t1 == nt_real    || t2 == nt_real   ) ST.push((float)real(c));
    else					   ST.push((int)real(c));
    return OKAY;
}

int
f_div(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    c = b / a;

    if      (t1 == nt_complex || t2 == nt_complex) ST.push(c);
    else if (t1 == nt_real    || t2 == nt_real   ) ST.push((float)real(c));
    else					   ST.push((int)real(c));
    return OKAY;
}

int
f_mod(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    int 	a, b, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    c = b % a;

    ST.push(c);
    return OKAY;
}

int
f_neg(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a, c;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    c = -a;

    if      (t1 == nt_complex) ST.push(c);
    else if (t1 == nt_real   ) ST.push((float)real(c));
    else		       ST.push((int)real(c));
    return OKAY;
}

int
f_sin(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(sin(a));
    else		  ST.push((float)sin(real(a)));
    return OKAY;
}

int
f_cos(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(cos(a));
    else		  ST.push((float)cos(real(a)));
    return OKAY;
}

int
f_tan(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(sin(a)/cos(a));
    else 		  ST.push((float)tan(real(a)));
    return OKAY;
}

int
f_asin(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) return FAIL;		// not supported
    else 		  ST.push((float)asin(real(a)));
    return OKAY;
}

int
f_acos(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) return FAIL;		// not supported
    else		  ST.push((float)acos(real(a)));
    return OKAY;
}

int
f_atan(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) return FAIL;		// not supported
    else		  ST.push((float)atan(real(a)));
    return OKAY;
}

int
f_sqrt(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(sqrt(a));
    else		  ST.push((float)sqrt(real(a)));
    return OKAY;
}

int
f_exp(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(exp(a));
    else		  ST.push((float)exp(real(a)));
    return OKAY;
}

int
f_log(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1;
    complex	a;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;

    if (t1 == nt_complex) ST.push(log(a));
    else		  ST.push((float)log(real(a)));
    return OKAY;
}

int
f_pow(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a,b;
    t1 = ST.sttype();
    if (ST.pop(a) == FAIL) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) == FAIL) return FAIL;

    if	    (t1 == nt_complex || t2 == nt_complex)
	ST.push(pow(b, a));
    else if (t1 == nt_real    || t2 == nt_real   )
	ST.push((float)pow(real(b), real(a)));
    else
	ST.push((int)pow(real(b), real(a)));
    return OKAY;
}
