/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  etc.cc
    ¾ν
*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <sys/time.h>
#include <pthread.h>
#include <errno.h>
#include "def.h"
#include "etc.h"
#include "com.h"
#include "protocol.h"
#include "msgqueue.h"

// ʬΥۥ֤̾
// ǽ get_hostname_first() Ƥ֤

#ifndef SYS_NMLN
#define SYS_NMLN 257
#endif

static char	myhostname[SYS_NMLN];

char*
get_hostname_first()
{
    struct utsname name;
    uname(&name);
    strcpy(myhostname, name.nodename);
    return myhostname;
}

char*
get_hostname()
{
    return myhostname;
}

#if (defined(__sun) && defined(__SVR4))
# define LOAD_RPC
#endif

// ֤ͤ
#ifdef LOAD_RPC
// ͤ뤿 RPC Ȥ
#include <rpc/rpc.h>
#include <rpcsvc/rstat.h>

#if defined(__sun)
extern "C" {
    enum clnt_stat rstat(char*, struct statstime*);
}
#endif

int
get_loadav()
{
    struct statstime statp;
    rstat(get_hostname(), &statp);
    return (int)statp.avenrun[0];
}

#else
# ifdef LOAD_GETLOADAVG

int
get_loadav()
{
    double loadav[3];
    getloadavg(loadav, 3);
    return (int)(loadav[0]*256);
}

# else

int
get_loadav()
{
    return 0;
}

# endif
#endif

// strdup  new Ƚ
char*
strdup_new(char* s)
{
    int   len;
    char* p;

    len = strlen(s);
    p = new char[len+1];
    strcpy(p, s);
    return p;
}

// "/name"  "name" Ѵΰ new char[size] ǳ
char*
strip_name(char* token)
{
    if (*token != '/') {
	// 顼
	return strdup_new(token);
    }
    return strdup_new(token+1);
}

// "(string)"  "string" Ѵΰ new char[size] ǳ
char*
strip_string(char* token)
{
    char*	s;
    int 	len;

    len = strlen(token);
    if (token[0] != '(' || token[len-1] != ')') {
	// 顼
	return strdup_new(token+1);
    }
    s = new char[len-2+1];
    strncpy(s, token+1, len-2);
    s[len-2] = '\0';
    return s;
}


// ߤλ֤
void
gettime(int& h, int& m, int& s, int& us)
{
    struct timeval	t;

    gettimeofday(&t, NULL);
    t.tv_sec += (9*3600); // JST-9, Ȥʤ٤Ĥ -9 ʤƤ⤤
    h  = (t.tv_sec % (24*3600)) / 3600;
    m  = (t.tv_sec % 3600) / 60;
    s  = t.tv_sec % 60;
    us = t.tv_usec;
}

// 2Ĥλκ֤
void
gettimes(int hs, int ms, int ss, int uss,
	 int he, int me, int se, int use,
	 int& ht, int& mt, int& st, int& ust)
{
    if (uss > use) {
	se--;
	use += 1000000;
    }
    ust = use - uss;
    if (ss > se) {
	me--;
	se += 60;
    }
    st = se - ss;
    if (ms > me) {
	he--;
	me += 60;
    }
    mt = me - ms;
    ht = he - hs;
}

void
time_print(char* message)
{
    timeval	t;
    int 	h,m,s,us;

    gettimeofday(&t, NULL);
    t.tv_sec += (9*3600); // JST-9, Ȥʤ٤Ĥ -9 ʤƤ⤤
                          //  -9 ȷᤦƤΤ⤪
                          // timezone Ĵ٤ΤɤΤ
    h  = (t.tv_sec % 86400)/3600;
    m  = (t.tv_sec % 3600) / 60;
    s  = t.tv_sec % 60;
    us = t.tv_usec;

    fprintf(stderr, "** %s: at %02d:%02d.%02d %06d\n", message, h,m,s,us);
}


#ifdef USE_POLL
#include <poll.h>
#endif
int
wait_for_read(int fd)
{
    for (;;) {
#ifdef USE_POLL
	struct pollfd p;

	p.fd = fd;
	p.events = POLLIN;
	if (poll(&p, 1, -1) < 0) {
	    if (errno != EINTR) {
		perror("wait_for_read: poll");
		return -1;
	    }
	} else {
	    break;
	}
#else
	fd_set readfd;
	FD_ZERO(&readfd);
	FD_SET(fd, &readfd);
	if (select(fd+1, &readfd, NULL, NULL, NULL) < 0) {
	    if (errno != EINTR) {
		perror("wait_for_read: select");
		return -1;
	    }
	} else {
	    break;
	}
#endif
    }
    return 0;
}


// ͥåȥɤ߽
void
setint(void* buf, int value)
{
    unsigned char* p = (unsigned char*)buf;
    p[0] = (value >> 24) & 0xff;
    p[1] = (value >> 16) & 0xff;
    p[2] = (value >>  8) & 0xff;
    p[3] = (value      ) & 0xff;
}

int
getint(void* buf)
{
    unsigned char* p = (unsigned char*)buf;
    return (p[0]<<24) | (p[1]<<16) | (p[2]<<8) | p[3];
}

// λ

static int ipu_server_fd;
static int ipu_port;
static int quit_stat;
static int quit_ok = 0;
static pthread_mutex_t quit_lock;
static pthread_cond_t  quit_cond;
static pthread_mutex_t quit_reent_lock;

extern int quit_flag;
extern msgqueue send_q, trans_q, lang_q, dmod_q, cache_q;

void
quit_set(int ipu_server_fd_param, int ipu_port_param)
{
    ipu_server_fd = ipu_server_fd_param;
    ipu_port      = ipu_port_param;
}

void
quit(int st)
{
    ipu_ipu msg;

    pthread_mutex_lock(&quit_reent_lock);
    if (quit_flag) {
	// Ǥ˽λʤʤˤ⤷ʤ
	pthread_mutex_unlock(&quit_reent_lock);
	pthread_exit(0);
	return;
    }
    quit_flag = 1;

    // ipu2om
    msg.to_om = new ipu_om;
    msg.to_om->type = p_io_quit;
    send_q.putmsg(msg);

    // IPU Ȥ³
    msg.t_msg = new trans_msg;
    msg.t_msg->type = p_ii_trans_quit;
    trans_q.putmsg(msg);

    // lang0 λ
    msg.i_info = NULL;
    lang_q.putmsg(msg);

    // compile λ
    msg.p_info = NULL;
    dmod_q.putmsg(msg);

    int fd;
    if ((fd = connect_host(get_hostname(), ipu_port)) < 0) {
	printf("connect error %d!!\n", ipu_port);
    } else {
	close(fd);
    }
    close(ipu_server_fd);

    quit_stat = st;
    pthread_mutex_lock(&quit_lock);
    quit_ok = 1;
    pthread_cond_signal(&quit_cond);
    pthread_mutex_unlock(&quit_lock);

    pthread_mutex_unlock(&quit_reent_lock);

    pthread_exit(0);
}

int
quit_wait()
{
    pthread_mutex_lock(&quit_lock);
    while (!quit_ok) {
	pthread_cond_wait(&quit_cond, &quit_lock);
    }
    pthread_mutex_unlock(&quit_lock);
    return quit_stat;
}
