/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  relation.cc
*********************************************************************/

#include "comp.h"
#include "relation.h"
#include "stack.h"
#include "table.h"
#include "ObjList.h"
#include "def.h"

int
f_eq(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    complex	a, b;
    if (ST.pop(a) != OKAY) return FAIL;
    if (ST.pop(b) != OKAY) return FAIL;

    if (a == b) ST.push(ib_true);
    else	ST.push(ib_false);
    return OKAY;
}

int
f_ne(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    complex	a, b;
    if (ST.pop(a) != OKAY) return FAIL;
    if (ST.pop(b) != OKAY) return FAIL;

    if (a != b)	ST.push(ib_true);
    else	ST.push(ib_false);
    return OKAY;
}

int
f_ge(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b;
    t1 = ST.sttype();
    if (ST.pop(a) != OKAY) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) != OKAY) return FAIL;

    if (t1 == nt_complex || t2 == nt_complex)
	if (norm(a) <= norm(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    else
	if (real(a) <= real(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    return OKAY;
}

int
f_gt(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b;
    t1 = ST.sttype();
    if (ST.pop(a) != OKAY) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) != OKAY) return FAIL;

    if (t1 == nt_complex || t2 == nt_complex)
	if (norm(a) < norm(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    else
	if (real(a) < real(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    return OKAY;
}

int
f_le(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b;
    t1 = ST.sttype();
    if (ST.pop(a) != OKAY) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) != OKAY) return FAIL;

    if (t1 == nt_complex || t2 == nt_complex)
	if (norm(a) >= norm(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    else
	if (real(a) >= real(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    return OKAY;
}

int
f_lt(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t1, t2;
    complex	a, b;
    t1 = ST.sttype();
    if (ST.pop(a) != OKAY) return FAIL;
    t2 = ST.sttype();
    if (ST.pop(b) != OKAY) return FAIL;

    if (t1 == nt_complex || t2 == nt_complex)
	if (norm(a) > norm(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    else
	if (real(a) > real(b))	ST.push(ib_true);
	else			ST.push(ib_false);
    return OKAY;
}


int
f_andorxor(stack& ST, int op)
{
    node_type	t1, t2;
    ipu_p_bool	b1, b2;
    int 	i1, i2;

    t1 = ST.sttype();
    if (t1 == nt_bool) {
	// bool ξ ... 黻
	if (ST.pop(b1) != OKAY) return FAIL;
	t2 = ST.sttype();
	if (t2 == nt_bool) {
	    if (ST.pop(b2) != OKAY) return FAIL;
	    if (op == 1) {
		ST.push(((b1==ib_true) && (b2==ib_true)) ? ib_true : ib_false);
	    } else if (op == 2) {
		ST.push(((b1==ib_true) || (b2==ib_true)) ? ib_true : ib_false);
	    } else {
		ST.push((((b1==ib_true)  && (b2==ib_true) ) ||
			 ((b1==ib_false) && (b2==ib_false)))
			? ib_false : ib_true);
	    }
	    return OKAY;
	} else {
	    // äƤʤ
	    ST.push(b1);
	    return FAIL;
	}
    } else if (t1 == nt_int) {
	// int ξ ... ӥåȱ黻
	if (ST.pop(i1) != OKAY) return FAIL;
	t2 = ST.sttype();
	if (t2 == nt_int) {
	    if (ST.pop(i2) != OKAY) return FAIL;
	    if (op == 1) {
		ST.push(i1 & i2);
	    } else if (op == 2) {
		ST.push(i1 | i2);
	    } else {
		ST.push(i1 ^ i2);
	    }
	    return OKAY;
	} else {
	    // äƤʤ
	    ST.push(i1);
	    return FAIL;
	}
    } else {
	return FAIL;
    }
}

int
f_and(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    return f_andorxor(ST, 1);
}

int
f_or(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    return f_andorxor(ST, 2);
}

int
f_xor(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    return f_andorxor(ST, 3);
}

int
f_not(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    node_type	t;
    ipu_p_bool	b;
    int 	i;

    t = ST.sttype();
    if (t == nt_bool) {
	if (ST.pop(b) != OKAY) return FAIL;
	ST.push((b == ib_true) ? ib_false : ib_true);
	return OKAY;
    } else if (t == nt_int) {
	if (ST.pop(i) != OKAY) return FAIL;
	ST.push(~i);
	return OKAY;
    } else {
	return FAIL;
    }
}

int
f_false(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    ST.push(ib_false);
    return OKAY;
}

int
f_true(stack& ST, table& TA, ftable& FT, ObjectList& OL)
{
    (void)TA; (void)FT; (void)OL;
    ST.push(ib_true);
    return OKAY;
}
