/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  trans.cc
    IPU ̿
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "Image.h"
#include "GrayImage.h"
#include "ObjList.h"
#include "def.h"
#include "protocol.h"
#include "msgqueue.h"
#include "etc.h"
#include "com.h"
#include "vom.h"
#include "trans.h"
#include "array.h"
#include "ipu_list.h"
#include "waitlist.h"
#include "unify.h"
#include "priodef.h"

#ifdef DEBUG
#define DEBUG2
#endif

// IPU ̿Υإå (ǽ1)
// ID + IPU ֹ
#define IPU_HEAD_LENGTH    (8+4)
#define IPU_HEAD_ID        "IPU-IPU\n"
#define IPU_HEAD_ID_LENGTH 8

extern ObjectList	OL;
extern msgqueue		image_wait_q;
extern msgqueue		trans_q;
extern msgqueue		send_q;
extern ipu_list		ipus;
extern int 		quit_flag;
extern waitlist		array_wait_list;
extern unify_list unify_l;

#define MAXCONNECT	256

static pthread_mutex_t	lock; // max_connect, h_fd Υå
static int		max_connect = 0;
static struct host_fd {
    int  hostno;        //
    int  fd;            // һ
    int  is_server;     // Фɤ TRUE or FALSE
    int  ipu_no;
} h_fd[MAXCONNECT];

static msgqueue	write_q[MAXCONNECT];

#ifdef DEBUG2
static char*
type2string(p_ii_t type)
{
    switch (type) {
      case p_ii_image_request:		return "image_request";
      case p_ii_image_return:		return "image_return";
      case p_ii_array_request:		return "array_request";
      case p_ii_array_return:		return "array_return";
      case p_ii_array_read:		return "array_read";
      case p_ii_array_write:		return "array_write";
      case p_ii_array_lock:		return "array_lock";
      case p_ii_array_lock_done:	return "array_lock_done";
      case p_ii_array_unlock:		return "array_unlock";
      case p_ii_array_unify:		return "array_unify";
      case p_ii_part_image:		return "part_image";
      case p_ii_trans_quit:		return "trans_quit";
      default:				return NULL;
    }
}
#endif

// ꡢʬäƤؽ񤭹
static int transed_image(image_info* img_info, void* buf);

// ȥ󥹥Ф쥷֥ФƤФ
// ɤ߹ѥåɥȴؿ
// ¾ IPU Ȥ³1ĤˤĤ1ĥåɤĤ
static void
trans_read(struct host_fd* fdp)
{
    int 		hostno = fdp->hostno;
    int 		fd = fdp->fd;
    int 		ipu_no = fdp->ipu_no;
    ipu_ipu		msg;
    trans_msg		recv;
    array_id_info	a_info;
    array_data_info	ad_info;
    char*		buf;
    image_info		img_info;
    Image*		img;

    for (;;) {
	wait_for_read(fd);
	if (net_read(fd, &recv, sizeof(trans_msg)) == FAIL) {
	    fprintf(stderr, "trans_read: error in mesage read\n");
	    return;
	}

#ifdef DEBUG2
	fprintf(stderr, "==== trans_read: from ipu %d, type ", ipu_no);
	char* comstr = type2string(recv.type);
	if (comstr != NULL) {
	    fprintf(stderr, "%s\n", comstr);
	} else {
	    fprintf(stderr, "unknown (%d)\n", recv.type);
	}
#endif

	switch (recv.type) {
	  case p_ii_image_request:
	    // ž׵᤬褿
	    // ֤
	    //time_print("ž׵");
	    if (net_read(fd, &img_info, sizeof(image_info)) == FAIL) {
		fprintf(stderr, "trans_read: error in mesage read\n");
		return;
	    }
	    msg.t_msg = new trans_msg;
	    msg.t_msg->type = p_ii_image_return;
	    msg.img_info = new image_info;
	    msg.img_info->image_id = img_info.image_id;
	    msg.img_info->div_id   = img_info.div_id;

	    img = OL.get(msg.img_info->image_id, msg.img_info->div_id);
	    if (img == NULL) {
		fprintf(stderr, "trans_read: OL.get() failed\n");
		abort();
		break;
	    }
	    img->get_image_info(*msg.img_info);
	    msg.t_msg->size  = img->get_image_size();

	    write_q[hostno].putmsg(msg);
	    break;

	  case p_ii_image_return:
	    // ž׵α褿
	    // ɤ߹ OL ϿΤ餻
	    if (net_read(fd, &img_info, sizeof(image_info)) == FAIL) {
		fprintf(stderr, "trans_read: error in mesage read\n");
		return;
	    }
	    img = new GrayImage(img_info.image_id, img_info.div_id);
	    img->setparam(img_info.width, img_info.height,
			  img_info.hot_x, img_info.hot_y,
			  (pixel_type)img_info.pixel_type, img_info.cache);
	    img->setorgsize(img_info.org_width, img_info.org_height);
	    OL.append(img);
#ifdef DEBUG2
	    img->print();
#endif
	    void* imgdata;
	    imgdata = img->get_image();

	    if (net_read(fd, imgdata, recv.size) == FAIL) {
		fprintf(stderr,"trans_read: error in mesage read\n");
		break;
	    }
#ifdef DEBUG
	    fprintf(stderr, "trans_read: image_data read done %d %d\n",
		    img_info.image_id, img_info.div_id);
#endif
	    image_wait_q.putmsg(msg);
	    break;
	  case p_ii_array_request:
	    // ž׵᤬褿
	    // ֤
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_request read error\n");
		return;
	    }
	    msg.t_msg = new trans_msg;
	    msg.t_msg->type = p_ii_array_return;
	    msg.a_info = new array_id_info;
	    msg.a_info->array_id = a_info.array_id;
	    msg.a_info->sub_id   = a_info.sub_id; // ž sub_id
	    write_q[hostno].putmsg(msg);
	    break;
	  case p_ii_array_return:
	    // ž׵α褿
	    // ɤ߹ AL ϿΤ餻
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_request read error\n");
		return;
	    }
	    array* arr;
	    arr = AL.search(a_info.array_id, a_info.sub_id);
	    if (arr == NULL) {
		// ʤʤ鿷
		arr = new array((array_type)a_info.type, a_info.size);
		arr->set_array_id(a_info.array_id);
		arr->set_sub_id(a_info.sub_id);
		AL.append(arr);
	    }
	    void* arrdata;
	    arrdata = arr->get_ptr();

	    if (net_read(fd, arrdata, a_info.size * a_info.width) == FAIL) {
		fprintf(stderr,"trans_read: error in mesage read\n");
		break;
	    }
	    fprintf(stderr, "trans_read: array read done %d %d\n",
		    a_info.array_id, a_info.sub_id);
	    array_wait_list.signal(NULL, a_info.array_id, a_info.sub_id);
	    break;
	  case p_ii_array_read:
	    // ΤǤɤ߹׵᤬褿
	    // Ǥ֤
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL ||
		net_read(fd, &ad_info, sizeof(array_data_info)) == FAIL)
	    {
		fprintf(stderr, "trans_read: array_read read error\n");
		break;
	    }
	    array_read(&a_info, &ad_info, ipu_no);
	    break;

	  case p_ii_array_write:
	    // ΤǤν񤭹ߤ褿
	    // Ǥ˽񤭹
	    buf = new char[recv.size];
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL ||
		net_read(fd, &ad_info, sizeof(array_data_info)) == FAIL ||
		net_read(fd, buf, recv.size) == FAIL)
	    {
		fprintf(stderr, "trans_read: array_write read error\n");
		break;
	    }
	    array_write(&a_info, &ad_info, buf);
	    delete[] buf;
	    break;

	  case p_ii_array_lock:
	    // å׵᤬
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_lock read error\n");
		break;
	    }
	    array_lock(&a_info, ipu_no);
	    break;

	  case p_ii_array_unlock:
	    // 󥢥å׵᤬
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_unlock read error\n");
		break;
	    }
	    array_unlock(&a_info, ipu_no);
	    break;

	  case p_ii_array_lock_done:
	    // åλ
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_lock_done read error\n");
		break;
	    }
	    array_lock_done(&a_info);
	    break;

	  case p_ii_array_unlock_done:
	    // 󥢥åλ
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_unlock_done read error\n");
		break;
	    }
	    array_unlock_done(&a_info);
	    break;

	  case p_ii_array_unify:
	    //  (ǡ)
	    if (net_read(fd, &a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_read: array_unify read error\n");
		break;
	    }
	    buf = new char[recv.size];
	    if (net_read(fd, buf, recv.size) == FAIL) {
		fprintf(stderr, "trans_read: array_unify read error\n");
		break;
	    }
	    unify_l.append(a_info.array_id, buf);
	    // no delete buf
	    break;

	  case p_ii_part_image:
#ifdef DEBUG
	    // ΰƤ
	    fprintf(stderr, "trans_read: part image!\n");
#endif
	    if (net_read(fd, &img_info, sizeof(image_info)) == FAIL) {
		fprintf(stderr, "trans_read: img_info read error\n");
		break;
	    }
	    buf = new char[recv.size];
	    if (net_read(fd, buf, recv.size) == FAIL) {
		fprintf(stderr, "trans_read: img_info data read error\n");
		break;
	    }
	    transed_image(&img_info, buf);
	    delete[] buf;
	    break;

	  case p_ii_trans_quit:
	    fprintf(stderr, "trans_read: client quitting...");
	    close(fd);
	    fprintf(stderr, "Done\n");
	    pthread_exit(0);
	  default:
	    fprintf(stderr, "trans_read: unknown type %d\n", recv.type);
	    break;
	}
    }
}

// ȥ󥹥Ф쥷֥ФƤФ
// 񤭹ѥåɥȴؿ
// ¾ IPU Ȥ³1ĤˤĤ1ĥåɤĤ
static void
trans_write(struct host_fd* fds)
{
    ipu_ipu	msg;
    int 	hostno = fds->hostno;
    int 	fd = fds->fd;
    int 	ipu_no = fds->ipu_no; (void)ipu_no;

    for (;;) {
	write_q[hostno].getmsg(msg);

	if (net_write(fd, msg.t_msg, sizeof(trans_msg)) == FAIL) {
	    fprintf(stderr, "trans_write: net_write failed\n");
	}

#ifdef DEBUG2
	fprintf(stderr, "==== trans_write:  to ipu %d, type ", ipu_no);
	char* comstr = type2string(msg.t_msg->type);
	if (comstr != NULL) {
	    fprintf(stderr, "%s\n", comstr);
	} else {
	    fprintf(stderr, "unknown (%d)\n", msg.t_msg->type);
	}
#endif

	switch (msg.t_msg->type) {
	  case p_ii_image_request:
	    // ž׵ (˲äƤ褦׵᤹)
	    if (net_write(fd, msg.img_info, sizeof(image_info)) == FAIL) {
		fprintf(stderr, "trans_write: net_write failed\n");
#ifdef DEBUG
	    } else {
		fprintf(stderr, "image request write OK\n");
#endif
	    }
	    delete msg.img_info;
	    break;
	  case p_ii_image_return:
	    // ž׵Ф
	    //time_print("ž");
	    Image* img;
	    img = OL.get(msg.img_info->image_id, msg.img_info->div_id);
	    if (img == NULL) {
		fprintf(stderr, "trans_write: OL.get() failed\n");
		abort();
	    }

	    if (net_write(fd, msg.img_info, sizeof(image_info)) == FAIL) {
		fprintf(stderr, "trans_write: net_write failed\n");
	    }
	    //time_print("žλ");

	    void* imgdata;
	    imgdata = img->get_image();
	    if (net_write(fd, imgdata, msg.t_msg->size) == FAIL) {
		fprintf(stderr, "image_data write error\n");
	    }
#ifdef DEBUG
	    fprintf(stderr, "β IPU %d žޤ\n", ipu_no);
	    img->print();
	    fprintf(stderr, "image_data write done %d %d\n",
		    msg.img_info->image_id,
		    msg.img_info->div_id);
#endif
	    delete msg.img_info;
	    break;
	  case p_ii_array_request:
	    // ž׵
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_write: net_write failed\n");
	    }
	    break;
	  case p_ii_array_return:
	    // ž׵Ф
	    array* arr;
	    arr = AL.search(msg.a_info->array_id, -1);
	    if (arr != NULL) {
		msg.a_info->size  = arr->get_size();
		msg.a_info->type  = (int)arr->get_type();
		msg.a_info->width = arr->get_width();
	    } else {
		// 顼
		fprintf(stderr, "array_return, no such id %d\n",
			msg.a_info->array_id);
		msg.a_info->size  = 0;
		msg.a_info->type  = 0;
		msg.a_info->width = 0;
	    }

	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_write: net_write failed\n");
	    }

	    if (msg.a_info->size > 0) {
		void* arrdata = arr->get_ptr();
		if (net_write(fd, arrdata, msg.a_info->size*msg.a_info->width)
		    == FAIL)
		{
		    fprintf(stderr, "trans_write: array_return write error\n");
		    break;
		}
	    }
#ifdef DEBUG
	    fprintf(stderr, "arrdata %d write done (size = %d)\n",
		    msg.a_info->array_id, msg.a_info->size*msg.a_info->width);
#endif
	    break;
	  case p_ii_array_read:
	    // ΤǤɤ߹׵
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL ||
		net_write(fd, msg.ad_info, sizeof(array_data_info)) == FAIL)
	    {
		fprintf(stderr, "trans_write: array write error\n");
	    }
	    break;

	  case p_ii_array_write:
	    // ΤǤν񤭹
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL ||
		net_write(fd, msg.ad_info, sizeof(array_data_info)) == FAIL ||
		net_write(fd, msg.buf, msg.t_msg->size) == FAIL)
	    {
		fprintf(stderr, "trans_write: array write error\n");
	    }
	    break;

	  case p_ii_array_lock:
	    // å׵
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_write: array_lock write error\n");
	    }
	    break;
	  case p_ii_array_unlock:
	    // 󥢥å׵
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_write: array_unlock write error\n");
	    }
	    break;

	  case p_ii_array_unify:
	    //  (ǡ)
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL ||
		net_write(fd, msg.buf, msg.t_msg->size) == FAIL)
	    {
		fprintf(stderr, "trans_write: array_unify write error\n");
	    }
	    break;

	  case p_ii_array_lock_done:
	    // åλ
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr, "trans_write: array_lock_done write error\n");
	    }
	    break;

	  case p_ii_array_unlock_done:
	    // 󥢥åλ
	    if (net_write(fd, msg.a_info, sizeof(array_id_info)) == FAIL) {
		fprintf(stderr,"trans_write: array_unlock_done write error\n");
	    }
	    break;

	  case p_ii_part_image:
	    // ΰ
	    if (net_write(fd, msg.img_info, sizeof(image_info)) == FAIL ||
		net_write(fd, msg.buf, msg.t_msg->size) == FAIL)
	    {
		fprintf(stderr,"trans_write: img_info write error\n");
	    }
	    delete[] msg.buf;
	    delete msg.img_info;
	    break;

	  case p_ii_trans_quit:
	    fprintf(stderr, "trans server quitting...");
	    close(fd);
	    fprintf(stderr, "Done\n");
	    pthread_exit(0);
	  default:
	    fprintf(stderr, "trans_write: error\n");
	    break;
	}
	delete msg.t_msg;
    }
}

// IPU ̿κǽ˹Ԥʬ
static int
send_header(int fd)
{
    char buf[IPU_HEAD_LENGTH];
    extern int my_ipu_no;
    strcpy(buf, IPU_HEAD_ID);
    setint(buf+IPU_HEAD_ID_LENGTH, my_ipu_no);
    if (net_write(fd, buf, IPU_HEAD_LENGTH) == FAIL) return -1;
    return 0;
}
static int
receive_header(int fd)
{
    char buf[IPU_HEAD_LENGTH];
    if (net_read(fd, buf, IPU_HEAD_LENGTH) == FAIL) return -1;
    if (strncmp(buf, IPU_HEAD_ID, IPU_HEAD_ID_LENGTH) != 0) {
	return -1;
    }
    return getint(buf+IPU_HEAD_ID_LENGTH); // ipu_no ֤
}

// ȥ󥹥ХåɤΥȴؿ
// IPU ¾ΥåɤΥǡԤ
// ǡ褿餽Υǡꤹ IPU 
// ȤȥͥĥƤʤХͥĥ
void
trans_server(int* port_param)
{
    int 		port = *port_param; delete port_param;
    int 		i;
    struct ipu_ipu 	msg;

    for (;;) {
	trans_q.getmsg(msg);

	if (msg.t_msg->type == p_ii_trans_quit) {
	    // λå褿
	    pthread_mutex_lock(&lock);
	    for (i = 0; i < max_connect; i++) {
		if (h_fd[i].is_server)
		    write_q[i].putmsg(msg);
	    }
	    pthread_mutex_unlock(&lock);
	    break;
	}

	// Ǥ³Ƥ뤫Ĵ٤

	pthread_mutex_lock(&lock);

	for (i = 0; i < max_connect; i++) {
	    if (h_fd[i].ipu_no == msg.t_msg->ipu_no) {
		pthread_mutex_unlock(&lock);
		break;
	    }
	}

	if (i == max_connect) {
	    // ޤ³ƤʤΤǡ³
	    int fd;
	    char* hostname = ipus.get_hostname(msg.t_msg->ipu_no);
	    if ((fd = connect_host(hostname, port)) < 0) {
		fprintf(stderr, "trans_server: cannot connect\n");
		pthread_mutex_unlock(&lock);
		continue;
	    }
	    // h_fd Ͽ
	    h_fd[i].fd = fd;
	    h_fd[i].hostno = i;
	    h_fd[i].is_server = FALSE;
	    h_fd[i].ipu_no = msg.t_msg->ipu_no;
	    max_connect++;
	    pthread_mutex_unlock(&lock);

	    // ̿ϥإå
	    send_header(h_fd[i].fd);

#ifdef DEBUG
	    fprintf(stderr, "trans_server: connect, host = %s (ipu no %d)\n",
		    hostname, msg.t_msg->ipu_no);
#endif

	    // ɤ߽񤭥å
	    pthread_t tid;
	    pthread_attr_t attr;
	    struct sched_param s_param;
	    pthread_attr_init(&attr);
	    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	    s_param.sched_priority = PRIO_TRANS;
	    pthread_attr_setschedparam(&attr, &s_param);
	    pthread_create(&tid, &attr, (void*(*)(void*))trans_write,&h_fd[i]);
	    pthread_create(&tid, &attr, (void*(*)(void*))trans_read ,&h_fd[i]);
	}

	// ͥåλ
	// ºݤ

	write_q[i].putmsg(msg);
    }
#ifdef DEBUG2
    fprintf(stderr, "thread trans_server end\n");
#endif
}

// 쥷֥ФΥåɤΥȴؿ
// ¾ IPU ³׵Ԥ³Ƥɤ߽2ĤΥåɤ
void
receive_server(int* fd_param)
{
    int 	listen_fd = *fd_param; delete fd_param;
    int 	fd;
    int 	i;

    // ³Ԥ̵¥롼
    for (;;) {

	// ³դ
	for (;;) {
	    if ((fd = accept(listen_fd, NULL, NULL)) < 0) {
		if (quit_flag) {
		    goto end;
		}
		if (errno != EINTR) {
		    perror("receive_server: accept failed");
		    goto err;
		}
	    } else {
		break;
	    }
	}

	if (quit_flag) {
	    close(fd);
	    goto end;
	}

	// ̿ϥإå
	int ipu_no;
	if ((ipu_no = receive_header(fd)) < 0) {
	    // 
	    close(fd);
	    continue;
	}

	// h_fd Ͽ
	pthread_mutex_lock(&lock);
	i = max_connect++;
	h_fd[i].hostno = i;
	h_fd[i].fd = fd;
	h_fd[i].is_server = TRUE;
	h_fd[i].ipu_no = ipu_no;
	pthread_mutex_unlock(&lock);

	pthread_t tid;
	pthread_attr_t attr;
	struct sched_param s_param;
	pthread_attr_init(&attr);
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	s_param.sched_priority = PRIO_TRANS;
	pthread_attr_setschedparam(&attr, &s_param);
	pthread_create(&tid, &attr, (void* (*)(void*))trans_write, &h_fd[i]);
	pthread_create(&tid, &attr, (void* (*)(void*))trans_read , &h_fd[i]);
    }

    return;

  err:

    if (listen_fd >= 0)
	close(listen_fd);

    fprintf(stderr, "receive_server: faital error, cannot connect to IPU\n");
    exit(1);
    return;

  end:
#ifdef DEBUG2
    fprintf(stderr, "thread receive_server end\n");
#endif
    return;
}

// ΰä
// ʬäƤ˽񤭹
static int
transed_image(image_info* img_info, void* buf)
{
    int width    = img_info->width;
    int height   = img_info->height;
    int hot_x    = img_info->hot_x;
    int hot_y    = img_info->hot_y;
    int image_id = img_info->image_id;
    int div_id   = img_info->div_id;
    int task_id  = img_info->div_type;

    Image* img = OL.get(image_id, div_id);

    if (img == NULL) {
	fprintf(stderr, "transed_image: OL.get failed\n");
	return -1;
    }
    int x, y;
    int i = 0;
    pixel_type pt = img->get_type();
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    if (pt == pt_int) {
		img->asetpixel(hot_x + x, hot_y + y, ((int*)buf)[i]);
	    } else if (pt == pt_real) {
		img->asetpixel(hot_x + x, hot_y + y, ((float*)buf)[i]);
	    } else if (pt == pt_complex) {
		img->asetpixel(hot_x + x, hot_y + y, ((complex*)buf)[i]);
	    } else {
		fprintf(stderr, "transed_image: bad pixel type %d\n", pt);
		return -1;
	    }
	    i++;
	}
    }

    // OM 
    ipu_ipu msg;
    msg.to_om = new ipu_om;
    msg.to_om->type = p_io_sync_cache_done;
    msg.sc_info = new sync_cache_info;
    msg.sc_info->task_id  = task_id;
    msg.sc_info->image_id = image_id;
    send_q.putmsg(msg);
    return 0;
}
