/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  ModuleList.cc
    ⥸塼ꥹ (⥸塼)
*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "msgqueue.h"
#include "ModuleList.h"
#include "IpuList.h"
#include "etc.h"

extern IpuList ip;

#ifndef MODULE_DATA_FILE
#define MODULE_DATA_FILE VIOSLIBDIR "/module.dat"
#endif

ModuleList::ModuleList()
{
    next = NULL;

    name = NULL;
    in_num = out_num = param_num = -1;
    divtype = NULL;
    boxsize = NULL;
    cache = NULL;
    etime = 0;
    compile = 0;

    {
	div_type dt[10];
	int      box[20];
	int      ca[10];

	dt[0] = dt[1] = dt[2] = dt_all;
	box[0] = box[1] = box[2] = box[3] = box[4] = box[5] = 0;
	ca[0] = ca[1] = ca[2] = 0;
	append_no_write("load", 0, 1, 1, dt, box, ca, 1, 100, 1, 1);
	append_no_write("save", 1, 0, 1, dt, box, ca, 1, 100, 1, 1);
	append_no_write("disp", 1, 0, 0, dt, box, ca, 1, 100, 1, 1);

	append_no_write("cload", 0, 3, 1, dt, box, ca, 1, 100, 1, 1);
	append_no_write("csave", 3, 0, 1, dt, box, ca, 1, 100, 1, 1);
	append_no_write("cdisp", 3, 0, 0, dt, box, ca, 1, 100, 1, 1);

	append_no_write("waitarray", 0, 0, 1, dt, box, ca, 1, 100, 1, 1);
    }
    read_file();
}

ModuleList::ModuleList(ModuleList* next_, char* name_,
	int in_num_, int out_num_, int param_num_,
	div_type* divtype_, int* boxsize_, int* cache_, int rtype_,
	int etime_, int compile_, int inter_)
{
    next          = next_;
    name          = strdup_new(name_);
    in_num        = in_num_;
    out_num       = out_num_;
    param_num     = param_num_;
    etime         = etime_;
    compile       = compile_;
    internal_flag = inter_;
    rundivtype    = rtype_;

    int n = in_num_ + out_num_;

    if (n > 0) {
	divtype    = new div_type[n];
	boxsize    = new int[n*2];
	cache      = new int[n];
    } else {
	divtype    = NULL;
	boxsize    = NULL;
	cache      = NULL;
    }

    for (int i = 0; i < n; i++) {
	divtype[i]     = divtype_[i];
	boxsize[i*2]   = boxsize_[i*2];
	boxsize[i*2+1] = boxsize_[i*2+1];
	cache[i]       = cache_[i];
    }
}

ModuleList::~ModuleList()
{
    if (name    != NULL) delete[] name;
    if (divtype != NULL) delete[] divtype;
    if (boxsize != NULL) delete[] boxsize;
    if (cache   != NULL) delete[] cache;
    if (next    != NULL) delete next;
}

ModuleList*
ModuleList::search(char* name_)
{
    ModuleList* cur = next;

    while (cur != NULL) {
	if (strcmp(name_, cur->name) == 0)
	    return cur;
	cur = cur->next;
    }
    return NULL;
}

void
ModuleList::append_no_write(char* name_, int in_num_, int out_num_,
		int param_num_, div_type* divtype_, int* boxsize_,
		int* cache_, int rtype_, int etime_, int comp_, int inter_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	// ⤷Ʊ̾Υ⥸塼뤬ä
	remove(name_);
    }
    next = new ModuleList(next, name_, in_num_, out_num_, param_num_,
		divtype_, boxsize_, cache_, rtype_, etime_, comp_, inter_);
    if (comp_ != 0) ipu_dfdef(next);
}

extern msgqueue send_q[];

void
ModuleList::ipu_dfdef(int ipu_no, ModuleList* mp)
{
    om_om msg;
    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_dfdef;
    msg.p_info = new prog_info;
    msg.p_info->in_num    = mp->in_num;
    msg.p_info->out_num   = mp->out_num;
    msg.p_info->param_num = mp->param_num;
    strcpy(msg.p_info->name, mp->name);
    send_q[ipu_no].putmsg(msg);
}

void
ModuleList::ipu_dfdef(ModuleList* mp)
{
    int   i;
    for (i = 0; i < ip.get_ipu_num(); i++) {
	ipu_dfdef(i, mp);
    }
}

// ѥ֥å᥽å

void
ModuleList::append(char* name_, int in_num_, int out_num_, int param_num_,
		   div_type* divtype_, int* boxsize_, int* cache_, int rtype_,
		   int etime_, int comp_)
{
    append_no_write(name_, in_num_, out_num_, param_num_,
		    divtype_, boxsize_, cache_, rtype_, etime_, comp_, 0);
    write_file();
}

void
ModuleList::ipu_dfdef_all(int ipu_no)
{
    ModuleList* mp;
    for (mp = next; mp != NULL; mp = mp->next) {
	if (mp->compile != 0) ipu_dfdef(ipu_no, mp);
    }
}

void
ModuleList::remove_all()
{
    delete next;
    next = NULL;
}

void
ModuleList::remove(char* name_)
{
    ModuleList* cur = next;
    ModuleList* pre = this;

    while (cur != NULL) {
	if (strcmp(name_, cur->name) == 0) {
	    pre->next = cur->next;
	    cur->next = NULL;
	    delete cur;
	    write_file();
	    return;
	}
	pre = cur;
	cur = cur->next;
    }
    fprintf(stderr, "ModuleList::remove: %s not found\n", name_);
}

int
ModuleList::exist(char* name_)
{
    return (search(name_) != NULL) ? OKAY : FAIL;
}

// ϲξ᥽å

int
ModuleList::get_in_divide_types(char* name_, div_type* divtype_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	for (int i = 0; i < mp->in_num; i++)
	    divtype_[i] = mp->divtype[i];
	return OKAY;
    } else {
	fprintf(stderr, "ModuleList::get_in_divide_types: %s not found\n",
		name_);
	return FAIL;
    }
}

div_type
ModuleList::get_in_divide_type(char* name_, int i)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->in_num) {
	    return mp->divtype[i];
	} else {
	    fprintf(stderr, "ModuleList::get_in_divide_type: i overflow\n");
	    return dt_null;
	}
    } else {
	fprintf(stderr, "ModuleList::get_in_divide_type: %s not found\n",
		name_);
	return dt_null;
    }
}

int
ModuleList::get_in_box_size(char* name_, int i, int* width, int* height)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->in_num) {
	    *width  = mp->boxsize[i*2];
	    *height = mp->boxsize[i*2+1];
	    return OKAY;
	} else {
	    fprintf(stderr, "ModuleList::get_in_box_size: i overflow\n");
	    return FAIL;
	}
    } else {
	fprintf(stderr, "ModuleList::get_in_box_size: %s not found\n", name_);
	return FAIL;
    }
}

int
ModuleList::get_in_cache(char* name_, int i)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->in_num) {
	    return mp->cache[i];
	} else {
	    fprintf(stderr, "ModuleList::get_in_cache: i overflow\n");
	    return -1;
	}
    } else {
	fprintf(stderr, "ModuleList::get_in_cache: %s not found\n", name_);
	return -1;
    }
}

// ϲξ᥽å

int
ModuleList::get_out_divide_types(char* name_, div_type* divtype_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	for (int i = 0; i < mp->out_num; i++)
	    divtype_[i] = mp->divtype[mp->in_num + i];
	return OKAY;
    } else {
	fprintf(stderr, "ModuleList::get_out_divide_types: %s not found\n",
		name_);
	return FAIL;
    }
}

div_type
ModuleList::get_out_divide_type(char* name_, int i)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->out_num) {
	    return mp->divtype[mp->in_num + i];
	} else {
	    fprintf(stderr, "ModuleList::get_out_divide_type: i overflow\n");
	    return dt_null;
	}
    } else {
	fprintf(stderr, "ModuleList::get_out_divide_type: %s not found\n",
		name_);
	return dt_null;
    }
}

int
ModuleList::get_out_cache(char* name_, int i)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->out_num) {
	    return mp->cache[mp->in_num + i];
	} else {
	    fprintf(stderr, "ModuleList::get_out_cache: i overflow\n");
	    return -1;
	}
    } else {
	fprintf(stderr, "ModuleList::get_out_cache: %s not found\n", name_);
	return -1;
    }
}

int
ModuleList::get_out_box_size(char* name_, int i, int* width, int* height)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (i >= 0 && i < mp->out_num) {
	    *width  = mp->boxsize[(mp->in_num + i)*2];
	    *height = mp->boxsize[(mp->in_num + i)*2+1];
	    return OKAY;
	} else {
	    fprintf(stderr, "ModuleList::get_out_box_size: i overflow\n");
	    return FAIL;
	}
    } else {
	fprintf(stderr, "ModuleList::get_out_box_size: %s not found\n", name_);
	return FAIL;
    }
}

int
ModuleList::get_rtype(char* name_)
{
    ModuleList* mp = search(name_);
    if (mp == NULL) {
	fprintf(stderr, "ModuleList::get_rtype: %s not found\n", name_);
	return -1;
    }
    return mp->rundivtype;
}

int
ModuleList::get_in_num(char* name_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	return mp->in_num;
    } else {
	fprintf(stderr, "ModuleList::get_in_num: %s not found\n", name_);
	return -1;
    }
}

int
ModuleList::get_out_num(char* name_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	return mp->out_num;
    } else {
	fprintf(stderr, "ModuleList::get_out_num: %s not found\n", name_);
	return -1;
    }
}

int
ModuleList::get_param_num(char* name_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	return mp->param_num;
    } else {
	fprintf(stderr, "ModuleList::get_param_num: %s not found\n", name_);
	return -1;
    }
}

int
ModuleList::get_exec_time(char* name_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	return mp->etime;
    } else {
	fprintf(stderr, "ModuleList::get_exec_time: %s not found\n", name_);
	return -1;
    }
}

int
ModuleList::set_compile(char* name_, int compile_)
{
    ModuleList* mp = search(name_);
    if (mp != NULL) {
	if (mp->compile != 0 && compile_ != 0) {
	    // Ǥ˥åȤƤ
	    return 0;
	}
	mp->compile = compile_;
	write_file();
	if (compile_ != 0) ipu_dfdef(mp);
	return 0;
    } else {
	fprintf(stderr, "ModuleList::set_compile: %s not found\n", name_);
	return -1;
    }
}

// ------------ module data եɤ߽ ------------

static div_type
name2type(char* name)
{
    if (strcmp(name, "pixel")  == 0) return dt_pixel;
    if (strcmp(name, "column") == 0) return dt_column;
    if (strcmp(name, "row")    == 0) return dt_row;
    if (strcmp(name, "box")    == 0) return dt_box;
    if (strcmp(name, "all")    == 0) return dt_all;
    return dt_null;
}

static char*
type2name(div_type dt)
{
    switch (dt) {
      case dt_pixel:  return "pixel";
      case dt_column: return "column";
      case dt_row:    return "row";
      case dt_box:    return "box";
      case dt_all:    return "all";
      default:        return "unknown";
    }
}

static void
skip_line(FILE* fp)
{
    int c;
    while ((c = getc(fp)) != '\n' && c != EOF) ;
}

int
ModuleList::read_module_data_line(FILE* fp)
{
    char      name_[256];
    int       in_num_, out_num_, param_num_;
    div_type* divtype_ = NULL;
    int*      boxsize_ = NULL;
    int*      cache_   = NULL;
    int       rtype_;
    int       etime_, comp_;
    int       n;
    int       i;
    int       st = -1;

    // ⥸塼̾Ȱοɤ߹
    if (fscanf(fp, "%s%d%d%d", name_, &in_num_, &out_num_, &param_num_) != 4) {
	goto end;
    }
    n = in_num_ + out_num_;
    if (n > 0) {
	divtype_ = new div_type[n];
	boxsize_ = new int[n*2];
	cache_   = new int[n];
    }

    // divtype ɤ߹
    for (i = 0; i < n; i++) {
	char divname[256];
	if (fscanf(fp, "%s", divname) != 1) {
	    goto end;
	}
	div_type dt;
	dt = name2type(divname);
	if (dt == dt_null) {
	    goto end;
	}
	divtype_[i] = dt;
	if (dt == dt_box) {
	    int w, h;
	    if (fscanf(fp, "%d%d", &w, &h) != 2) {
		goto end;
	    }
	    boxsize_[i*2]   = w;
	    boxsize_[i*2+1] = h;
	} else {
	    boxsize_[i*2]   = 0;
	    boxsize_[i*2+1] = 0;
	}
	if (fscanf(fp, "%d", &cache_[i]) != 1) {
	    goto end;
	}
    }

    // ¹Ի֤ȥѥ֤ɤ߹
    if (fscanf(fp, "%d%d%d", &rtype_, &etime_, &comp_) != 3) {
	goto end;
    }

    // ɤ߹ߴλ

    append_no_write(name_, in_num_, out_num_, param_num_,
		    divtype_, boxsize_, cache_, rtype_, etime_, comp_, 0);
    st = 0;

  end:
    if (divtype_ != NULL) delete[] divtype_;
    if (boxsize_ != NULL) delete[] boxsize_;
    if (cache_   != NULL) delete[] cache_;
    skip_line(fp);
    return st;
}

int
ModuleList::read_file()
{
    FILE* fp;

    if ((fp = fopen(MODULE_DATA_FILE, "r")) == NULL) {
	return -1;
    }

    do {
	read_module_data_line(fp);
    } while (!feof(fp));

    fclose(fp);

    return 0;
}


int
ModuleList::write_file()
{
    FILE* fp;

    if ((fp = fopen(MODULE_DATA_FILE, "w")) == NULL) {
	perror(MODULE_DATA_FILE);
	return -1;
    }

    ModuleList* mp;
    int i, n;
    for (mp = next; mp != NULL; mp = mp->next) {
	if (mp->internal_flag) continue; // ⥸塼ϥե˽񤫤ʤ
	fprintf(fp, "%s %d %d %d",
		mp->name, mp->in_num, mp->out_num, mp->param_num);
	n = mp->in_num + mp->out_num;
	for (i = 0; i < n; i++) {
	    fprintf(fp, " %s", type2name(mp->divtype[i]));
	    if (mp->divtype[i] == dt_box) {
		fprintf(fp, " %d %d", mp->boxsize[i*2], mp->boxsize[i*2+1]);
	    }
	    fprintf(fp, " %d", mp->cache[i]);
	}
	fprintf(fp, " %d %d %d\n", mp->rundivtype, mp->etime, mp->compile);
    }

    fclose(fp);

    return 0;
}
