/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  ObjectList.cc
    δ
*********************************************************************/

#include <stdio.h>
#include <string.h>
#include "ObjectList.h"
#include "DivisionList.h"
#include "etc.h"

ObjectList::ObjectList(int i_id, int w_count, ObjectList* n):div_img()
{
    next = n;
    image_id = i_id;
    wait_count = w_count;
    pthread_mutex_init(&lock, NULL);
}

ObjectList::ObjectList():div_img()
{
    next = NULL;
    image_id = -1;
    pthread_mutex_init(&lock, NULL);
}

ObjectList::~ObjectList()
{
    delete next;
}

ObjectList*
ObjectList::search(int i_id)
{
    ObjectList* p;
    for (p = next; p != NULL; p = p->next) {
	if (p->image_id == i_id)
	    return p;
    }
    return NULL;
}

// ---- ѥ֥å᥽å ----

void
ObjectList::new_image(int i_id, int w_count)
{
    next = new ObjectList(i_id, w_count, next);
}

void
ObjectList::decr_wait_count(int i_id)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	p->wait_count--;
	return;
    }
    fprintf(stderr, "ObjectList::decr_wait_count: i_id %d not found\n", i_id);
}

int
ObjectList::get_wait_count(int i_id)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->wait_count;
    }
    fprintf(stderr, "ObjectList::get_wait_count: i_id %d not found\n", i_id);
    return 0;
}

void
ObjectList::append(int i_id, int d_id, div_type d_type, int w, int h,
		   int hx, int hy, int ca, int ipu_no)
    // i_id:    ID
    // d_id:   ʬ ID
    // d_type: ʬ䥿
    // w:      
    // h:      ι⤵
    // hx:     ۥåȥݥ
    // hy:     ۥåȥݥ
    // ca:     
    // ipu_no: äƤ IPU ֹ
{
    pthread_mutex_lock(&lock);
    ObjectList* p;
    for (p = next; p != NULL; p = p->next) {
	pthread_mutex_lock(&p->lock);
	if (p->image_id == i_id) {
	    p->div_img.append(d_id, d_type, w, h, hx, hy, ca, ipu_no);
	    pthread_mutex_unlock(&p->lock);
	    pthread_mutex_unlock(&lock);
	    return;
	}
	pthread_mutex_unlock(&p->lock);
    }
    fprintf(stderr, "ObjectList::append: i_id %d not found\n", i_id);
    pthread_mutex_unlock(&lock);
    return;
}

void
ObjectList::append(int i_id, int d_id, int ipu_no)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	p->div_img.append(d_id, ipu_no);
    }
}

void
ObjectList::remove_all_object()
{
    ObjectList* cur;
    while (next != NULL){
	cur = next;
	next = next->next;
	cur->next = NULL;
	delete cur;
    }
}

void
ObjectList::remove_host(int i_id, div_type d_type, int d_id, int ipu_no)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	p->div_img.remove_host(d_type, d_id, ipu_no);
	return;
    }
    fprintf(stderr, "ObjectList::remove_host: i_id %d not found\n", i_id);
}

void
ObjectList::remove_objpart(int i_id, div_type d_type, int d_id)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	p->div_img.remove_objpart(d_type, d_id);
	return;
    }
    fprintf(stderr, "ObjectList::remove_objpart: i_id %d not found\n", i_id);
}

void
ObjectList::remove_division(int i_id, div_type d_type)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	p->div_img.remove_division(d_type);
	return;
    }
    fprintf(stderr, "ObjectList::remove_division: i_id %d not found\n", i_id);
}

void
ObjectList::remove_object(int i_id)
{
    ObjectList* cur;
    ObjectList* pre;
    for (cur = next, pre = this; cur != NULL; pre = cur, cur = cur->next) {
	if (cur->image_id == i_id) {
	    pre->next = cur->next;
	    cur->next = NULL;
	    delete cur;
	    return;
	}
    }
    fprintf(stderr, "ObjectList::remove_object: i_id %d not found\n", i_id);
}

// ʬΥۥȤˤ뤫ɤ
//  OKAY ʤ FAIL ֤
int
ObjectList::obj_part_exist_this_host(int i_id, int d_id, int ipu_no)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.obj_part_exist_this_host(d_id, ipu_no);
    }
    return FAIL;
}

// ʬäƤۥ֤̾
//  OKAY ʤ FAIL ֤
int
ObjectList::get_part_image_host(int i_id, div_type d_type, int d_id,
				int* ipu_no_ret)
{
    ObjectList* p;
    if ((p = search(i_id)) != NULL) {
	return p->div_img.get_part_image_host(d_type, d_id, ipu_no_ret);
    }
    return FAIL;
}

// ʬäƤۥ֤̾
//  OKAY ʤ FAIL ֤
int
ObjectList::get_part_image_host(int i_id, int d_id, int* ipu_no_ret)
    // i_id:        ID ()
    // d_id:       ʬ ID ()
    // ipu_no_ret: IPU ۥ̾ ()
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.get_part_image_host(d_id, ipu_no_ret);
    }
    fprintf(stderr, "ObjectList::get_part_image_host: i_id %d is not found\n",
	    i_id);
    return FAIL;
}

// ʬäƤۥ֤̾
//  OKAY ʤ FAIL ֤
int
ObjectList::get_part_image_host(int i_id, div_type d_type, int x, int y,
				int* ipu_no_ret, int* div_id_ret)
{
    ObjectList* p;
    if ((p = search(i_id)) != NULL) {
	return p->div_img.get_part_image_host(d_type, x, y, ipu_no_ret,
					      div_id_ret);
    }
    return FAIL;
}

// βΥۥåȥݥȤ֤
int
ObjectList::get_image_hot(int i_id, div_type d_type, int d_id, int* hx,int* hy)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.get_image_hot(d_type, d_id, hx, hy);
    }
    fprintf(stderr, "ObjectList::get_image_hot: i_id %d is not found\n",
	    i_id);
    return FAIL;
}

// β礭֤
int
ObjectList::get_image_size(int i_id, div_type d_type, int d_id, int* w, int* h)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.get_image_size(d_type, d_id, w, h);
    }
    fprintf(stderr, "ObjectList::get_image_size: i_id %d is not found\n",
	    i_id);
    return FAIL;
}

// β cache ֤
int
ObjectList::get_image_cache(int i_id, div_type d_type, int d_id, int* ca)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.get_image_cache(d_type, d_id, ca);
    }
    fprintf(stderr, "ObjectList::get_image_cache: i_id %d is not found\n",
	    i_id);
    return FAIL;
}

// βλʬñ̤뤫ɤ
//  OKAY ʤ FAIL ֤
int
ObjectList::DivideTypeExist(int i_id, div_type d_type)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.DivideTypeExist(d_type);
    }
    fprintf(stderr, "ObjectList::DivideType: i_id %d not found\n", i_id);
    return FAIL;
}

// β뤫ɤ  OKAY ʤ FAIL
int
ObjectList::ObjExist(int i_id)
{
    if (search(i_id) != NULL) {
	return OKAY;
    }
    return FAIL;
}

// βλʬñ̤뤫ɤ֤
// 0:¸ߤʤ 1: 2:¸ߤ
int
ObjectList::get_divide_type_exist(int i_id, div_type d_type)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.get_divide_type_exist(d_type);
    }
    fprintf(stderr, "ObjectList::get_divide_type_exist: i_id %d not found\n",
	    i_id);
    return 0;
}

// βλʬñ̤¸ߤ򥻥åȤ
int
ObjectList::set_divide_type_exist(int i_id, div_type d_type, int exist)
{
    ObjectList* p = search(i_id);
    if (p != NULL) {
	return p->div_img.set_divide_type_exist(d_type, exist);
    }
    fprintf(stderr, "ObjectList::set_divide_type_exist: i_id %d not found\n",
	    i_id);
    return FAIL;
}

// ǥХå
void
ObjectList::print(int indent_level)
{
    fprintf(stderr, "ObjectList::print() begin\n");
    if (next != NULL) {
	next->print_sub(indent_level);
    }
    fprintf(stderr, "ObjectList::print() end\n");
}

void
ObjectList::print_sub(int indent_level)
{
    int i;
    if (next != NULL) next->print_sub(indent_level);
    for (i = 0; i < indent_level; i++) fprintf(stderr, " ");
    fprintf(stderr, "image_id = %d\n", image_id);
    for (i = 0; i < indent_level; i++) fprintf(stderr, " ");
    fprintf(stderr, "div_img (image id %d) begin\n", image_id);
    div_img.print(indent_level+1);
    for (i = 0; i < indent_level; i++) fprintf(stderr, " ");
    fprintf(stderr, "div_img (image id %d) end\n", image_id);
}
