/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  comvpe.cc
    OM <--> VPE ̿
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <pthread.h>
#include "protocol.h"
#include "msgqueue.h"
#include "com.h"
#include "comvpe.h"
#include "vios.h"
#include "IpuList.h"
#include "etc.h"

extern IpuList ip;

static void om2vpe(int* fdp);
static void vpe2om(int* fdp);

extern msgqueue com_q, send_q[64], vpe_send_q, sche_q, done_q;

// VPE ̿륹åɤ
void
comm_vpe(int fd)
{
    int* fdp1 = new int;
    int* fdp2 = new int;
    *fdp1 = *fdp2 = fd;

    pthread_t tid;
    pthread_attr_t attr;
    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
    pthread_create(&tid, &attr, (void* (*)(void*))om2vpe, fdp1);
    pthread_create(&tid, &attr, (void* (*)(void*))vpe2om, fdp2);
    pthread_attr_destroy(&attr);
}

// OM  VPE  message ؿ
// åɥȴؿ
#ifdef DEBUG
static char* omvpe_type2name(int type);
#endif
static void
om2vpe(int* vpe_fd_p)
{
    om_om	msg;
    int 	fd = *vpe_fd_p;
    char	buf[8];
    delete vpe_fd_p;

#ifdef DEBUG
    fprintf(stderr, "### Thread for VPE sender\n");
#endif

    for (;;) {
	vpe_send_q.getmsg(msg);

#ifdef DEBUG
	char* type_name;
	type_name = omvpe_type2name(msg.to_vpe->type);
	fprintf(stderr, "---- OM -> VPE    message type = ");
	if (type_name != NULL) {
	    fprintf(stderr, "%s\n", type_name);
	} else {
	    fprintf(stderr, "unkown(%d)\n", msg.to_vpe->type);
	}
#endif

	if (msg.to_vpe->type == p_ov_close) {
	    delete msg.to_vpe;
	    return;
	}

	setint(buf + 0, msg.to_vpe->type);
	setint(buf + 4, msg.to_vpe->size);
	net_write(fd, buf, 8);

	switch (msg.to_vpe->type) {
	  case p_ov_main_translate_stat:
	  case p_ov_module_translate_stat:
	  case p_ov_debug_translate_stat:
	  case p_ov_debug_translate:
	  case p_ov_compile_stat:
	  case p_ov_ipu_create:
	  case p_ov_load_data:
	  case p_ov_module_done:
	  case p_ov_image:
	    net_write(fd, msg.buf, msg.to_vpe->size);
	    delete   msg.to_vpe;
	    delete[] msg.buf;
	    break;
	  default:
	    fprintf(stderr, "om2vpe: unknown type %d\n", msg.to_vpe->type);
	    break;
	}
    }
}

#ifdef DEBUG
static char*
omvpe_type2name(int type)
{
    switch (type) {
      case p_ov_close:			return "close";
      case p_ov_ipu_create:		return "ipu_create";
      case p_ov_load_data:		return "load_data";
      case p_ov_main_translate_stat:	return "main_translate_stat";
      case p_ov_module_translate_stat:	return "module_translate_stat";
      case p_ov_debug_translate_stat:	return "debug_translate_stat";
      case p_ov_debug_translate:	return "debug_translate";
      case p_ov_compile_stat:		return "compile_stat";
      case p_ov_ipu_p_main:		return "ipu_p_main";
      case p_ov_image_reply:		return "image_reply";
      case p_ov_variable_reply:		return "variable_reply";
      case p_ov_module_done:		return "module_done";
      case p_ov_image:			return "image";
      default:				return NULL;
    }
}
#endif

// VPE  message ؿ
// åɥȴؿ
#ifdef DEBUG
static char* vpeom_type2name(int type);
#endif
static void
vpe2om(int* vpe_fd_p)
{
    vpe_om*	vpe_msg;
    char*	prog;
    om_om	msg;
    int 	i;
    int 	fd = *vpe_fd_p;

    delete vpe_fd_p;

    /* message read */
#ifdef DEBUG
    fprintf(stderr, "### Thread for VPE receiver\n");
#endif

    for (;;) {
	vpe_msg = new vpe_om;
	unsigned char buf[8];
	if (net_read(fd, buf, 8) == FAIL) {
	    fprintf(stderr, "vpe2om: net_read failed\n");
	    exit(1);
	}
	vpe_msg->type = (p_vo_t)getint(buf + 0);
	vpe_msg->size =         getint(buf + 4);

#ifdef DEBUG
	char* type_name;
	type_name = vpeom_type2name(vpe_msg->type);
	fprintf(stderr, "---- VPE -> OM    message type = ");
	if (type_name != NULL) {
	    fprintf(stderr, "%s\n", type_name);
	} else {
	    fprintf(stderr, "unknown(%d)\n", vpe_msg->type);
	}
#endif

	switch (vpe_msg->type) {
	  case p_vo_quit:
	    // VPE 齪λ׵

	    delete vpe_msg;

	    // IPU ؽλå
	    int ipu_num;
	    ipu_num = ip.get_ipu_num();

	    for (i = 0; i < ipu_num; i++) {
		msg.to_ipu = new om_ipu;
		msg.to_ipu->type = p_oi_quit;
		send_q[i].putmsg(msg);
	    }

	    // ѥؽλå
	    msg.from_vpe = new vpe_om;
	    msg.from_vpe->type = p_vo_quit;
	    com_q.putmsg(msg);

	    // 塼ؽλå
	    msg.done = new ipu_om;
	    msg.done->type = p_io_quit;
	    sche_q.putmsg(msg);

	    // done_report åɤؽλå
	    msg.done = new ipu_om;
	    msg.done->type = p_io_quit;
	    done_q.putmsg(msg);

	    // VPE Ȥ³
	    sleep(1);
	    close(fd);

#ifdef DEBUG
	    fprintf(stderr, "### Thread for VPE receiver end\n");
#endif

	    exit(0);
	    break;

	  case p_vo_close:
	    // VPE ׵
	    // åɽλ
	    {
		delete vpe_msg;
		msg.to_vpe = new om_vpe;
		msg.to_vpe->type = p_ov_close;
		msg.to_vpe->size = 0;
		vpe_send_q.putmsg(msg);
		return;
	    }
	    break;

	  case p_vo_vpe_p_main:
	  case p_vo_vpe_p_module:
	  case p_vo_vpe_p_debug:
	    // VPE  VPE-p ǽ񤫤줿ץबƤΤ
	    // ɤ߹ߡcom_q 
	    prog = new char[vpe_msg->size + 1];
	    if (net_read(fd, prog, vpe_msg->size) == FAIL) {
		fprintf(stderr, "vpe2om: net_read failed\n");
		exit(1);
	    }
	    prog[vpe_msg->size] = '\0';
	    msg.from_vpe = vpe_msg;
	    msg.buf = prog;
	    com_q.putmsg(msg);
	    break;
	  case p_vo_ipu_p:
	    // VPE  IPU-p ǽ񤫤줿ץबƤΤ
	    // ɤ߹ߡsend_q[0] 
	    prog = new char[vpe_msg->size + 1];
	    if (net_read(fd, prog, vpe_msg->size) == FAIL) {
		fprintf(stderr, "vpe2om: net_read failed\n");
		exit(1);
	    }
	    prog[vpe_msg->size] = '\0';
	    msg.buf = prog;
	    msg.to_ipu = new om_ipu;
	    msg.to_ipu->type = p_oi_ipu_p;
	    msg.to_ipu->size = vpe_msg->size;
	    send_q[0].putmsg(msg);
	    delete vpe_msg;
	    break;
	  default:
	    fprintf(stderr, "vpe2om: protocol error, message type=%d\n",
		    vpe_msg->type);
	    exit(1);
	}
    }
    close(fd);

    exit(0);
}

#ifdef DEBUG
static char*
vpeom_type2name(int type)
{
    switch (type) {
      case p_vo_quit:			return "quit";
      case p_vo_close:			return "close";
      case p_vo_vpe_p_main:		return "vpe_p_main";
      case p_vo_vpe_p_module:		return "vpe_p_module";
      case p_vo_vpe_p_debug:		return "vpe_p_debug";
      case p_vo_ipu_p:			return "ipu_p";
      case p_vo_variable_request:	return "variable_request";
      default:				return NULL;
    }
}
#endif
