/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  done_report.cc
    IPU 
*********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <sys/time.h>
#include <unistd.h>
#include "msgqueue.h"
#include "vios.h"
#include "schedule.h"
#include "protocol.h"
#include "done_report.h"
#include "vpe_p.h"
#include "ObjectList.h"
#include "ModuleList.h"
#include "IpuList.h"
#include "VarList.h"
#include "etc.h"
#include "TaskData.h"
#include "sync_cache.h"

extern IpuList		ip;
extern VarList		va;
extern VarList		va_array;
extern ModuleList	mo;
extern ObjectList	ob;
extern TaskData		td;
extern msgqueue		done_q, sche_q, send_q[], vpe_send_q;

//###############################################
//	done reportν
//###############################################

// load, split ʤɤΥ⥸塼˴ؤ done report
// 桼Υ⥸塼ޤ
static void
module(int ipu, ipu_om* msg_param, out_info* out, image_info* img_info)
{
    int 	width, height;
    int 	hot_x, hot_y;
    ipu_om& 	msg = *msg_param;
    out_info&	o_info = *out;
    char*	hostname = ip.get_hostname(ipu);
    div_type	d_type;

    if (msg.result != OKAY) {
	fprintf(stderr,"module: `%s' error at %s\n",
		o_info.out_module, hostname);
	return;
    }

#ifdef DEBUG
    fprintf(stderr, "#### module `%s' done at %s\n",
	    o_info.out_module, hostname);
#endif

    for (int i = 0; i < o_info.out_num; i++) {
	width  = img_info[i].width;
	height = img_info[i].height;
	hot_x  = img_info[i].hot_x;
	hot_y  = img_info[i].hot_y;

	if (strcmp(o_info.out_module, "split") == 0) {
	    d_type = (div_type)img_info[i].div_type;
	} else if (strcmp(o_info.out_module, "join") == 0) {
	    d_type = dt_all;
	} else if (img_info[i].div_id == -1) {
	    d_type = dt_all;
	} else {
	    d_type = mo.get_out_divide_type(o_info.out_module, i);
	}

#ifdef DEBUG
	fprintf(stderr,
		"done data: dtype:%s %dx%d (%d,%d) ca:%d %s iid:%d did:%d\n",
		dt2string(d_type), img_info[i].width, img_info[i].height,
		img_info[i].hot_x, img_info[i].hot_y,
		img_info[i].cache, hostname,
		img_info[i].image_id, img_info[i].div_id);
#endif
	ob.append(img_info[i].image_id, img_info[i].div_id, d_type,
		  width, height, hot_x, hot_y,
		  img_info[i].cache, ipu);

	if (strcmp(o_info.out_module, "load") == 0 ||
	    strcmp(o_info.out_module, "cload") == 0 ||
	    strcmp(o_info.out_module, "split") == 0 ||
	    strcmp(o_info.out_module, "join") == 0)
	{
	    ob.set_divide_type_exist(img_info[i].image_id, d_type, 2);
	}
    }

    // 塼
    om_om tmp;
    tmp.done = new ipu_om;
    tmp.done->type = p_io_module;
    tmp.o_info = new out_info;
    *tmp.o_info = o_info;
    sche_q.putmsg(tmp);
}

// IPU ưƤۥȤ٤򹹿
static void
remote_stat(int result, int ipu_no, int r_load)
{
    if (result == OKAY) {
	ip.set_ruser_load(ipu_no, r_load);

	// VPE پ
	om_om msg;
	msg.to_vpe = new om_vpe;
	msg.to_vpe->type = p_ov_load_data;
	msg.to_vpe->size = 8;
	msg.buf = new char[8];
	setint(msg.buf + 0, ipu_no);
	setint(msg.buf + 4, r_load);
	vpe_send_q.putmsg(msg);
    }
#ifdef DEBUG
    fprintf(stderr,"rstat=%d host \"%s\"\n",
	    ip.get_ruser_load(ipu_no), ip.get_hostname(ipu_no));
#endif
}

#define MAXTASK	1024

static void
sync_done(int ipu, int task_id, int ip1_num)
{
    om_om	msg;

    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_sync;
    msg.s_info = new sync_info;
    msg.s_info->task_id = task_id;
    msg.s_info->ip1_num = ip1_num;

    send_q[ipu].putmsg(msg);
}

static void
module_sync(int task_id)
{
    static int	sync_count[MAXTASK];

    sync_count[task_id]++;

    if (sync_count[task_id] == td.get_running_number(task_id)) {
	// sync λ
	int ip1_n = ip.get_total_ip1_num();
	sync_count[task_id] = 0;
	int ipu, ip1;
	ipu = 0;

	ipu = 0;
	ip1 = 0;
	for (int i = 0; i < ip1_n; i++) {
	    sync_done(ipu, task_id, i);
	    ip1++;
	    if (ip1 >= ip.get_ip1_num(ipu)) {
		ip1 = 0;
		ipu++;
	    }
	}
    }
}


// array ID
static int
get_new_array_id()
{
    static int	array_id = 0; // array_id 
    static pthread_mutex_t array_id_lock;

    pthread_mutex_lock(&array_id_lock);
    int id = array_id++;
    pthread_mutex_unlock(&array_id_lock);

    return id;
}

// IPU 鿷 array ׵
static void
send_array_id(int ipu)
{
    om_om	msg;
    int 	id;

    //  ID 
    id = get_new_array_id();

    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_array_id_return;
    msg.to_ipu->size = id;
    send_q[ipu].putmsg(msg);

    // ꥹȤϿ
    va_array.append(id, ipu);
}

// IPU 鿷 ID ׵
static void
send_image_id(int ipu)
{
    static int	image_id = 0; // image_id 
    om_om	msg;
    int 	id;

    //  ID 
    id = image_id++;

    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_image_id_return;
    msg.to_ipu->size = id;
    send_q[ipu].putmsg(msg);

    // ꥹȤϿ
    ob.new_image(id, -1);
}

#define jobid_to_ip1no(j)		((j) / 10000)
#define jobid_to_taskid(j)		((j) % 10000)
#define taskid_ip1no_to_jobid(t,i)	\
	(((t >= 10000) && taskid_error()), (i)*10000 + (t))
#include <stdlib.h>
static int
taskid_error()
{
    fprintf(stderr, "task id overflow\n");
    abort();
    return 0;
}

// Ƥ ip1 ʳ ip1  parallel_signal 
static void
send_para_signal(int job_id, int para_id)
{
    om_om	msg;
    int 	i;
    int 	ip1;
    int 	ipu;
    int 	from_ip1 = jobid_to_ip1no(job_id);
    int 	task_id  = jobid_to_taskid(job_id);
    int 	to_jobid;

    ipu = 0;
    ip1 = 0;
    for (i = 0; i < ip.get_total_ip1_num(); i++) {
	if (i != from_ip1) {
	    to_jobid = taskid_ip1no_to_jobid(task_id, i);
	    msg.to_ipu = new om_ipu;
	    msg.to_ipu->type = p_oi_signal_para;
	    msg.para_info = new para_id_info;
	    msg.para_info->job_id = to_jobid;
	    msg.para_info->para_id = para_id;
	    send_q[ipu].putmsg(msg);
	}

	ip1++;
	if (ip1 >= ip.get_ip1_num(ipu)) {
	    ip1 = 0;
	    ipu++;
	}
    }
}


// ⥸塼ǤΥХѿ

static struct {
    int array_id;
    int flag;
} global_ids[1000]; // Ȥꤢ

static void
create_global_var(int ipu, array_id_info* a_info)
{
    int job_id;
    int task_id;
    int ip1_no;
    int arr_no; // ⥸塼ǤΥХѿֹ
    int idx;

    job_id  = a_info->job_id;
    task_id = jobid_to_taskid(job_id);
    ip1_no  = jobid_to_ip1no(job_id);
    arr_no  = a_info->sub_id;

    idx = task_id*10 + arr_no;
    if (arr_no >= 10) {
	fprintf(stderr, "create_global_var: bad arr_no\n");
	exit(1);
    }
    if (global_ids[idx].flag == 0) {
	// ǽ׵
	global_ids[idx].array_id = get_new_array_id();
	global_ids[idx].flag = 1;
    }

    // ID ֤ (a_info ХåեϤΤޤ޻Ȥ)
    a_info->array_id = global_ids[idx].array_id;
    a_info->sub_id   = ip1_no;
    a_info->job_id   = ip.get_total_ip1_num();
    a_info->type     = arr_no; // Ȥꤢ
    a_info->width    = job_id; //
    om_om msg;
    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_global_var_id;
    msg.a_info = a_info;
    send_q[ipu].putmsg(msg);
}

// βɤ IPU äƤ뤫Ĵ٤
// ߤĤ OKAYߤĤʤ FAIL ֤
static int
search_image_ipu(int image_id, div_type d_type, int x, int y,
		 int* ipu_ret, int* div_id_ret)
{
    int w, h;
    if (ob.get_image_size(image_id, dt_all, -1, &w, &h) == OKAY) {
	if (x < 0) x = 0;
	if (x >= w) x = w - 1;
	if (y < 0) y = 0;
	if (y >= h) y = h - 1;
    }
    return ob.get_part_image_host(image_id, d_type, x, y, ipu_ret, div_id_ret);
}

// ʬΰ賰β׵
static void
image_request(int ipu, om_om& msg)
{
    int job_id = msg.img_info->job_id;
    int task_id = jobid_to_taskid(job_id);
    int image_no = msg.img_info->image_no;
    div_type d_type;
    int inimg_number = td.get_inimg_number(task_id);
    if (image_no < inimg_number) {
	d_type = td.get_in_divide_type(task_id, image_no);
    } else {
	d_type = td.get_out_divide_type(task_id, image_no - inimg_number);
    }
    int image_id = msg.img_info->image_id;
    int x        = msg.img_info->hot_x;
    int y        = msg.img_info->hot_y;
    int st;
    int image_ipu, div_id;
    st = search_image_ipu(image_id, d_type, x, y, &image_ipu, &div_id);
    if (st == OKAY) {
	msg.to_ipu = new om_ipu;
	msg.to_ipu->type = p_oi_image_request;
	// msg.img_info ϤΤޤ޻Ȥ
	msg.img_info->div_id = div_id;
	msg.img_info->image_no = ipu; //  IPU ֹƤ
	send_q[image_ipu].putmsg(msg);
    } else {
	fprintf(stderr, "image_id %d, (%d, %d) not found\n",
		image_id, x, y);
    }
}

// ʬΰ賰β׵α
static void
image_return(om_om& msg)
{
    int to_ipu = msg.img_info->image_no;
    msg.to_ipu = new om_ipu;
    msg.to_ipu->type = p_oi_image_return;
    msg.to_ipu->size = msg.done->size;
    // msg.img_info msg.buf ϤΤޤ޻Ȥ
    send_q[to_ipu].putmsg(msg);
}

// 桼̿
static void
usercom(om_om& msg)
{
    //  IPU 
    int task_id  = msg.uc_info->task_id;
    div_type d_type;
    int outimg_number = td.get_outimg_number(task_id);
    int inimg_number = td.get_inimg_number(task_id);
    int image_id = msg.uc_info->image_id;
    if (outimg_number > 0 && 0) {
	//image_id = td.GetOutimgId(task_id, 0);
	d_type = td.get_out_divide_type(task_id, 0);
    } else if (inimg_number > 0) {
	//image_id = td.GetInimgId(task_id, 0);
	d_type = td.get_in_divide_type(task_id, 0);
    } else {
	fprintf(stderr, "usercom: ERROR, no image found\n");
	return;
    }
    int x = msg.uc_info->to_x;
    int y = msg.uc_info->to_y;
#ifdef DEBUG
    fprintf(stderr,"%d:%d\n",x,y);
#endif
    int st;
    int image_ipu, div_id;
    int w, h;
    if (ob.get_image_size(image_id, dt_all, -1, &w, &h) == OKAY) {
     	if (x < 0 || x >= w || y < 0 || y >= h) {
	 /*  if(msg.uc_info->from_y == 174 && msg.uc_info->from_x == 149) {
	    fprintf(stderr,"Ƥ衢ǤϤ߽Фä(;_;)\n");
	  } */
	    // Ϥ߽ФƤ
	  return;
	}
    }
    st = search_image_ipu(image_id, d_type, x, y, &image_ipu, &div_id);
    if (st == OKAY) {
      msg.to_ipu = new om_ipu;
      msg.to_ipu->type = p_oi_usercom;
      msg.to_ipu->size = msg.done->size;
      // msg.uc_info msg.buf ϤΤޤ޻Ȥ
      send_q[image_ipu].putmsg(msg);
    } else {
	fprintf(stderr, "usercom: image_id %d, (%d, %d) not found\n",
		image_id, x, y);
    }
}

// done report åɥȴؿ
void*
done_report(void* param)
{
    (void)param;

    int 		ipu;
    om_om		msg;

#ifdef DEBUG
    fprintf(stderr, "### Thread for done report\n");
#endif

    for (;;) {
	ipu = done_q.getmsg(msg);

	if (msg.done->type == p_io_quit) {
	    delete msg.done;
	    break;
	}

	switch (msg.done->type) {
	  case p_io_ipu_create:
	    fprintf(stderr, "done_report: protocol error (ipu_create)\n");
	    break;
	  case p_io_module:
	    // ⥸塼μ¹Ԥä
	    char buf[256];
	    sprintf(buf, "module `%s' done", msg.o_info->out_module);
	    time_print(buf);
	    module(ipu, msg.done, msg.o_info, msg.img_info);
	    remote_stat(msg.done->result, ipu, msg.done->r_stat);
	    ip.sub_ipu_load(ipu);
	    if (msg.done->size != 0)
		delete[] msg.img_info;
	    delete msg.o_info;
	    break;
	  case p_io_fdef:
	    // IPU ưŪ⥸塼˴ؤ done report
	    // msg.buf ˤϥ⥸塼̾äƤ (new char[size])
	    remote_stat(msg.done->result, ipu, msg.done->r_stat);
	    if (msg.done->result == OKAY) {
		mo.set_compile(msg.buf, 1);
	    } else {
		mo.remove(msg.buf);
	    }
#ifdef DEBUG
	    fprintf(stderr,
		    "done_report: fdef received from %s, modue %s, %s\n",
		    ip.get_hostname(ipu), msg.buf,
		    msg.done->result == OKAY ? "succeeded" : "failed");
#endif
	    delete[] msg.buf;
	    break;
	  case p_io_compile_stat:
	    // ⥸塼륳ѥ
	    if (msg.done->result == 0) {
		mo.set_compile(msg.p_info->name, 1);
	    } else {
		mo.remove(msg.p_info->name);
	    }

	    // VPE 
	    // Ƭ4Х: ѥνλ
	    // Ĥ:        ѥΥå
	    {
		char* msg_buf_tmp = msg.buf;
		msg.to_vpe = new om_vpe;
		msg.to_vpe->type = p_ov_compile_stat;
		msg.to_vpe->size = 4 + msg.done->size;
		msg.buf = new char[4 + msg.done->size];
		setint(msg.buf, msg.done->result);
		memcpy(msg.buf+4, msg_buf_tmp, msg.done->size);
		delete[] msg_buf_tmp;
		vpe_send_q.putmsg(msg);
	    }
	    break;
	  case p_io_stat:
	    remote_stat(msg.done->result, ipu, msg.done->r_stat);
	    break;
	  case p_io_mod_sync:
	    module_sync(msg.done->result); // result  task_id
	    break;
	  case p_io_module_done:
	    // ⥸塼¹Խλä֤
	    // VPE 
	    msg.to_vpe = new om_vpe;
	    msg.to_vpe->type = p_ov_module_done;
	    msg.to_vpe->size = msg.done->size;
	    // msg.buf ϤΤޤ޻Ȥ
	    vpe_send_q.putmsg(msg);
	    break;
	  case p_io_image:
	    // VPE 
	    msg.to_vpe = new om_vpe;
	    msg.to_vpe->type = p_ov_image;
	    msg.to_vpe->size = msg.done->size;
	    // msg.buf ϤΤޤ޻Ȥ
	    vpe_send_q.putmsg(msg);
	    break;
	  case p_io_dsync:
	    break;
	  case p_io_array_id_request:
	    // 󿷵׵
	    send_array_id(ipu);
	    break;
	  case p_io_image_id_request:
	    //  ID ׵
	    send_image_id(ipu);
	    break;
	  case p_io_signal_para:
	    // parallel signal
	    send_para_signal(msg.para_info->job_id, msg.para_info->para_id);
	    delete msg.para_info;
	    break;
	  case p_io_create_global_var:
	    // ⥸塼ǤΥХѿ
	    create_global_var(ipu, msg.a_info);
	    // no delete msg.a_info;
	    break;

	  case p_io_image_request:
	    image_request(ipu, msg);
	    // no delete msg.img_info
	    break;
	  case p_io_image_return:
	    image_return(msg);
	    // no delete msg.buf, msg.img_info
	    break;
	  case p_io_usercom:
	    usercom(msg);
	    // no delete msg.buf, msg.uc_info
	    break;

	  case p_io_sync_cache:
	    sync_cache(ipu, msg);
	    // no delete msg.sc_info
	    break;
	  case p_io_sync_cache_done:
	    sync_cache_done(msg.sc_info->task_id, msg.sc_info->image_id);
	    delete msg.sc_info;
	    break;

	  default:
	    fprintf(stderr, "done_report: unknown type %d\n",
		    msg.done->type);
	}
	delete msg.done;
    }

    return 0;
}
