/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  protocol.h (VPE, OM, IPU )
*********************************************************************/

#ifndef PROTOCOL_H
#define PROTOCOL_H

#ifdef __cplusplus
# define dummy_union_name
#endif

#define OM_PORT		9998
#define IPU_PORT	9999

/* VPE -> OM */
typedef enum p_vo {
    p_vo_quit,
    p_vo_close,
    p_vo_vpe_p_main,
    p_vo_vpe_p_module,
    p_vo_vpe_p_debug,
    p_vo_ipu_p,
    p_vo_variable_request
} p_vo_t;

/* OM -> VPE */
typedef enum p_ov {
    p_ov_close,
    p_ov_ipu_create,
    p_ov_load_data,
    p_ov_main_translate_stat,
    p_ov_module_translate_stat,
    p_ov_debug_translate_stat,
    p_ov_debug_translate,
    p_ov_compile_stat,
    p_ov_ipu_p_main,
    p_ov_image_reply,
    p_ov_variable_reply,
    p_ov_module_done,    /* ¹Ի: int() int(ޥ) */
    p_ov_image           /* ɽ */
} p_ov_t;

/* OM -> IPU */
typedef enum p_oi {
    p_oi_quit,
    p_oi_ipu_create,
    p_oi_sync,
    p_oi_cxx_module,
    p_oi_ipu_p,
    p_oi_dfdef,
    p_oi_array_id_return,
    p_oi_image_id_return,
    p_oi_signal_para,
    p_oi_global_var_id,
    p_oi_ipu_no,
    p_oi_wait_array_done,
    p_oi_image_request,
    p_oi_image_return,
    p_oi_usercom,
    p_oi_trans_image,
    p_oi_sync_cache_done
} p_oi_t;

/* IPU -> OM */
typedef enum p_io {
    p_io_quit,
    p_io_ipu_create,
    p_io_module,
    p_io_fdef,
    p_io_stat,
    p_io_mod_sync,
    p_io_mod_exec,
    p_io_module_done,   /* ¹Ի: int() int(ޥ) */
    p_io_image,         /* ɽ */
    p_io_compile_stat,
    p_io_dsync,
    p_io_array_id_request,
    p_io_image_id_request,
    p_io_signal_para,
    p_io_create_global_var,
    p_io_wait_array,
    p_io_image_request,
    p_io_image_return,
    p_io_usercom,
    p_io_sync_cache,
    p_io_sync_cache_done
} p_io_t;

/*######################################################################
**
**	VPE  OM 
**
**######################################################################*/

typedef struct vpe_om {
    p_vo_t	type;
    int 	size;
} vpe_om;

/*######################################################################
**
**	OM  VPE 
**
**######################################################################*/

typedef struct om_vpe {
    p_ov_t	type;
    int 	size;
} om_vpe;

/*######################################################################
**
**	OM  IPU 
**
**######################################################################*/

typedef struct om_ipu {
    p_oi_t	type;
    int 	size;
} om_ipu;

typedef struct prog_info {	/* type  cxx, dfdef ΤȤ˻Ȥ */
    int 	in_num;
    int 	out_num;
    int 	param_num;
    char	name[64]; /* ⥸塼̾ */
} prog_info;

typedef struct sync_info {
    int 	task_id;
    int 	ip1_num;
} sync_info;

/*######################################################################
**
**	IPU  OM 
**
**######################################################################*/

typedef struct ipu_om {
    p_io_t	type;    /* åμ */
    int 	result;	 /* OKAY or FAIL */
    int 	r_stat;	 /* rstat */
    int 	size;
} ipu_om;

typedef struct out_info {
    char	out_module[64];		/* ɤ module ˤϤ */
    int 	job_id;
    int 	out_num;		/* ϲο */
} out_info;

/*######################################################################
**
**	IPU  OM 
**
**######################################################################*/

typedef struct image_info {
    int 	image_id;        /*  ID */
    int 	div_id;          /* β ID Ǥʬ ID */
    int 	width;           /*  */
    int 	height;          /* ι⤵ */
    int 	org_width;       /* ʬβ */
    int 	org_height;      /* ʬβι⤵ */
    int 	hot_x;           /* ۥåȥݥȤ x ɸ */
    int 	hot_y;           /* ۥåȥݥȤ y ɸ */
    int 	cache;           /* Ȥ礭 */
    int 	pixel_type;      /* 1ǤΥǡ (pt_int ʤ) */
    int 	image_type;      /* °륯饹 (it_gray ʤ) */

    int		div_type;        /* ʬˡ (split  done report Τ߻) */
    int 	job_id;          /* boundary Τ߻ */
    int 	image_no;        /* ⥸塼Ǥβֹ, boundary Τ */
                                 /* sync_cache Ȥ */
} image_info;

/* 桼̿ */
typedef struct usercom_info {
    int 	task_id;
    int 	image_id; /* ɤβФɸ */
    int 	from_x;
    int 	from_y;
    int 	to_x;
    int 	to_y;
} usercom_info;

/* sync cache */
typedef struct sync_cache_info {
    int 	task_id;
    int 	image_id;
} sync_cache_info;

/*######################################################################
**
**	IPU  IPU 
**
**######################################################################*/

typedef enum p_ii_t {
    p_ii_trans_quit,
    p_ii_image_request,    /*  () ž׵ */
    p_ii_image_return,     /* α */
    p_ii_array_request,    /*  () ž׵ */
    p_ii_array_return,     /* α */
    p_ii_array_read,
    p_ii_array_write,
    p_ii_array_lock,
    p_ii_array_lock_done,
    p_ii_array_unlock,
    p_ii_array_unlock_done,
    p_ii_array_unify,
    p_ii_part_image        /* ΰž */
} p_ii_t;

typedef struct trans_msg {
    p_ii_t	type;      /* åμ */
    int 	ipu_no;    /*  IPU ֹ */
    int 	size;
} trans_msg;

typedef struct para_id_info {
    int job_id;
    int para_id;
} para_id_info;

typedef struct array_id_info {
    int array_id;
    int sub_id;
    int job_id;
    int type;  /*           3ĤϾä */
    int width; /* 1Ĥ礭                   */
    int size;  /*                               */
} array_id_info;

typedef struct array_data_info {
    int index;       /* ź */
    int width;       /* 1Ĥ礭 */
    int number;      /* Ŀ */
    int need_signal; /* 񤭹ȤΤ餻ɬפ뤫 */
} array_data_info;

/*######################################################################
**
**	OM ̿
**
**######################################################################*/

typedef struct om_om {
    om_ipu*		to_ipu;
    ipu_om*		done;
    om_vpe*		to_vpe;
    vpe_om*		from_vpe;
    out_info*		o_info;
    prog_info*		p_info;
    image_info*		img_info;
    para_id_info*	para_info;
    array_id_info*	a_info;
    array_data_info*	ad_info;
    usercom_info*	uc_info;
    sync_cache_info*	sc_info;
    sync_info*		s_info;
    char*		buf;
} om_om;

/*######################################################################
**
**	IPU ̿
**
**######################################################################*/

typedef struct ipu_p_info {
    int size;
} ipu_p_info;

typedef struct ipu_ipu {
    ipu_om*			to_om;
    union {
	prog_info*		p_info;
	out_info*		o_info;
	ipu_p_info*		i_info;
	trans_msg*		t_msg;
	para_id_info*		para_info;
	usercom_info*		uc_info;
	sync_cache_info*	sc_info;
    } dummy_union_name;
    array_id_info*		a_info;
    image_info* 		img_info;
    array_data_info*		ad_info;
    char*			buf;
} ipu_ipu;

#endif /* PROTOCOL_H */
