/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  declare.c
    򰷤ץ
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "trans.h"
#include "transdefs.h"
#include "tree.h"
#include "pragma.h"

static char *current_module_name;

/***********************
* for all declarations *
***********************/

#define DECL_MAXLEVEL  32	/* ({,})κ */
#define DECL_TABLESIZE 1024	/* 1ĤΥơ֥礭 */
static struct decl_table {
    char *name;
    Tree *type;
} decl_table[DECL_MAXLEVEL][DECL_TABLESIZE];
static int decl_count[DECL_MAXLEVEL];
static int decl_level;

static char *internal_functions[] = {
    "width", "height", "waku", "hotx", "hoty",
    "sync_cache",
    NULL
};
static char *c_functions[] = {
    /* math.h */
    "acos", "asin", "atan", "atan2", "cos", "sin", "tan",
    "cosh", "sinh", "tanh",
    "exp", "frexp", "ldexp", "log", "log10", "modf",
    "pow", "sqrt",
    "ceil", "fabs", "floor", "fmod",
    /* ¾ */
    "abs", "printf", "fprintf", "fflush", "stdout", "stderr",
    /* ѿ */
    "com",
    NULL
};
static char *c_macros[] = {
    "M_E", "M_LOG2E", "M_LOG10E", "M_LN2", "M_LN10",
    "M_PI", "M_PI_2", "M_PI_4", "M_1_PI", "M_2_PI",
    "M_2_SQRTPI", "M_SQRT2", "M_SQRT1_2",
    NULL
};
static Tree internal_function_type;
static Tree c_function_type;
static Tree c_macro_type;

int
decl_levelup(void)
{
    if (decl_level >= DECL_MAXLEVEL) {
	trans_error("declaration level overflow");
	return 1;
    }
    decl_level++;
    decl_count[decl_level] = 0;
    return 0;
}

int
decl_leveldown(void)
{
    int i;

    if (decl_level <= 0) {
	trans_error("internal error: declaration level underflow");
    }
    for (i = 0; i < decl_count[decl_level]; i++) {
	free_tree(decl_table[decl_level][i].type);
	free(decl_table[decl_level][i].name);
    }
    decl_count[decl_level] = 0;
    decl_level--;
    return 0;
}

/*
 * Ͽ
 *  name  type ϥԡϿ
 */
int
regist_declaration(char *name, Tree *type)
{
    int i;
    char buf[256];

    if (decl_count[decl_level] >= DECL_TABLESIZE) {
	trans_error("declaration table overflow");
	return 1;
    }
    for (i = 0; i < decl_count[decl_level]; i++) {
	if (strcmp(name, decl_table[decl_level][i].name) == 0) {
	    sprintf(buf, "redeclaretion of `%s'", name);
	    trans_error(buf);
	    return 1;
	}
    }
    decl_table[decl_level][decl_count[decl_level]].name = strdup(name);
    decl_table[decl_level][decl_count[decl_level]].type = treedup(type);
    decl_count[decl_level]++;
    return 0;
}

Tree *
search_declarations(char *name)
{
    int level, i;

    for (level = decl_level; level >= 0; level--) {
	for (i = 0; i < decl_count[level]; i++) {
	    if (strcmp(name, decl_table[level][i].name) == 0) {
		return decl_table[level][i].type;
	    }
	}
    }
    /* Ȥ߹ߴؿ, Cοشؿ, CΥޥ */
    for (i = 0; internal_functions[i] != NULL; i++) {
	if (strcmp(name, internal_functions[i]) == 0) {
	    return &internal_function_type;
	}
    }
    for (i = 0; c_functions[i] != NULL; i++) {
	if (strcmp(name, c_functions[i]) == 0) {
	    return &c_function_type;
	}
    }
    for (i = 0; c_macros[i] != NULL; i++) {
	if (strcmp(name, c_macros[i]) == 0) {
	    return &c_macro_type;
	}
    }
    return NULL;
}

/* ۤδؿ */
int
implicit_declare(char *funcname)
{
    return regist_declaration(funcname, &c_function_type);
}

/****************
* for parameter *
****************/

/*  */
static int
check_isimage(Tree *type)
{
    if (type->ch2 != NULL) return 1; /* on ʲΤǲ */
    switch (type->node.type.primary) {
      case tp_int_image:
      case tp_float_image:
      case tp_complex_image:
	return 1; /* Ū˲ǤȤäƤΤǲ */
      default:
	return 0; /* ǤʤФǥѥ᡼ */
    }
}
static int
convert_type(int type)
{
    switch (type) {
      case tp_int_image:	return tp_int;
      case tp_float_image:	return tp_float;
      case tp_complex_image:	return tp_complex;
      default:			return type;
    }
}

static char *get_name_from_type(Tree *type);

/*
 * ⥸塼ΰơ֥Ͽ
 *   param1()ˤ줿
 *   param2ϥ⥸塼̾{δ֤ˤ줿
 */
int
regist_parameter(char *module_name, Tree *param1, Tree *param2)
    /* param1: () ΰ */
    /* param2: ⥸塼̾ { δ֤ΰ */
{
    Tree *list1, *list2;
    char buf[256];
    int  i, j, errflag = 0;
    int  in_num, out_num, param_num;
    Tree *type;
    char *name;
    int  oldtype, newtype, isimage;
    div_type divtypes[256];
    int      boxsizes[256*2];
    int      caches[256];
    int      rtype;

    in_num = out_num = param_num = 0;

    /* () ΰĴ٤ */
    for (i = j = 0; i < 3; i++) { /* i==0:input, i==1:output, i==2:param */
	list1 = (i == 0) ? param1->ch1 : ((i == 1)? param1->ch2 : param1->ch3);
	for (; list1 != NULL; list1 = list1->ch1, j++) {
	    for (list2 = param2; list2 != NULL; list2 = list2->ch2) {
		if (strcmp(list1->node.string,
			   get_name_from_type(list2->ch1)) == 0)
		    break;
	    }
	    if (list2 == NULL) {
		sprintf(buf, "parameter `%s' is not declared",
			list1->node.string);
		trans_error(buf);
		errflag = 1;
		continue;
	    }

	    /* list2 ()  () */
	    type = list2->ch1;
	    name = list1->node.string;
	    isimage = (i == 2) ? 0 : check_isimage(type);
	    oldtype = type->node.type.primary;
	    newtype = convert_type(oldtype);
	    type->node.type.primary = newtype;
	    type->node.type.isimage = isimage;
	    type->node.type.isinparam = (i == 0);
	    type->node.type.isoutparam = (i == 1);

	    if (isimage && type->ch2 == NULL) {
		/* on ʲά줿  on all Ȥ */
		type->ch2 = newnode();
		type->ch2->ch1 = newnodei(dv_all);
		type->ch2->ch2 = NULL;
	    }
	    if (isimage && type->ch2->ch2 == NULL) {
		/* cache ά줿  0 Ȥ */
		type->ch2->ch2 = newnodei(0);
	    }
	    if (!isimage && type->ch2 != NULL) {
		sprintf(buf, "`%s' is not image, but has `on ...'", name);
		trans_error(buf);
		errflag = 1;
	    }
	    if (isimage) {
		regist_declaration(name, type);
	    } else {
		/* ̾ѿ type ˹碌 */
		Tree *tmp;
		tmp = newnode();
		tmp->ch1 = type->ch1;
		type->ch1 = tmp;
		regist_declaration(name, type);
		type->ch1 = tmp->ch1;
		tmp->ch1 = NULL;

		/* ѿХåեδɽϿ */
		tmp->ch2  = newnode();
		tmp->ch2->ch1 = type;
		tmp->ch2->ch2 = NULL;
		gbuf_declare(tmp);
		tmp->ch2->ch1 = NULL;
		free_tree(tmp);
	    }

	    if (i == 0) {
		in_num++;
	    } else if (i == 1) {
		out_num++;
	    } else /* i == 2 */ {
		param_num++;
	    }

	    if (isimage) {
		div_type ptype;
		switch (type->ch2->ch1->node.ival) {
		  case dv_pixel:  ptype = dt_pixel;  break;
		  case dv_column: ptype = dt_column; break;
		  case dv_row:    ptype = dt_row;    break;
		  case dv_box:    ptype = dt_box;    break;
		  case dv_all:    ptype = dt_all;    break;
		  default:        ptype = dt_null; /* something wrong */
		}
		divtypes[j] = ptype;
		caches[j] = type->ch2->ch2->node.ival;
		if (ptype == dt_box) {
		    boxsizes[j*2]   = type->ch2->ch1->ch1->node.ival;
		    boxsizes[j*2+1] = type->ch2->ch1->ch2->node.ival;
		}
	    } else {
		divtypes[j] = dt_null;
		caches[j] = 0;
	    }
	}
    }

    /* 顼å, ٤ param2 ǥ롼פޤ魯 */
    for (list2 = param2; list2 != NULL; list2 = list2->ch2) {
	int found;

	name = get_name_from_type(list2->ch1);
	found = 0;
	for (i = 0; i < 3; i++) { /* i==0:input, i==1:output, i==2:param */
	    list1 = (i == 0) ? param1->ch1 :
		               ((i == 1)? param1->ch2 : param1->ch3);
	    for (; list1 != NULL; list1 = list1->ch1) {
		if (strcmp(list1->node.string, name) == 0) {
		    found = 1;
		    break;
		}
	    }
	    if (found) break;
	}
	if (!found) {
	    sprintf(buf, "`%s': no such argument", name);
	    trans_error(buf);
	    errflag = 1;
	}
    }

    rtype = search_pragma_divtype(module_name);
    if (rtype == PG_DIVTYPE_DEFAULT) {
	rtype = PG_DIVTYPE_BOX;
    }

    /* OM  */
    module_set_param(module_name, in_num, out_num, param_num,
		     divtypes, boxsizes, caches, rtype, 100);
    return errflag;
}

static char *
get_name_from_type(Tree *type)
{
    Tree *p;

    for (p = type->ch1; p != NULL; p = p->ch1) {
	if (p->nodetype == nt_string)
	    return p->node.string;
    }
    return NULL;
}

/* iܤϡʽϡ˲֤̾ */
char *
get_input_image_name(int i)
{
    Tree *l, *module_params, *mtype;
    char *name;
    Tree *type;
    int  found;

    if ((mtype = search_declarations(current_module_name)) == NULL)
	return NULL;
    if ((module_params = mtype->ch1) == NULL)
	return NULL;

    found = 0;
    for (l = module_params->ch1; l != NULL; l = l->ch1) {
	name = l->node.string;
	if ((type = search_declarations(name)) != NULL) {
	    if (type->node.type.isimage) {
		if (found == i) {
		    return name;
		}
		found++;
	    }
	}
    }
    return NULL;
}

char *
get_output_image_name(int i)
{
    Tree *l, *module_params, *mtype;
    char *name;
    Tree *type;
    int found;

    if ((mtype = search_declarations(current_module_name)) == NULL)
	return NULL;
    if ((module_params = mtype->ch1) == NULL)
	return NULL;

    found = 0;
    for (l = module_params->ch2; l != NULL; l = l->ch1) {
	name = l->node.string;
	if ((type = search_declarations(name)) != NULL) {
	    if (type->node.type.isimage) {
		if (found == i) {
		    return name;
		}
		found++;
	    }
	}
    }
    return NULL;
}

/** for module name **/

int
regist_module(char *name, Tree *params)
{
    Tree *mtype;
    int  ret;

    current_module_name = name;
    mtype = newnode();
    mtype->nodetype = nt_type;
    mtype->node.type.ismodule = 1;
    mtype->ch1 = params;
    ret = regist_declaration(name, mtype);
    mtype->ch1 = NULL;
    free_tree(mtype);
    return ret;
}

/*
 * declarations
 */

static int
declare(Tree *decls) /* TYPE init_decl_list SM */
{
    Tree *list, *arrs;
    Tree *type;
    char *name;

    type = treedup(decls->ch1);
    for (list = decls->ch2; list != NULL; list = list->ch2) {
	/* ꥹȤ򤿤ɤä̾õ */
	for (arrs = list->ch1->ch1; arrs->nodetype != nt_string;
	     arrs = arrs->ch1);
	name = arrs->node.string;
	type->ch1 = list->ch1;
	regist_declaration(name, type);
    }
    type->ch1 = NULL;
    free_tree(type);

    return 0;
}

/*
 * global declarations
 */

int
global_declare(Tree *decls_top) /* TYPE init_decl_list SM Υꥹ */
{
    Tree *p, *decls;
    for (p = decls_top; p != NULL; p = p->ch2) {
	decls = p->ch1; /* TYPE init_decl_list SM */
	decls->ch1->node.type.isglobal = 1;
	gbuf_declare(decls); /* åѤϿ */
	declare(decls);
    }
    return 0;
}

/*
 * local declarations
 */

int
local_declare(Tree *decls) /* TYPE init_decl_list SM */
{
    decls->ch1->node.type.isglobal = 0;
    return declare(decls);
}

/** initialize **/

int
declare_init(void)
{
    decl_level = 0;
    decl_count[0] = 0;
    return 0;
}

/** terminate **/

int
declare_term(void)
{
    int i;
    while (decl_level >= 0) {
	for (i = 0; i < decl_count[decl_level]; i++) {
	    free_tree(decl_table[decl_level][i].type);
	    free(decl_table[decl_level][i].name);
	}
	decl_count[decl_level] = 0;
	decl_level--;
    }
    decl_level = 0;
    return 0;
}
