/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  gbuf.c
    ץޥ֥ѿХåեɤ߽
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "transdefs.h"
#include "pragma.h"

#define CACHE_GLOBAL_VAR_TABLE_SIZE 1024

#define MD_CHECK 0
#define MD_READ  1
#define MD_WRITE 2

static struct {
    char *name;
    int  read_flag;
    int  write_flag;
} cache_global_var_table[CACHE_GLOBAL_VAR_TABLE_SIZE];

static int cache_global_var_table_count;

int
gbuf_init(void)
{
    cache_global_var_table_count = 0;
    return 0;
}

int
gbuf_term(void)
{
    int i;

    for (i = 0; i < cache_global_var_table_count; i++) {
	free(cache_global_var_table[i].name);
    }
    cache_global_var_table_count = 0;
    return 0;
}

int
gbuf_declare(Tree *decls) /* ХѿɽϿ */
    /* decls->ch1: , ch2: ѿꥹ */
{
    Tree *list, *arrs;

    if (cache_global_var_table_count >= CACHE_GLOBAL_VAR_TABLE_SIZE) {
	trans_error("cache global var table overflow");
	return 1;
    }

    for (list = decls->ch2; list != NULL; list = list->ch2) {
	for (arrs = list->ch1->ch1; arrs->nodetype != nt_string;
	     arrs = arrs->ch1);
	cache_global_var_table[cache_global_var_table_count].name =
	    strdup(arrs->node.string);
	cache_global_var_table_count++;
    }
    return 0;
}

static int
clear_mark(void)
{
    int i;

    for (i = 0; i < cache_global_var_table_count; i++) {
	cache_global_var_table[i].read_flag = 0;
	cache_global_var_table[i].write_flag = 0;
    }
    return 0;
}

static int
mark_read(char *varname)
{
    int i;

    for (i = 0; i < cache_global_var_table_count; i++) {
	if (strcmp(cache_global_var_table[i].name, varname) == 0) {
	    cache_global_var_table[i].read_flag = 1;
	    return 0;
	}
    }
    return 1; /* not found */
}

static int
mark_write(char *varname)
{
    int i;

    for (i = 0; i < cache_global_var_table_count; i++) {
	if (strcmp(cache_global_var_table[i].name, varname) == 0) {
	    cache_global_var_table[i].write_flag = 1;
	    return 0;
	}
    }
    return 1; /* not found */
}

static int
gen_lock_using_table(void)
{
    int  i, read_flag, write_flag, pg_mode;
    char *varname;

    for (i = 0; i < cache_global_var_table_count; i++) {
	varname    = cache_global_var_table[i].name;
	read_flag  = cache_global_var_table[i].read_flag;
	write_flag = cache_global_var_table[i].write_flag;
	pg_mode = search_pragma_mutex(varname);
	if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	    continue;
	if (read_flag && write_flag) {
	    gen_indent();
	    gen_outf("VIOS_Global_%s->lock();\n", varname);
	}
    }
    return 0;
}

static int
gen_unlock_using_table(void)
{
    int  i, read_flag, write_flag, pg_mode;
    char *varname;

    for (i = 0; i < cache_global_var_table_count; i++) {
	varname    = cache_global_var_table[i].name;
	read_flag  = cache_global_var_table[i].read_flag;
	write_flag = cache_global_var_table[i].write_flag;
	pg_mode = search_pragma_mutex(varname);
	if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	    continue;
	if (read_flag && write_flag) {
	    gen_indent();
	    gen_outf("VIOS_Global_%s->unlock();\n", varname);
	}
    }
    return 0;
}
static int
gen_lock_using_table2(void)
{
    int  i, read_flag, write_flag, pg_mode;
    char *varname;

    for (i = 0; i < cache_global_var_table_count; i++) {
	varname    = cache_global_var_table[i].name;
	read_flag  = cache_global_var_table[i].read_flag;
	write_flag = cache_global_var_table[i].write_flag;
	pg_mode = search_pragma_mutex(varname);
	if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	    continue;
	if (read_flag && write_flag) {
	    gen_outf("VIOS_Global_%s->lock(),", varname);
	}
    }
    return 0;
}

static int
gen_unlock_using_table2(void)
{
    int  i, read_flag, write_flag, pg_mode;
    char *varname;

    for (i = 0; i < cache_global_var_table_count; i++) {
	varname    = cache_global_var_table[i].name;
	read_flag  = cache_global_var_table[i].read_flag;
	write_flag = cache_global_var_table[i].write_flag;
	pg_mode = search_pragma_mutex(varname);
	if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	    continue;
	if (read_flag && write_flag) {
	    gen_outf(",VIOS_Global_%s->unlock()", varname);
	}
    }
    return 0;
}

static char *
get_global_var_name(Tree *ex)
{
    if (ex->node.op.op1 == op_id)    return ex->ch1->node.string;
    if (ex->node.op.op1 == op_array) return get_global_var_name(ex->ch1);
    return NULL;
}

static int
expand_global_array_index_sub(Tree *ex, Tree *type, int lv, int n)
    /* ex:    */
    /* type: η (礭뤿˻Ȥ) */
    /* lv:   Ƶο */
    /* n:    1礭 */
{
    if (type == NULL || type->ch2 == NULL) {
	gen_outf("0");
	return 0;
	/*
	trans_error("global array `%s' dimension error",
		    get_global_var_name(ex));
	return 1;
	*/
    }
    gen_outf("(");
    if (ex->ch1->node.op.op1 == op_array) {
	expand_global_array_index_sub(ex->ch1, type->ch1, lv+1,
				      type->ch2->ch1->node.ival);
	gen_outf(" + ");
    }
    gen_outf("(");
    if (ex->ch2 != NULL) {
	gen_exp(ex->ch2);
    } else {
	gen_outf("VIOS_foreach_i%d", lv);
    }
    gen_outf(")*%d)", n);
    return 0;
}

static int
expand_global_array_index(Tree *ex, Tree *type)
{
    if (type != NULL && type->ch1 != NULL && type->ch1->ch1 != NULL)
	expand_global_array_index_sub(ex, type->ch1->ch1, 0, 1);
    return 0;
}

static int
read_write(Tree *ex, int assign, int mode)
{
    char *varname;
    int  pg_mode;
    Tree *type;

    varname = get_global_var_name(ex);
    type = search_declarations(varname);

    pg_mode = search_pragma_mutex(varname);
    if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	return 0;

    if (mode == MD_READ && (assign == 0 || assign == 2)) {
	gen_indent();
	gen_outf("VIOS_Global_%s->read(", varname);
	if (type != NULL) expand_global_array_index(ex, type);
	gen_outf(");\n");
    }
    if (mode == MD_WRITE && (assign == 1 || assign == 2)) {
	gen_indent();
	gen_outf("VIOS_Global_%s->write(", varname);
	if (type != NULL) expand_global_array_index(ex, type);
	gen_outf(");\n");
    }
    return 0;
}
/* ʸGlobalΤλ */
static int
read_write2(Tree *ex, int assign, int mode)
{
    char *varname;
    int  pg_mode;
    Tree *type;

    varname = get_global_var_name(ex);
    type = search_declarations(varname);

    pg_mode = search_pragma_mutex(varname);
    if (pg_mode != PG_MUTEX_DEFAULT && pg_mode != PG_MUTEX_ON)
	return 0;

    if (mode == MD_READ && (assign == 0 || assign == 2)) {
	gen_outf("VIOS_Global_%s->read(", varname);
	if (type != NULL) expand_global_array_index(ex, type);
	gen_outf("),");
    }
    if (mode == MD_WRITE && (assign == 1 || assign == 2)) {
	gen_outf(",VIOS_Global_%s->write(", varname);
	if (type != NULL) expand_global_array_index(ex, type);
	gen_outf("),*%s",varname);
    }
    return 0;
}


static int
check_global_array(Tree *ex)
{
    Tree *type;

    if (ex->node.op.op1 == op_id) {
	type = search_declarations(ex->ch1->node.string);
	if (type != NULL &&
	    type->node.type.isglobal && !type->node.type.isimage)
	    return 1;
    }
    if (ex->node.op.op1 == op_array) return check_global_array(ex->ch1);
    return 0;
}

static int read_write_func(Tree *params, int mode);
static int read_write_func2(Tree *params, int mode);

static int
read_write_check(Tree *ex, int assign, int mode)
    /* assign: 0: ʤ, 1: = ˤ, 2: += ˤ */
    /* mode: MD_CHECK or MD_READ or MD_WRITE */
{
    int ret = 0;

    switch (ex->node.op.op1) {
      case op_id:
	{
	    Tree *type;
	    char *varname;

	    varname = ex->ch1->node.string;
	    if ((type = search_declarations(varname)) != NULL &&
		type->node.type.isglobal && !type->node.type.isimage)
	    {
		if (assign == 0) {
		    mark_read(varname);
		} else if (assign == 1) {
		    mark_write(varname);
		} else {
		    mark_read(varname);
		    mark_write(varname);
		}
		if (mode == MD_READ || mode == MD_WRITE) {
		    read_write(ex, assign, mode);
		}
	    }
	}
	break;
      case op_constant:
      case op_fconstant:
	break;
      case op_parent:
	ret = read_write_check(ex->ch1, assign, mode);
	break;
      case op_unop:
	ret = read_write_check(ex->ch1, 0, mode);
	break;
      case op_incop:
	if (ex->ch1 != NULL)
	    ret = read_write_check(ex->ch1, 2, mode);
	else
	    ret = read_write_check(ex->ch2, 2, mode);
	break;
      case op_cast:
	ret = read_write_check(ex->ch2, 0, mode);
	break;
      case op_mul:
      case op_div:
      case op_mod:
      case op_addop:
      case op_shift:
      case op_compop:
      case op_eqop:
      case op_and:
      case op_xor:
      case op_or:
      case op_andand:
      case op_oror:
	ret = read_write_check(ex->ch1, 0, mode);
	ret += read_write_check(ex->ch2, 0, mode);
	break;
      case op_quest:
	ret = read_write_check(ex->ch1, 0, mode);
	ret += read_write_check(ex->ch2, 0, mode);
	ret += read_write_check(ex->ch3, 0, mode);
	break;
      case op_assign:  /* = () */
	ret = read_write_check(ex->ch1, 1, mode);
	ret += read_write_check(ex->ch2, 0, mode);
	break;
      case op_assignop: /* += ʤ */
	ret = read_write_check(ex->ch1, 2, mode);
	ret += read_write_check(ex->ch2, 0, mode);
	break;
      case op_comma:
	ret  = read_write_check(ex->ch1, 0, mode);
	ret += read_write_check(ex->ch2, 0, mode);
	break;
      case op_array:
	if ((mode == MD_READ || mode == MD_WRITE) && check_global_array(ex)) {
	    read_write(ex, assign, mode);
	} else {
	    read_write_check(ex->ch1, assign, mode);
	}
	if (ex->ch2 != NULL) read_write_check(ex->ch2, 0, mode);
	break;
      case op_func:
	ret  = read_write_check(ex->ch1, 0, mode);
	ret += read_write_func(ex->ch2, mode);
	break;
      default: /* something wrong */
	break;
    }
    return ret;
}
/* ʸGlobalΤλ */
static int
read_write_check2(Tree *ex, int assign, int mode)
    /* assign: 0: ʤ, 1: = ˤ, 2: += ˤ */
    /* mode: MD_CHECK or MD_READ or MD_WRITE */
{
    int ret = 0;

    switch (ex->node.op.op1) {
      case op_id:
	{
	    Tree *type;
	    char *varname;

	    varname = ex->ch1->node.string;
	    if ((type = search_declarations(varname)) != NULL &&
		type->node.type.isglobal && !type->node.type.isimage)
	    {
		if (assign == 0) {
		    mark_read(varname);
		} else if (assign == 1) {
		    mark_write(varname);
		} else {
		    mark_read(varname);
		    mark_write(varname);
		}
		if (mode == MD_READ || mode == MD_WRITE) {
		    read_write2(ex, assign, mode);
		}
	    }
	}
	break;
      case op_constant:
      case op_fconstant:
	break;
      case op_parent:
	ret = read_write_check2(ex->ch1, assign, mode);
	break;
      case op_unop:
	ret = read_write_check2(ex->ch1, 0, mode);
	break;
      case op_incop:
	if (ex->ch1 != NULL)
	    ret = read_write_check2(ex->ch1, 2, mode);
	else
	    ret = read_write_check2(ex->ch2, 2, mode);
	break;
      case op_cast:
	ret = read_write_check2(ex->ch2, 0, mode);
	break;
      case op_mul:
      case op_div:
      case op_mod:
      case op_addop:
      case op_shift:
      case op_compop:
      case op_eqop:
      case op_and:
      case op_xor:
      case op_or:
      case op_andand:
      case op_oror:
	ret = read_write_check2(ex->ch1, 0, mode);
	ret += read_write_check2(ex->ch2, 0, mode);
	break;
      case op_quest:
	ret = read_write_check2(ex->ch1, 0, mode);
	ret += read_write_check2(ex->ch2, 0, mode);
	ret += read_write_check2(ex->ch3, 0, mode);
	break;
      case op_assign:  /* = () */
	ret = read_write_check2(ex->ch1, 1, mode);
	ret += read_write_check2(ex->ch2, 0, mode);
	break;
      case op_assignop: /* += ʤ */
	ret = read_write_check2(ex->ch1, 2, mode);
	ret += read_write_check2(ex->ch2, 0, mode);
	break;
      case op_comma:
	ret  = read_write_check2(ex->ch1, 0, mode);
	ret += read_write_check2(ex->ch2, 0, mode);
	break;
      case op_array:
	if ((mode == MD_READ || mode == MD_WRITE) && check_global_array(ex)) {
	    read_write2(ex, assign, mode);
	} else {
	    read_write_check2(ex->ch1, assign, mode);
	}
	if (ex->ch2 != NULL) read_write_check2(ex->ch2, 0, mode);
	break;
      case op_func:
	ret  = read_write_check2(ex->ch1, 0, mode);
	ret += read_write_func2(ex->ch2, mode);
	break;
      default: /* something wrong */
	break;
    }
    return ret;
}

static int
read_write_func(Tree *params, int mode)
    /* mode: MD_CHECK or MD_READ or MD_WRITE */
{
    int ret = 0;
    for (; params != NULL; params = params->ch2) {
	ret += read_write_check(params->ch1, 0, mode);
    }
    return ret;
}
/* ʸGlobalΤλ */
static int
read_write_func2(Tree *params, int mode)
    /* mode: MD_CHECK or MD_READ or MD_WRITE */
{
    int ret = 0;
    for (; params != NULL; params = params->ch2) {
	ret += read_write_check2(params->ch1, 0, mode);
    }
    return ret;
}

/* gen_exp() ƤФ롢󥿡ե */
/* ɬ gen_read, gen_write ν˸Ƥ֤  */

int
gen_read(Tree *ex)
{
    /* ơ֥ͤ򥻥å */
    clear_mark();
    read_write_check(ex, 0, MD_CHECK);

    /* readɤ */
    gen_lock_using_table();
    read_write_check(ex, 0, MD_READ);
    return 0;
}

int
gen_write(Tree *ex)
{
    /* writeɤ */
    read_write_check(ex, 0, MD_WRITE);
    gen_unlock_using_table();
    return 0;
}
/* ʸGlobalѿβΤλ */
int
gen_read2(Tree *ex)
{
    /* ơ֥ͤ򥻥å */
    clear_mark();
    read_write_check(ex, 0, MD_CHECK);

    /* readɤ */
    gen_lock_using_table2();
    read_write_check2(ex, 0, MD_READ);
    return 0;
}

int
gen_write2(Tree *ex)
{
    /* writeɤ */
    read_write_check2(ex, 0, MD_WRITE);
    gen_unlock_using_table2();
    return 0;
}

/**** ༡κǽǤΥХѿɤ߹߽ ****/

int
cache_gen_read_global_variable(void)
{
    int  i;
    char *name;
    int  mode;
    int  write_flag = 0;

    for (i = 0; i < cache_global_var_table_count; i++) {
	name = cache_global_var_table[i].name;
	mode = search_pragma_mutex(name);
	if (mode == PG_MUTEX_ON ||
	    mode == PG_MUTEX_DEFAULT ||
	    mode == PG_MUTEX_OFF) continue;
	gen_indent();
	gen_outf("VIOS_Global_%s->read();\n", name);
	write_flag = 1;
    }
    if (write_flag) {
	gen_outf("\n");
    }

    return 0;
}

/**** ༡κǸǤΥХѿν񤭽Ф ****/

int
cache_gen_write_global_variable(void)
{
    int  i;
    char *name;
    int  mode;
    int  write_flag = 0;

    for (i = 0; i < cache_global_var_table_count; i++) {
	name = cache_global_var_table[i].name;
	mode = search_pragma_mutex(name);
	if (mode == PG_MUTEX_ON ||
	    mode == PG_MUTEX_DEFAULT) continue;
	gen_indent();
	gen_outf("VIOS_Global_%s->write();\n", name);
	write_flag = 1;
    }
    if (write_flag) {
	gen_outf("\n");
    }

    return 0;
}

/**** κǸǤΥХѿ ****/

int
cache_gen_unify_global_variable(void)
{
    int  i;
    char *name;
    int  mode;
    char *s;
    int  write_flag = 0;

    gen_outf("\n");
    for (i = 0; i < cache_global_var_table_count; i++) {
	name = cache_global_var_table[i].name;
	mode = search_pragma_mutex(name);
	if (mode == PG_MUTEX_ON || mode == PG_MUTEX_DEFAULT) continue;
	switch (mode) {
	  case PG_MUTEX_OFF_MAX:
	    s = "max";
	    break;
	  case PG_MUTEX_OFF_MIN:
	    s = "min";
	    break;
	case PG_MUTEX_OFF_DMAX:
	    s = "dmax";
	    break;
	case PG_MUTEX_OFF_DMIN:
	    s = "dmin";
	    break;
	case PG_MUTEX_OFF_ADD:
	    s = "add";
	    break;
	  default:
	    trans_internal_error("unify: unknown mode");
	    s = "null";
	    break;
	}
	gen_indent();
	gen_outf("VIOS_Global_%s->unify(uf_%s);\n", name, s);
	write_flag = 1;
    }
    if (write_flag) {
	gen_outf("\n");
    }
    return 0;
}
