/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  genexp.c
    Υ
*********************************************************************/

#include <stdio.h>
#include <string.h>
#include "transdefs.h"
#include "pragma.h"

extern char *foreach_varname;

static int
isimage(Tree *ex)
{
    Tree *type;
    char *name;

    if (ex->node.ival == op_id) {
	name = ex->ch1->node.string;
	if ((type = search_declarations(name)) == NULL) {
	    return 0;
	}
	if (type->node.type.isimage) {
	    return 1;
	}
    }
    return 0;
}

static int
isarray(Tree *ex)
{
    if (ex->node.ival == op_array) {
	return 1;
    } else {
	return 0;
    }
}

static int gen_exp_sub(Tree *tr);

/* ؿ */
static int
gen_exp_funcparams(Tree *params)
{
    int ret;

    if (params == NULL) return 0;
    ret = gen_exp_sub(params->ch1);
    if (params->ch2 != NULL) gen_outf(", ");
    ret += gen_exp_funcparams(params->ch2);
    return ret;
}

static int
gen_charcter_sub(int c)
{
    switch (c) {
      case '\n':
	gen_outf("\\n");
	break;
      case '\'':
      case '\"': /* " */
	gen_outf("\\%c", c);
	break;
      default:
	gen_outf("%c", c);
	break;
    }
    return 0;
}

static int
gen_charcter(int c)
{
    gen_outf("\'");
    gen_charcter_sub(c);
    gen_outf("\'");
    return 0;
}

static int
gen_string(char *s)
{
    gen_outf("\"");
    for (; *s != '\0'; s++) {
	gen_charcter_sub(*s);
    }
    gen_outf("\"");
    return 0;
}

static int gen_exp_image(Tree *);

static int
gen_exp_sub(Tree *tr)
{
    int ret = 0;

    switch (tr->node.op.op1) {
      case op_id:
	{
	    Tree *type;

	    if ((type = search_declarations(tr->ch1->node.string)) == NULL) {
		trans_error("`%s' undeclared", tr->ch1->node.string);
	    }
	    if (type != NULL &&
		type->node.type.isglobal && !type->node.type.isimage) {
		gen_outf("(*%s)", tr->ch1->node.string);
	    } else {
		gen_outf("%s", tr->ch1->node.string);
	    }
	}
	break;
      case op_constant:
	gen_outf("%s", tr->ch2->node.string);
	break;
      case op_fconstant:
	gen_outf("%s", tr->ch2->node.string);
	break;
      case op_string:
	gen_string(tr->ch1->node.string);
	break;
      case op_charcter:
	gen_charcter(tr->ch1->node.ival);
	break;
      case op_parent:
	gen_outf("(");
	ret = gen_exp_sub(tr->ch1);
	gen_outf(")");
	break;
      case op_unop:
	gen_outf("%c", tr->node.op.op2);
	ret = gen_exp_sub(tr->ch1);
	break;
      case op_incop:
	if (tr->ch1 != NULL) {
	    gen_outf("%c%c", tr->node.op.op2, tr->node.op.op2);
	    ret = gen_exp_sub(tr->ch1);
	} else {
	    ret = gen_exp_sub(tr->ch2);
	    gen_outf("%c%c", tr->node.op.op2, tr->node.op.op2);
	}
	break;
      case op_cast:
	gen_outf("(%s)", type2string(tr->ch1->node.ival));
	ret = gen_exp_sub(tr->ch2);
	break;
      case op_mul:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("*");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_div:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("/");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_mod:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("%%");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_addop:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" %c ", tr->node.op.op2);
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_shift:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("%c%c", tr->node.op.op2, tr->node.op.op2);
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_compop:
	ret = gen_exp_sub(tr->ch1);
	switch (tr->node.op.op2) {
	  case '>': gen_outf(" > " ); break;
	  case ')': gen_outf(" >= "); break;
	  case '<': gen_outf(" < " ); break;
	  case '(': gen_outf(" <= "); break;
	}
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_eqop:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" %c= ", tr->node.op.op2);
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_and:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("&");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_xor:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("^");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_or:
	ret = gen_exp_sub(tr->ch1);
	gen_outf("|");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_andand:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" && ");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_oror:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" || ");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_quest:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" ? ");
	ret += gen_exp_sub(tr->ch2);
	gen_outf(" : ");
	ret += gen_exp_sub(tr->ch3);
	break;
      case op_assign:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(" = ");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_assignop:
	ret = gen_exp_sub(tr->ch1);
	switch (tr->node.op.op2) {
	  case '<': gen_outf(" <<= "); break;
	  case '>': gen_outf(" >>= "); break;
	  default:  gen_outf(" %c= ", tr->node.op.op2); break;
	}
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_comma:
	ret = gen_exp_sub(tr->ch1);
	gen_outf(", ");
	ret += gen_exp_sub(tr->ch2);
	break;
      case op_array:
	{
	    Tree *id_n;

	    for (id_n = tr; isarray(id_n); id_n = id_n->ch1) ;

	    if (isimage(id_n)) {
		/* Image ξ */
		ret = gen_exp_image(tr);
	    } else {
		/* Image ʳ̤ѿ */
		if (tr->node.op.op2) {
		    trans_error("[[]] is only for image");
		}
		ret = gen_exp_sub(tr->ch1);
		gen_outf("[");
		if (tr->ch2 != NULL) {
		    ret += gen_exp_sub(tr->ch2);
		} else {
		    trans_error("[] error");
		    ret = 1;
		}
		gen_outf("]");
	    }
	}
	break;
      case op_func:
	/* Ȥ߹ߴؿå */
	if (tr->ch1->node.op.op1 == op_id &&
	    tr->ch2 != NULL &&
	    tr->ch2->ch2 == NULL &&
	    tr->ch2->ch1->node.op.op1 == op_id)
	{
	    /* ؿ̾(ѿ̾) ηξ */
	    char *funcname = tr->ch1->ch1->node.string;
	    char *varname  = tr->ch2->ch1->ch1->node.string;
	    Tree *type;

	    if ((type = search_declarations(varname)) != NULL &&
			type->node.type.isimage) {
		if (strcmp(funcname, "width") == 0) {
		    gen_outf("VIOS_image_%s->get_org_width()", varname);
		    break;
		} else if (strcmp(funcname, "height") == 0) {
		    gen_outf("VIOS_image_%s->get_org_height()", varname);
		    break;
		} else if (strcmp(funcname, "waku") == 0) {
		    gen_outf("VIOS_image_%s->get_cache()", varname);
		    break;
		} else if (strcmp(funcname, "hotx") == 0) {
		    gen_outf("(VIOS_parallel_x + VIOS_image_%s->get_hot_x())",
			     varname);
		    break;
		} else if (strcmp(funcname, "hoty") == 0) {
		    gen_outf("(VIOS_parallel_y + VIOS_image_%s->get_hot_y())",
			     varname);
		    break;
		} else if (strcmp(funcname, "sync_cache") == 0) {
		    gen_outf("VIOS_sync_cache(jobid, VIOS_image_%s)", varname);
		    break;
		}
	    }
	}

	/* Ȥ߹ߴؿå */
	if (tr->ch1->node.op.op1 == op_id && tr->ch2 == NULL) {
	    /* ؿ̾(ʤ) ηξ */
	    char *funcname = tr->ch1->ch1->node.string;
	    if (strcmp(funcname, "sync_cache") == 0) {
		gen_outf("VIOS_sync_cache(jobid)");
		break;
	    }
	}

	/* Ȥ߹ߴؿǤϤʤ */
	if (tr->ch1->node.op.op1 == op_id) {
	    char *funcname = tr->ch1->ch1->node.string;

	    if (search_declarations(funcname) == NULL) {
		trans_warning("implicit declaration of function `%s'",
			      funcname);
		implicit_declare(funcname);
	    }
	}

	ret = gen_exp_sub(tr->ch1);
	gen_outf("(");
	ret += gen_exp_funcparams(tr->ch2);
	gen_outf(")");
	break;
      default: /* something wrong */
	trans_internal_error("gen_exp_sub: unknown op %d", tr->node.op.op1);
	break;
    }
    return ret;
}

int
gen_exp(Tree *ex)
{
    if (ex == NULL) return 0;
    return gen_exp_sub(ex);
}

/****  ****/

static int
gen_exp_image_coor(Tree *tr, char *imagename, int x_or_y, int div_type)
{
    if (tr->node.op.op2 == 0) {
	/* кɸ */
	if (div_type == dv_pixel || div_type == dv_box ||
	    (x_or_y == 'x' && div_type == dv_column) ||
	    (x_or_y == 'y' && div_type == dv_row))
	{
	    gen_outf("VIOS_parallel_%c + ", x_or_y);
	}
    }

    /* ɸμ */
    if (tr->ch2 != NULL) {
	gen_outf("(");
	gen_exp_sub(tr->ch2);
	gen_outf(")");
    } else {
	gen_outf("0");
    }

    if (tr->node.op.op2 != 0) {
	/* кɸ */
	gen_outf(" - VIOS_image_%s->get_hot_%c()", imagename, x_or_y);
    }
    return 0;
}

static int
gen_exp_image(Tree *tr)
{
    int  ret = 0;
    Tree *id_n;
    int  dim;
    int  i;
    char *imagename;
    int  imageno;
    Tree *type;
    int  div_type;
    int  bd_mode;
    char *const_value;

    for (dim = 0, id_n = tr; isarray(id_n); dim++, id_n = id_n->ch1) ;

    imagename = id_n->ch1->node.string;
    if (dim != 2) {
	trans_error("image `%s' must be 2 dimension", imagename);
	return 1;
    }
    type = search_declarations(imagename);
    if (type == NULL) {
	trans_internal_error("genexp: case op_array: type NULL");
	return 1;
    }
    div_type = type->ch2->ch1->node.ival;
    bd_mode = search_pragma_boundary(imagename, &const_value);

    if (bd_mode == PG_BOUNDARY_NOCHECK || bd_mode == PG_BOUNDARY_DEFAULT) {
	gen_exp_sub(tr->ch1->ch1); /* ѿ̾ */
	gen_outf("[(VIOS_image_%s_cache + ", imagename);
	gen_exp_image_coor(tr, imagename, 'y', div_type); /* y ɸ */
	gen_outf(") * VIOS_image_%s_cwidth", imagename);
	gen_outf(" + VIOS_image_%s_cache + ", imagename);
	gen_exp_image_coor(tr->ch1, imagename, 'x', div_type); /* x ɸ */
	gen_outf("]");
    } else {
	char *bd_mode_str;
	char *type_str;
	type_str =  type2string(type->node.ival);
	if (type_str == NULL) {
	    trans_internal_error("gen_exp_image: unknown type %d",
			type->node.ival);
	    type_str = "unknown";
	}
	switch (bd_mode) {
	  case PG_BOUNDARY_GET:      bd_mode_str = "bd_get"; break;
	  case PG_BOUNDARY_COMPLETE: bd_mode_str = "bd_complete"; break;
	  case PG_BOUNDARY_CONST:    bd_mode_str = "bd_const"; break;
	  default:                   bd_mode_str = "bd_null";
	    trans_internal_error("gen_exp_image: unknown boundary mode %d",
				 bd_mode);
	    break;
	}
	imageno = -1;
	for (i = 0; ; i++) {
	    char *iname = get_input_image_name(i);
	    if (iname == NULL) break;
	    if (strcmp(iname, imagename) == 0) {
		imageno = i;
		break;
	    }
	}
	if (imageno < 0) {
	    imageno = i; /* ϲο */
	    for (i = 0; ; i++) {
		char *iname = get_output_image_name(i);
		if (iname == NULL) {
		    /* ߤĤʤä */
		    trans_internal_error("gen_exp_image: image %s not found",
					 imagename);
		    imageno = -1;
		    break;
		}
		if (strcmp(iname, imagename) == 0) {
		    imageno += i;
		    break;
		}
	    }
	}
	gen_outf("VIOS_image_%s->read%s(jobid, %d, %s, ",
		 imagename, type_str, imageno, bd_mode_str);
	gen_exp_image_coor(tr->ch1, imagename, 'x', div_type); /* x ɸ */
	gen_outf(", ");
	gen_exp_image_coor(tr, imagename, 'y', div_type); /* y ɸ */
	if (bd_mode == PG_BOUNDARY_CONST) {
	    gen_outf(", (%s)%s", type_str, const_value);
	}
	gen_outf(")");
    }
    return ret;
}

/**** eval constant expression ****/

int
eval_constant_exp(Tree *ex)
{
    switch(ex->node.op.op1) {
      case op_constant:
      case op_charcter:
	return ex->ch1->node.ival;
      case op_parent:
	return eval_constant_exp(ex->ch1);
      case op_addop:
	if (ex->node.op.op2 == '+')
	    return eval_constant_exp(ex->ch1) + eval_constant_exp(ex->ch2);
	else
	    return eval_constant_exp(ex->ch1) - eval_constant_exp(ex->ch2);
      case op_mul:
	return eval_constant_exp(ex->ch1) * eval_constant_exp(ex->ch2);
      case op_div:
      case op_mod:
	{
	    int tmp1, tmp2;

	    tmp1 = eval_constant_exp(ex->ch1);
	    tmp2 = eval_constant_exp(ex->ch2);
	    if (tmp2 == 0) {
		trans_error("divided by 0");
		return 0;
	    } else {
		if (ex->node.op.op1 == op_div)
		    return tmp1 / tmp2;
		else
		    return tmp1 % tmp2;
	    }
	}
      case op_unop:
	switch (ex->node.op.op2) {
	  case '+':
	    return eval_constant_exp(ex->ch1);
	  case '-':
	    return -eval_constant_exp(ex->ch1);
	  default:
	    trans_error("unknown op");
	    return 0;
	}
      default:
	trans_error("unknown op");
	return 0;
    }
}
