/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  genmain.c
    main  
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "transdefs.h"
#include "tree.h"

#define genmain_outf		gen_outf
#define genmain_indent		gen_indent
#define genmain_indent_level_up	gen_indent_level_up
#define genmain_indent_level_down gen_indent_level_down

/* genmain_expression_sub ΰ */
typedef enum {
    em_val,   /* åͤĤ    ɤ */
    em_noval, /*               Ĥʤɤ */
    em_noout  /* ɤʤ (Ĵ٤Ȥ˻Ȥ) */
} exp_mode_t;


/* ؿΥץȥ */

static Tree *search_declare(char *);
static int  genmain_expression(Tree *);
static int  genmain_expression_sub(Tree *, exp_mode_t);
static int  genmain_expression_left(Tree *, exp_mode_t);
static void internal_func_new(Tree *);
static void internal_func_load(Tree *);
static void internal_func_save(Tree *);
static void internal_func_disp(Tree *);
static void genmain_module_call(char *mname, Tree *params);
static void module_call(char *, Tree *, Tree *);
static void module_call_sub(Tree *, Tree *);
static Tree *pre_module_call(Tree *);
static int  genmain_expression_cast(Tree *, exp_mode_t);
static int  isload(Tree *ex);
static int  genmain_load(Tree *assign_ex, exp_mode_t mode);

/* ơ֥ */
#define MAIN_DECLARE_TABLE_SIZE 1024

static struct {
    char *name;
    Tree *type;
} declare_table[MAIN_DECLARE_TABLE_SIZE];
static int declare_count;

/****  ****/
int
genmain_init(void)
{
    declare_count = 0;
    return 0;
}

/****  ****/
int
genmain_term(void)
{
    int i;

    for (i = 0; i < declare_count; i++) {
	free(declare_table[i].name);
	free_tree(declare_table[i].type);
    }
    declare_count = 0;
    return 0;
}

/****  ****/

static int regist_declare(char *, Tree *);

int
genmain_declare(Tree *decls)
{
    Tree *list, *arrs;
    Tree *type;
    char *name;

    type = treedup(decls->ch1);
    if (type->node.type.primary == tp_image) {
	type->node.type.isimage = 1;
    }
    for (list = decls->ch2; list != NULL; list = list->ch2) {
	/* ꥹȤ򤿤ɤä̾õ */
	for (arrs = list->ch1->ch1; arrs->nodetype != nt_string;
	     arrs = arrs->ch1);
	name = arrs->node.string;
	type->ch1 = list->ch1;
	regist_declare(name, type);

	if (type->ch1->ch1->nodetype == nt_string) {
	    /* Ǥʤñѿ */
	    switch (type->node.type.primary) {
	      case tp_int:
		genmain_indent();
		genmain_outf("/%s 0 def\n", name);
		break;
	      case tp_float:
		genmain_indent();
		genmain_outf("/%s 0.0 def\n", name);
		break;
	      case tp_image:
		genmain_indent();
		genmain_outf("/%s newimage def\n", name);
		break;
	      default:
		trans_error("unknown type");
		break;
	    }
	} else {
	    /*  */
	    int num;
	    Tree *tmp;

	    num = 1;
	    for (tmp = type->ch1->ch1; tmp->nodetype != nt_string;
		 tmp = tmp->ch1)
	    {
		if (tmp->ch2 == NULL || tmp->ch2->nodetype != nt_op)
		{
		    trans_error("main declare: array error");
		} else {
		    num *= eval_constant_exp(tmp->ch2);
		}
	    }
	    genmain_indent();
	    if (type->node.type.primary == tp_int) {
		genmain_outf("/%s %d intarray def\n", name, num);
	    } else if (type->node.type.primary == tp_float) {
		genmain_outf("/%s %d realarray def\n", name, num);
	    } else {
		trans_error("array type error");
	    }
	}
    }
    type->ch1 = NULL;
    free_tree(type);

    return 0;
}

static int
regist_declare(char *name, Tree *type)
{
    if (search_declare(name) != NULL) {
	trans_error("redeclaretion of `%s'", name);
	return 1;
    }
    if (declare_count >= MAIN_DECLARE_TABLE_SIZE) {
	trans_error("main declare table overflow");
	return 1;
    }
    declare_table[declare_count].name = strdup(name);
    declare_table[declare_count].type = treedup(type);
    declare_count++;
    return 0;
}

static Tree *
search_declare(char *varname)
{
    int i;

    for (i = 0; i < declare_count; i++) {
	if (strcmp(declare_table[i].name, varname) == 0) {
	    return declare_table[i].type;
	}
    }
    return NULL;
}

/****  ****/

/** Υ **/

/* ͤη顼å */
#define EXP_TYPE_NONE	1	/* ʤ       */
#define EXP_TYPE_INT	2	/*        */
#define EXP_TYPE_FLOAT	3	/* ư */
#define EXP_TYPE_IMAGE	4	/* Image      */
#define EXP_TYPE_BOOL	5	/* bool       */
#define EXP_TYPE_ARRAY	5	/*        */
#define EXP_TYPE_ERROR	(-1)	/* 顼     */

static int isarray(Tree *ex);
static int genmain_expression_array(Tree *ex, exp_mode_t);

static int
genmain_expression(Tree *ex)
{
    return genmain_expression_sub(ex, em_val);
}

/* 2黻Ҥη int:int -> int, float:int, int:float, float:float -> float */
static int
binop_typeconv(int t1, int t2)
{
    /* int or float only */
    if (t1 == EXP_TYPE_INT   && t2 == EXP_TYPE_INT)   return EXP_TYPE_INT;
    if (t1 == EXP_TYPE_INT   && t2 == EXP_TYPE_FLOAT) return EXP_TYPE_FLOAT;
    if (t1 == EXP_TYPE_FLOAT && t2 == EXP_TYPE_INT)   return EXP_TYPE_FLOAT;
    if (t1 == EXP_TYPE_FLOAT && t2 == EXP_TYPE_FLOAT) return EXP_TYPE_FLOAT;
    return EXP_TYPE_ERROR;
}

/* ͤη֤ */
static int
genmain_expression_sub(Tree *ex, exp_mode_t mode)
{
    char buf[256];
    int  pushed_flag = 1;	/* åͤѤ */
    int  tp1, tp2, rettp;	/* դηդη̤η */
    int  need_type_check = 1;   /* å򤹤뤫ɤ */
    int  binop_type_check = 0;  /* 2黻ҷηå򤹤뤫 */
    exp_mode_t mode_binop;

    tp1 = tp2 = 0; /* 0 ɬפϤʤɷٹΤ */
    rettp = EXP_TYPE_INT;	/* ̤ Image ˤʤΤϤäˤʤΤ */
				/* ǥեȤǿˤƤ */

    mode_binop = (mode == em_noout) ? em_noout : em_val;

    if (ex == NULL) return EXP_TYPE_NONE;
    switch (ex->node.op.op1) {
      case op_constant:
	if (mode != em_noout) genmain_outf("%d ", ex->ch1->node.ival);
	need_type_check = 0;
	rettp = EXP_TYPE_INT;
	break;
      case op_fconstant:
	if (mode != em_noout) genmain_outf("%s ", ex->ch2->node.string);
	need_type_check = 0;
	rettp = EXP_TYPE_FLOAT;
	break;
      case op_string:
	if (mode != em_noout)
	    genmain_outf("(%s) ", ex->ch1->node.string);
	rettp = EXP_TYPE_NONE;
	need_type_check = 0;
	break;
      case op_charcter:
	if (mode != em_noout) genmain_outf("%d ", ex->ch1->node.ival);
	need_type_check = 0;
	rettp = EXP_TYPE_INT;
	break;
      case op_id:
	{
	    char *varname;
	    Tree *type;
	    /* ѿͤλ */
	    varname = ex->ch1->node.string;
	    if (mode != em_noout) genmain_outf("%s ", varname);
	    type = search_declare(varname);
	    if (type == NULL) {
		if (mode != em_noout)
		    trans_error("`%s' not declared", varname);
		rettp = EXP_TYPE_ERROR;
	    } else {
		if (type->ch1->ch1->nodetype == nt_string) {
		    /* ñѿ */
		    switch (type->node.type.primary) {
		      case tp_image: rettp = EXP_TYPE_IMAGE; break;
		      case tp_float: rettp = EXP_TYPE_FLOAT; break;
		      case tp_int:   rettp = EXP_TYPE_INT;   break;
		      default:       rettp = EXP_TYPE_INT;   break;
		    }
		} else {
		    /* ѿ */
		    rettp = EXP_TYPE_ARRAY;
		}
	    }
	    need_type_check = 0;
	    break;
	}
      case op_parent: /* ( ) */
	rettp = genmain_expression_sub(ex->ch1, mode);
	pushed_flag = (mode == em_val);
	need_type_check = 0;
	break;
      case op_addop:
	tp1 = genmain_expression_sub(ex->ch1, mode_binop);
	tp2 = genmain_expression_sub(ex->ch2, mode_binop);
	rettp = binop_typeconv(tp1, tp2);
	binop_type_check = 1;
	if (mode != em_noout) {
	    if (ex->node.op.op2 == '+') {
		genmain_outf("add ");
	    } else { /* '-' */
		genmain_outf("sub ");
	    }
	}
	break;
      case op_mul:
	tp1 = genmain_expression_sub(ex->ch1, mode_binop);
	tp2 = genmain_expression_sub(ex->ch2, mode_binop);
	rettp = binop_typeconv(tp1, tp2);
	binop_type_check = 1;
	if (mode != em_noout) genmain_outf("mul ");
	break;
      case op_div:
	tp1 = genmain_expression_sub(ex->ch1, mode_binop);
	tp2 = genmain_expression_sub(ex->ch2, mode_binop);
	rettp = binop_typeconv(tp1, tp2);
	binop_type_check = 1;
	if (mode != em_noout) genmain_outf("div ");
	break;
      case op_mod:
	tp1 = genmain_expression_sub(ex->ch1, mode_binop);
	tp2 = genmain_expression_sub(ex->ch2, mode_binop);
	if (tp1 != EXP_TYPE_INT || tp2 != EXP_TYPE_INT) {
	    if (mode != em_noout) trans_error("mod type error");
	}
	rettp = EXP_TYPE_INT;
	if (mode != em_noout) genmain_outf("mod ");
	break;
      case op_compop:
	tp1 = genmain_expression_sub(ex->ch1, mode_binop);
	tp2 = genmain_expression_sub(ex->ch2, mode_binop);
	rettp = EXP_TYPE_BOOL;
	if (mode != em_noout) {
	    switch (ex->node.op.op2) {
	      case '<': genmain_outf("lt "); break;
	      case '(': genmain_outf("le "); break;
	      case '>': genmain_outf("gt "); break;
	      case ')': genmain_outf("ge "); break;
	      default:
		trans_internal_error("unknown op_compop (op-code2 = %d)",
				     ex->node.op.op2);
		return 1;
	    }
	}
	break;
      case op_unop:
	rettp = genmain_expression_sub(ex->ch1, mode_binop);
	if (mode != em_noout) {
	    switch (ex->node.op.op2) {
	      case '+':
		break;
	      case '-':
		genmain_outf("neg ");
		break;
	      default:
		trans_error("unary operator `%c' is not supported",
			    ex->node.op.op2);
	    }
	}
	need_type_check = 0;
	break;
      case op_comma:
	if (mode != em_noout)
	    tp1 = genmain_expression_sub(ex->ch1, em_noval);
	tp2 = genmain_expression_sub(ex->ch2, mode);
	pushed_flag = (mode == em_val);
	need_type_check = 0;
	rettp = tp2;
	break;
      case op_assign:
	if (mode == em_val) { /* ͤĤ */
	    if (isarray(ex->ch1)) {
		/* դξ */
		tp1 = genmain_expression_array(ex->ch1, mode);
		tp2 = genmain_expression_sub(ex->ch2, em_val);
		genmain_outf("dup 4 1 roll putarray ");
	    } else if (isload(ex->ch2)) {
		/* դ load() ξ */
		genmain_load(ex, mode);
	    } else {
		tp2 = genmain_expression_sub(ex->ch2, em_val);
		genmain_outf("dup ");
		tp1 = genmain_expression_left(ex->ch1, mode);
		genmain_outf("exch def ");
	    }
	    pushed_flag = 1;
	} else if (mode == em_noval) { /* ͤĤʤ */
	    if (isarray(ex->ch1)) {
		/* դξ */
		tp1 = genmain_expression_array(ex->ch1, mode);
		tp2 = genmain_expression_sub(ex->ch2, em_val);
		genmain_outf("putarray ");
	    } else if (isload(ex->ch2)) {
		/* դ load() ξ */
		genmain_load(ex, mode);
	    } else {
		tp1 = genmain_expression_left(ex->ch1, mode);
		tp2 = genmain_expression_sub(ex->ch2, em_val);
		genmain_outf("def ");
	    }
	    pushed_flag = 0;
	} else { /* Ĵ٤ */
	    if (isarray(ex->ch1)) {
		/* դξ */
		tp1 = genmain_expression_array(ex->ch1, mode);
	    } else if (isload(ex->ch2)) {
		/* դ load() ξ */
		tp1 = EXP_TYPE_IMAGE;
	    } else {
		tp1 = genmain_expression_left(ex->ch1, mode);
	    }
	    pushed_flag = 0;
	}
	rettp = tp1;
	break;
      case op_func:
	{
	    char *funcname;
	    Tree *ex_list;

	    if (ex->ch1->node.op.op1 != op_id) {
		if (mode != em_noout) 
		    trans_error("function call error. func(...) only");
		return EXP_TYPE_ERROR;
	    }
	    funcname = ex->ch1->ch1->node.string;	/* ؿ̾ */
	    ex_list  = ex->ch2;				/* ؿΰ */
	    if (strcmp(funcname, "new") == 0) {
		trans_error("new ɬפʤʤޤ");
		if (mode != em_noout) internal_func_new(ex_list);
		rettp = EXP_TYPE_IMAGE;
		pushed_flag = 1;
	    } else if (strcmp(funcname, "load") == 0) {
		if (mode != em_noout) internal_func_load(ex_list);
		rettp = EXP_TYPE_IMAGE;
		pushed_flag = 1;
	    } else if (strcmp(funcname, "save") == 0) {
		if (mode != em_noout) internal_func_save(ex_list);
		rettp = EXP_TYPE_NONE;
		pushed_flag = 0;
	    } else if (strcmp(funcname, "disp") == 0) {
		if (mode != em_noout) internal_func_disp(ex_list);
		rettp = EXP_TYPE_NONE;
		pushed_flag = 0;
	    } else {
		/* ⥸塼ƤӤ */
		if (mode != em_noout)
		    genmain_module_call(funcname, ex_list);
		rettp = EXP_TYPE_NONE;
		pushed_flag = 0;
	    }
	    need_type_check = 0;
	}
	break;
      case op_array:
	tp1 = genmain_expression_array(ex, mode);
	if (mode != em_noout) genmain_outf("getarray ");
	rettp = tp1;
	need_type_check = 0;
	break;
      case op_cast:
	tp1 = genmain_expression_cast(ex, mode);
	rettp = tp1;
	pushed_flag = (mode != em_noout);
	need_type_check = 0;
	break;
      default:
	sprintf(buf, "not supported op (op-code = %d).", ex->node.op.op1);
	trans_error(buf);
	return EXP_TYPE_ERROR;
    }

    /* mode  noval ǳĥåͤѤޤ줿 pop  */
    if (mode == em_noval && pushed_flag) {
	genmain_outf("pop ");
    }

    /* 顼å */
    if (binop_type_check) {
	/* 2黻Ҥξ硢Ǥ˥顼ɤĴ٤Ƥ */
	/* åϤޤʤΤǤǽ */
	if (rettp == EXP_TYPE_ERROR) {
	    if (mode != em_noout) trans_error("type miss match");
	}
    } else if (need_type_check) {
	/* ¾ξ */
	/* EXP_TYPE_ERROR ξϤǤ̤Υ顼ƤΤ */
	/* åϽФʤ */
	if ((tp1 != tp2 || tp1 != rettp) &&
	    tp1 != EXP_TYPE_ERROR && tp2 != EXP_TYPE_ERROR &&
	    rettp != EXP_TYPE_ERROR)
	{
	    if (mode != em_noout) trans_error("type miss match");
	}
    }

    return rettp;
}

static int
genmain_expression_left(Tree *ex, exp_mode_t mode) /* ͤ */
{
    char *varname;
    Tree *type;
    int  rettp;

    if (ex->node.op.op1 == op_id) {
	varname = ex->ch1->node.string;
	type = search_declare(varname);
	if (type == NULL) {
	    if (mode != em_noout) {
		trans_error("`%s' not declared", varname);
	    }
	    return EXP_TYPE_ERROR;
	}
	if (mode != em_noout) genmain_outf("/%s ", varname);
	switch (type->node.type.primary) {
	  case tp_image: rettp = EXP_TYPE_IMAGE; break;
	  case tp_float: rettp = EXP_TYPE_FLOAT; break;
	  default:       rettp = EXP_TYPE_INT;   break;
	}
	return rettp;
    }
    if (mode != em_noout)
	trans_error("assign requires left value");
    return EXP_TYPE_ERROR;
}

static int
isarray(Tree *ex)
{
    return ex->node.op.op1 == op_array;
}

static int genmain_array_sub(Tree *, Tree *);

static char *
genmain_get_array_name(Tree *ex)
{
    Tree *tmp;
    for (tmp = ex; tmp->node.op.op1 == op_array; tmp = tmp->ch1) ;
    if (tmp->node.op.op1 != op_id) {
	return NULL;
    }
    return tmp->ch1->node.string;
}

static int
genmain_expression_array(Tree *ex, exp_mode_t mode)
{
    char *arrayname;
    Tree *type;
    int  rettp;

    arrayname = genmain_get_array_name(ex);
    if (arrayname == NULL) {
	if (mode != em_noout) trans_error("array error");
	return EXP_TYPE_ERROR;
    }
    type = search_declare(arrayname);
    if (type == NULL) {
	if (mode != em_noout)
	    trans_error("`%s' not declared", arrayname);
	return EXP_TYPE_ERROR;
    }
    if (type->ch1 == NULL || type->ch1->ch1 == NULL) {
	if (mode != em_noout) 
	    trans_internal_error("genmain_exp_array: type NULL");
	return EXP_TYPE_ERROR;
    }
    if (type->ch1->ch1->nodetype == nt_string) {
	if (mode != em_noout) 
	    trans_error("`%s' is not array", arrayname);
	return EXP_TYPE_ERROR;
    }

    if (mode != em_noout)
	(void)genmain_array_sub(ex, type->ch1->ch1);

    switch (type->node.type.primary) {
      case tp_image: rettp = EXP_TYPE_IMAGE; break;
      case tp_float: rettp = EXP_TYPE_FLOAT; break;
      default:       rettp = EXP_TYPE_INT;   break;
    }
    return rettp;
}

/*
 int a[4][5][6];
 i = a[1][2][3];
 ((((0)*4+1)*5+2)*6+3)
*/
static int
genmain_array_sub(Tree *ex, Tree *type)
{
    int tp;
    if (type->nodetype == nt_string) {
	genmain_outf("%s ", type->node.string);
	return 0;
    } else {
	if (genmain_array_sub(ex->ch1, type->ch1) == 0) {
	    tp = genmain_expression(ex->ch2);
	} else {
	    genmain_outf("%d ", eval_constant_exp(type->ch2));
	    genmain_outf("mul ");
	    tp = genmain_expression(ex->ch2);
	    genmain_outf("add ");
	}
	if (tp != EXP_TYPE_INT) {
	    trans_error("array index type error");
	}
	return 1;
    }
}

/** load **/

/* ex  load ؿƤӽФĴ٤ */
static int
isload(Tree *ex)
{
    char *funcname;

    if (ex->node.op.op1 != op_func || ex->ch1->node.op.op1 != op_id) {
	return 0;
    }
    funcname = ex->ch1->ch1->node.string;
    if (strcmp(funcname, "load") == 0) {
	return 1;
    }
    return 0;
}

/* load  */
/* ex->ch1:  (ѿ) */
/* ex->ch2:  (load) */
static int
genmain_load(Tree *assign_ex, exp_mode_t mode)
{
    Tree *left_ex = assign_ex->ch1;
    Tree *func_ex = assign_ex->ch2;
    Tree *ex_list;
    char *funcname;

    if (func_ex == NULL || func_ex->node.op.op1 != op_func) {
	trans_internal_error("genmain_load: func_ex error");
	return -1;
    }
    if (func_ex->ch1->node.op.op1 != op_id) {
	trans_error("function call error. func(...) only");
	return -1;
    }
    funcname = func_ex->ch1->ch1->node.string;	/* ؿ̾ */
    ex_list  = func_ex->ch2;			/* ؿΰ */

    /* load("filename") η򤷤Ƥ뤫å */
    if (ex_list == NULL ||
	ex_list->ch1 == NULL ||
	ex_list->ch1->node.op.op1 != op_string ||
	ex_list->ch2 != NULL)
    {
	trans_error("load parameter error");
	return -1;
    }

    genmain_expression(left_ex);
    if (mode == em_val) {
	genmain_outf("dup ");
    }
    genmain_outf("(%s) load ", ex_list->ch1->ch1->node.string);
    return 0;
}

/** cast **/
/** Ѵ int arr[width][height] -> image */
static int
genmain_expression_cast(Tree *ex, exp_mode_t mode)
{
    char *arrayname;
    Tree *tmp;
    Tree *type;
    int  tp1, tp2;

    if (ex->ch1->node.ival != tp_image) {
	if (mode != em_noout)
	    trans_error("cast error (only array -> image)");
	return EXP_TYPE_ERROR;
    }

    for (tmp = ex->ch2; tmp->node.op.op1 == op_parent; tmp = tmp->ch1) ;
    if (tmp->node.op.op1 != op_id) {
	if (mode != em_noout) trans_error("array error (not id)");
	return EXP_TYPE_ERROR;
    }
    arrayname = tmp->ch1->node.string;

    type = search_declare(arrayname);
    if (type == NULL) {
	if (mode != em_noout)
	    trans_error("`%s' not declared", arrayname);
	return EXP_TYPE_ERROR;
    }
    if (type->ch1 == NULL || type->ch1->ch1 == NULL) {
	if (mode != em_noout) 
	    trans_internal_error("genmain_exp_cast: type NULL");
	return EXP_TYPE_ERROR;
    }

    /* 2Ĵ٤ */

    if (type->ch1->ch1->nodetype == nt_string ||
	type->ch1->ch1->ch1->nodetype == nt_string ||
	type->ch1->ch1->ch1->ch1->nodetype != nt_string)
    {
	if (mode != em_noout) 
	    trans_error("`%s' dimension error", arrayname);
	return EXP_TYPE_ERROR;
    }

    if (mode != em_noout) {
	(void)genmain_expression_sub(ex->ch2, em_val);
	genmain_outf("dup waitarray ");
	tp1 = genmain_expression_sub(type->ch1->ch1->ch1->ch2, em_val); /* x */
	tp2 = genmain_expression_sub(type->ch1->ch1->ch2, em_val); /* y */
	if (tp1 != EXP_TYPE_INT || tp2 != EXP_TYPE_INT) {
	    trans_error("array index type error");
	}
	genmain_outf("arraytoimage ");
    }
    return EXP_TYPE_IMAGE;
}

/** ⥸塼ƤӽФ **/

static void
genmain_module_call(char *mname, Tree *params)
{
    Tree *unique_number_list;

    unique_number_list = pre_module_call(params);
    module_call(mname, params, unique_number_list);
    free_tree(unique_number_list);
}

/* ⥸塼ƤˡѿΤߤΰʳΰ򤤤äѿˤ */

/* ѿ뤫ɤĴ٤ 0:ʤ 1: */
static int
need_tmp_var(Tree *ex)
{
    if (ex->node.op.op1 == op_id) {
	char *varname;
	Tree *type;
	varname = ex->ch1->node.string;
	type = search_declare(varname);
	if (type != NULL && type->node.type.primary == tp_image) {
	    /*  -> Τޤ */
	    return 0;
	}
	if (type != NULL && type->ch1 != NULL && type->ch1->ch1 != NULL &&
	    type->ch1->ch1->nodetype != nt_string)
	{
	    /*  -> Τޤ */
	    return 0;
	}

    }
    /* ʤ */
    return 0;

    /*  */
    /* return 1; */
}

static Tree *
pre_module_call(Tree *params)  /* ѿ̾Ѥ unique ʿΥꥹȤ֤ */
{
    Tree *ex;
    Tree *unique_number_list, *tmp;
    int  unique_number;
    int  tp;

    if (params == NULL) return NULL;
    unique_number_list = newnodei(0);

    ex = params->ch1; /*new*/

    /* ѿ뤫ɤĴ٤ */
    if (need_tmp_var(ex)) {
	/* ѿ ()  */
	unique_number = get_unique_number();
	unique_number_list->node.ival = unique_number;
	/* intarray  realarray  */
	tp = genmain_expression(params->ch1); /*new*/
	genmain_outf("/_tmpvar%d 1 ", unique_number);
	if (tp == EXP_TYPE_FLOAT) {
	    genmain_outf("realarray ");
	} else {
	    genmain_outf("intarray ");
	}
	genmain_outf("def _tmpvar%d exch 0 exch putarray\n", unique_number);
	genmain_indent();
    }

    tmp = pre_module_call(params->ch2); /*new*/
    unique_number_list->ch1 = tmp;

    return unique_number_list;
}

static void
module_call(char *mname, Tree *params, Tree *unique_number_list)
{
    module_call_sub(params, unique_number_list);
    genmain_outf("%s ", mname);
}

static void
module_call_sub(Tree *params, Tree *unique_number_list)
{
    Tree *ex;

    if (params == NULL) return;

    ex = params->ch1; /*new*/

    /* ѿƤ뤫ɤĴ٤ */
    if (need_tmp_var(ex) == 0) {
	/* Τޤ */
	if (ex->node.op.op1 == op_id) {
	    char *varname;
	    Tree *type;
	    varname = ex->ch1->node.string;
	    type = search_declare(varname);
	    if (type != NULL && type->node.type.primary == tp_image) {
		/*  */
		genmain_outf("%s ", ex->ch1->node.string);
	    } else {
		/*  */
		genmain_outf("%s ", ex->ch1->node.string);
	    }
	} else {
	    genmain_expression(ex);
	}
    } else {
	/* ѿäƤ */
	genmain_outf("_tmpvar%d ", unique_number_list->node.ival);
    }

    module_call_sub(params->ch2, unique_number_list->ch1); /*new*/

    return;
}

/** ؿ **/

static void
internal_func_new(Tree *ex_list)
{
    /* new("filename") η򤷤Ƥ뤫å */
    if (ex_list == NULL ||
	ex_list->ch1 == NULL ||
	ex_list->ch1->node.op.op1 != op_string ||
	ex_list->ch2 != NULL)
    {
	trans_error("new parameter error");
    } else {
	genmain_outf("(%s) obj ", ex_list->ch1->ch1->node.string);
    }
}

static void
internal_func_load(Tree *ex_list)
{
    /* load("filename") η򤷤Ƥ뤫å */
    if (ex_list == NULL ||
	ex_list->ch1 == NULL ||
	ex_list->ch1->node.op.op1 != op_string ||
	ex_list->ch2 != NULL)
    {
	trans_error("load parameter error");
    } else {
	genmain_outf("(%s) load ", ex_list->ch1->ch1->node.string);
    }
}

static void
internal_func_save(Tree *ex_list)
{
    /* save(Imagename, "filename") η򤷤Ƥ뤫å */
    if (ex_list == NULL ||			       /* ʤ */
	ex_list->ch1 == NULL ||			       /* 1 */
	ex_list->ch1->node.op.op1 != op_id ||          /* 1not̻ */
	ex_list->ch2 == NULL ||			       /* 1 */
	ex_list->ch2->ch1 == NULL ||		       /* 2 */
	ex_list->ch2->ch1->node.op.op1 != op_string || /* 2notʸ */
	ex_list->ch2->ch2 != NULL)		       /* 3İʾ */
    {
	trans_error("save parameter error");
    } else {
	char *imagename;
	char *filename;
	Tree *type;

	imagename = ex_list->ch1->ch1->node.string;
	filename  = ex_list->ch2->ch1->ch1->node.string;
	type = search_declare(imagename);
	if (type == NULL) {
	    trans_error("save type error. `%s' not declared", imagename);
	}
	if (type != NULL && type->node.type.primary != tp_image) {
	    trans_error("save type error. `%s' is not image", imagename);
	}
	genmain_outf("%s (%s) save ", imagename, filename);
    }
}

static void
internal_func_disp(Tree *ex_list)
{
    /* disp(Imagename) η򤷤Ƥ뤫å */
    if (ex_list == NULL ||
	ex_list->ch1 == NULL ||
	ex_list->ch1->node.op.op1 != op_id ||
	ex_list->ch2 != NULL)
    {
	trans_error("disp parameter error");
    } else {
	char *imagename = ex_list->ch1->ch1->node.string;
	char buf[256];
	Tree *type;

	type = search_declare(imagename);
	if (type != NULL && type->node.type.primary != tp_image) {
	    sprintf(buf, "disp type error. `%s' is not image", imagename);
	    trans_error(buf);
	}
	genmain_outf("%s disp ", imagename);
    }
}


/**** ƤӽФ ****/

void
genmain_begin(void)
{
    genmain_indent();
    genmain_outf("{\n");
    genmain_indent_level_up();
}

void
genmain_end(void)
{
    genmain_indent_level_down();
    genmain_indent();
    genmain_outf("} main\n");
}

/** ʸ **/

void
genmain_exp_stmt(Tree *ex)
{
    genmain_indent();
    (void)genmain_expression_sub(ex, em_noval);
    genmain_outf("\n");
}

/** 湽ʸ **/

void
genmain_ifhead(Tree *ex)
{
    genmain_indent();
    (void)genmain_expression(ex);
    genmain_outf("{\n");
    genmain_indent_level_up();
}

void
genmain_ifelse(void)
{
    genmain_indent_level_down();
    genmain_indent();
    genmain_outf("} {\n");
    genmain_indent_level_up();
}

void
genmain_iftail(void)
{
    genmain_indent_level_down();
    genmain_indent();
    genmain_outf("} if\n");
}

void
genmain_ifelsetail(void)
{
    genmain_indent_level_down();
    genmain_indent();
    genmain_outf("} ifelse\n");
}

void
genmain_forhead(Tree *ex1, Tree *ex2, Tree *ex3)
{
    int  start, end, inc;
    char *varname;

    if (ex1 != NULL && ex1->node.op.op1 == op_assign &&
	ex1->ch1->node.op.op1 == op_id &&
	ex1->ch2->node.op.op1 == op_constant)
    {
	start = ex1->ch2->ch1->node.ival;
	varname = ex1->ch1->ch1->node.string;
    } else {
	trans_error("for 1 ѿ =  ηǤϤޤ");
	start = 0;    /* dummy */
	varname = "";
    }

    if (ex2 != NULL && ex2->node.op.op1 == op_compop &&
	ex2->ch1->node.op.op1 == op_id &&
	ex2->ch2->node.op.op1 == op_constant)
    {
	if (strcmp(varname, ex2->ch1->ch1->node.string) != 0) {
	    trans_error("for ѿ12ǰ㤤ޤ");
	}
	end = ex2->ch2->ch1->node.ival;
	if (ex2->node.op.op2 == '<') end--;
	if (ex2->node.op.op2 == '>') end++;
    } else {
	trans_error("for 2 ѿ ӱ黻  ηǤϤޤ");
	end = 0;
    }

    if (ex3 != NULL && ex3->node.op.op1 == op_incop &&
	((ex3->ch1 != NULL && ex3->ch1->node.op.op1 == op_id) ||
	 (ex3->ch2 != NULL && ex3->ch2->node.op.op1 == op_id))
       )
    {
	inc = ex3->node.op.op2 == '+' ? 1 : -1;
	if ((ex3->ch1 != NULL &&
	     strcmp(varname, ex3->ch1->ch1->node.string) != 0) ||
	    (ex3->ch2 != NULL &&
	     strcmp(varname, ex3->ch2->ch1->node.string) != 0))
	{
	    trans_error("for ѿ13ǰ㤤ޤ");
	}
    } else if (ex3 != NULL && ex3->node.op.op1 == op_assignop &&
	       (ex3->node.op.op2 == '+' || ex3->node.op.op2 == '-') &&
	       ex3->ch1->node.op.op1 == op_id &&
	       ex3->ch2->node.op.op1 == op_constant)
    {
	if (strcmp(varname, ex3->ch1->ch1->node.string) != 0) {
	    trans_error("for ѿ13ǰ㤤ޤ");
	}
	inc = ex3->ch2->ch1->node.ival;
	if (ex3->node.op.op2 == '-') inc = -inc;
    } else {
	trans_error("for 3 ѿ++  ѿ+= ʤɤηǤϤޤ");
	inc = 1;
    }
    genmain_indent();
    genmain_outf("%d %d %d {\n", start, inc, end);
    genmain_indent_level_up();
    genmain_indent();
    genmain_outf("/%s exch def\n", varname);
    return;
}

void
genmain_fortail(void)
{
    genmain_indent_level_down();
    genmain_indent();
    genmain_outf("} for\n");
}
