/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  macro.c
    ޥ
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "transdefs.h"
#include "macro.h"
#include "tree.h"

#define MACRO_BODY_SIZE 65536
#define BUFSIZE 4096
#define MACRO_TABLE_SIZE 1024

static macro_t	macro_table[MACRO_TABLE_SIZE];
static int	macro_table_count;

/**  &  **/
int
macro_init(void)
{
    macro_table_count = 0;
    return 0;
}

int
macro_term(void)
{
    int i;
    for (i = 0; i < macro_table_count; i++) {
	free(macro_table[i].name);
	free_tree(macro_table[i].body);
    }
    macro_table_count = 0;
    return 0;
}

/** ޥơ֥Ͽ &  **/

macro_t *
search_macro(char *name)
{
    int i;

    for (i = 0; i < macro_table_count; i++) {
	if (strcmp(name, macro_table[i].name) == 0) {
	    return &macro_table[i];
	}
    }
    return NULL;
}

static int
add_macro(char *name, Tree *body, int param_n)
{
    macro_t *old;

    if ((old = search_macro(name)) != NULL) {
	trans_warning("macro `%s' redefined", name);
	free_tree(old->body);
	old->body = body;
	old->param_n = param_n;
	return 0;
    }

    if (macro_table_count >= MACRO_TABLE_SIZE) {
	trans_error("macro table overflow");
	return 1;
    }

    macro_table[macro_table_count].name    = strdup(name);
    macro_table[macro_table_count].body    = body;
    macro_table[macro_table_count].param_n = param_n;
    macro_table_count++;
    return 0;
}

static int
undef_macro(char *name)
{
    macro_t *p;

    if ((p = search_macro(name)) == NULL) return 0;
    free(p->name);
    free_tree(p->body);
    if (macro_table_count > 0) { /* if Ϥʤ */
	*p = macro_table[macro_table_count-1];
    }
    macro_table_count--;
    return 0;
}

/* misc functions */

static int (*get_p)(void),
	   (*unget_p)(int),
	   (*get_identifier_p)(char *, int),
	   (*get_string_p)(char *, int);

static int
unget(int c)
{
    return (*unget_p)(c);
}
static int
get(void)
{
    return (*get_p)();
}
static int
get_identifier(char *buf, int size)
{
    return (*get_identifier_p)(buf, size);
}
static int
get_string(char *buf, int size)
{
    return (*get_string_p)(buf, size);
}

static void
skip_space(void)
{
    int c;
    while ((c = get()), isspace(c) && c != '\n') ;
    unget(c);
    return;
}

static void
macro_parse_error(void)
{
    int c;

    trans_error("#define parse error");
    while ((c = get()) != '\n' && c != EOF) ;
    return;
}

static int
add_param(Tree *params, char *name)
{
    Tree *list;

    for (list = params; list->ch1 != NULL; list = list->ch1) ;
    list->ch1 = newnodes(strdup(name));
    return 0;
}

/* ޥΰʬɤ߹ */
static Tree *
read_params(void)
{
    int c;
    char buf[BUFSIZE];
    Tree *params;
    int  n; /* ο */

    c = get();
    if (c != '(') {
	/* ʤ */
	unget(c);
	return NULL;
    }

    n = 0;
    params = newnodei(0);
    for (;;) {
	skip_space();
	c = get();
	if (c == ')') {
	    break;
	} else if (isalpha(c) || c == '_') {
	    unget(c);
	    get_identifier(buf, BUFSIZE);
	    add_param(params, buf);
	    n++;
	    skip_space();
	    c = get();
	    if (c != ',') unget(c);
	} else {
	    macro_parse_error();
	    return params;
	}
    }
    params->node.ival = n;

    return params;
}

/* ꥹ params  name õ */
/* Ĥä餽ΰ֡(0, 1, 2, ...) Ĥʤä -1 ֤ */
static int
search_params(Tree *params, char *name)
{
    int i;
    Tree *l;

    if (params == NULL) return -1;
    for (i = 0, l = params->ch1; l != NULL; i++, l = l->ch1) {
	if (strcmp(name, l->node.string) == 0)
	    return i;
    }

    return -1;
}

/** ޥ **/
int
macro_define(int (*get_param)(void),
	     int (*unget_param)(int),
	     int (*get_identifier_param)(char *, int),
	     int (*get_string_param)(char *, int))
{
    int c;
    char buf[BUFSIZE];
    char name[BUFSIZE];
    Tree *params, *body, *body_p;
    static char body_string[MACRO_BODY_SIZE];
    int body_i;
    int n;

    get_p = get_param;
    unget_p = unget_param;
    get_identifier_p = get_identifier_param;
    get_string_p = get_string_param;

    skip_space();

    c = get();
    unget(c);
    if (!(isalpha(c) || c == '_')) {
	macro_parse_error();
	return 1;
    }
    get_identifier(name, BUFSIZE); /* ޥ̾ */

    /*  */
    params = read_params();

    skip_space();

    /* ޥ body */
    body = body_p = newnodes(NULL);
    body_i = 0;
    while ((c = get()) != '\n' && c != EOF) {
	if (isalpha(c) || c == '_') {
	    unget(c);
	    get_identifier(buf, BUFSIZE);
	    if ((n = search_params(params, buf)) >= 0) {
		/*  */
		if (body_i < MACRO_BODY_SIZE) {
		    body_string[body_i] = '\0';
		} else {
		    trans_error("macro buffer overflow");
		    body_string[MACRO_BODY_SIZE - 1] = '\0';
		}
		body_p->node.string = strdup(body_string);
		body_i = 0;
		body_p->ch1 = newnodei(n);
		body_p->ch1->ch1 = newnodes(NULL);
		body_p = body_p->ch1->ch1;
	    } else {
		/* ǤϤʤ */
		int len = strlen(buf);
		if (body_i + len < MACRO_BODY_SIZE)
		    strcpy(body_string + body_i, buf);
		body_i += len;
	    }
	} else if (c == '\"' || c == '\'') { /*"*/
	    int len;

	    unget(c);
	    get_string(buf, BUFSIZE);
	    len = strlen(buf);
	    if (body_i + len + 2 < MACRO_BODY_SIZE) {
		body_string[body_i++] = c;
		strcpy(body_string + body_i, buf);
		body_i += len;
		body_string[body_i++] = c;
	    } else {
		body_i += len + 2;
	    }
	} else {
	    if (body_i + 1 < MACRO_BODY_SIZE)
		body_string[body_i] = c;
	    body_i++;
	}
    }
    if (body_i < MACRO_BODY_SIZE) {
	body_string[body_i] = '\0';
    } else {
	trans_error("macro buffer overflow");
	body_string[MACRO_BODY_SIZE - 1] = '\0';
    }
    body_p->node.string = strdup(body_string);

    unget(c);   /* 顼åιֹΤˤ뤿 */
    add_macro(name, body, params == NULL ? -1 : params->node.ival);
    (void)get();

    free_tree(params);

    return 0;
}

/** ޥŸ **/

static char *
get_param(Tree *params, int i)
{
    int j;
    Tree *l;

    if (params != NULL) l = params->ch1; else l = NULL;
    for (j = 0; j < i; j++) {
	if (l != NULL) l = l->ch1;
    }
    if (l != NULL && l->nodetype == nt_string) {
	return l->node.string;
    } else {
	return " ";
    }
}

static Tree *
read_params2(void)
{
    int 	c;
    static char param_str[MACRO_BODY_SIZE];
    int 	param_str_i;
    Tree	*params;
    int 	par_level; /* åΥͥȤο */
    char	str[BUFSIZE];
    int 	n; /* ο */

    skip_space();
    if ((c = get()) != '(') {
	return NULL;
    }

    n = 0;
    params = newnodei(0);
    par_level = 0;
    param_str_i = 0;
    for (;;) {
	c = get();
	if (c == '(') {
	    par_level++;
	} else if (c == ')') {
	    if (par_level <= 0)
		break;
	    par_level--;
	} else if (c == ',' && par_level == 0) {
	    if (param_str_i < MACRO_BODY_SIZE)
		param_str[param_str_i] = '\0';
	    else {
		trans_error("macro replace buffer overflow");
		param_str[MACRO_BODY_SIZE - 1] = '\0';
	    }
	    add_param(params, param_str);
	    n++;
	    param_str_i = 0;
	} else if (c == '\"' || c == '\'') { /*"*/
	    int len;

	    unget(c);
	    get_string(str, BUFSIZE);
	    len = strlen(str);
	    if (param_str_i < MACRO_BODY_SIZE) {
		param_str[param_str_i++] = c;
		strcpy(param_str + param_str_i, str);
		param_str_i += len;
		param_str[param_str_i++] = c;
	    } else {
		param_str_i += len + 2;
	    }
	} else if (c == EOF) {
	    trans_error("macro not terminated");
	    break;
	} else {
	    if (param_str_i < MACRO_BODY_SIZE)
		param_str[param_str_i] = c;
	    param_str_i++;
	}
    }
    if (param_str_i < MACRO_BODY_SIZE) {
	param_str[param_str_i] = '\0';
    } else {
	trans_error("macro replace buffer overflow");
	param_str[MACRO_BODY_SIZE - 1] = '\0';
    }
    add_param(params, param_str);
    n++;
    params->node.ival = n;

    return params;
}

char *
replace_macro(macro_t *macro_p)
{
    Tree	*params;
    Tree	*body;
    static char replace_buf[MACRO_BODY_SIZE];
    int 	replace_buf_i;
    int 	len;
    char	*s;

    if (macro_p->param_n >= 0) {
	params = read_params2();
	if (params == NULL || params->node.ival != macro_p->param_n) {
	    trans_error("macro `%s' requires %d parameter%s",
			macro_p->name, macro_p->param_n,
			macro_p->param_n > 1 ? "s" : "");
	}
    } else {
	params = NULL;
    }

    replace_buf_i = 0;
    for (body = macro_p->body; body != NULL; body = body->ch1) {
	if (body->nodetype == nt_string) {
	    s = body->node.string;
	} else {
	    s = get_param(params, body->node.ival);
	}
	len = strlen(s);
	if (replace_buf_i + len < MACRO_BODY_SIZE) {
	    strcpy(replace_buf + replace_buf_i, s);
	}
	replace_buf_i += len;
    }
    if (replace_buf_i >= MACRO_BODY_SIZE) {
	trans_error("macro replace buffer overflow");
    }

    free_tree(params);

    return replace_buf;
}

/** undef **/

static void
macro_undef_parse_error(void)
{
    int c;
    trans_error("#undef parse error");
    while ((c = get()) != '\n' && c != EOF) ;
    return;
}

int
macro_undef(int (*get_param)(void),
	    int (*unget_param)(int),
	    int (*get_identifier_param)(char *, int))
{
    int c;
    char name[BUFSIZE];

    get_p = get_param;
    unget_p = unget_param;
    get_identifier_p = get_identifier_param;

    skip_space();

    c = get();
    unget(c);
    if (!(isalpha(c) || c == '_')) {
	macro_undef_parse_error();
	return 1;
    }
    get_identifier(name, BUFSIZE); /* ޥ̾ */

    undef_macro(name);

    skip_space();
    if ((c = get()) != '\n') {
	macro_undef_parse_error();
	return 1;
    }
    return 0;
}
