/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  pragma.c
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "transdefs.h"
#include "pragma.h"

#define PRAGMA_TABLE_SIZE 256
#define LINEBUFSIZE 4096

static struct pragma_mutex {
    char *varname;
    int  mode;
} pragma_mutex_table[PRAGMA_TABLE_SIZE];
static int pragma_mutex_table_count = 0;

static struct pragma_boundary {
    char *varname;
    int  mode;
    char *const_value;
} pragma_boundary_table[PRAGMA_TABLE_SIZE];
static int pragma_boundary_table_count = 0;

static struct pragma_divtype {
    char *module_name;
    int  type;
} pragma_divtype_table[PRAGMA_TABLE_SIZE];
static int pragma_divtype_table_count = 0;

/*  */
int
pragma_init(void)
{
    pragma_mutex_table_count = 0;
    pragma_boundary_table_count = 0;
    pragma_divtype_table_count = 0;
    return 0;
}

/*  */
int
pragma_term(void)
{
    int i;

    /* mutex */
    for (i = 0; i < pragma_mutex_table_count; i++) {
	free(pragma_mutex_table[i].varname);
    }
    pragma_mutex_table_count = 0;

    /* boundary */
    for (i = 0; i < pragma_boundary_table_count; i++) {
	free(pragma_boundary_table[i].varname);
	if (pragma_boundary_table[i].const_value != NULL)
	    free(pragma_boundary_table[i].const_value);
    }
    pragma_boundary_table_count = 0;

    /* divtype */
    for (i = 0; i < pragma_divtype_table_count; i++) {
	free(pragma_divtype_table[i].module_name);
    }
    pragma_divtype_table_count = 0;

    return 0;
}

/* Ͽ */
static int
regist_mutex(char *varname, int mode)
{
    int i;

    /* ǤϿƤ뤫Ĵ٤ */
    for (i = 0; i < pragma_mutex_table_count; i++) {
	if (strcmp(varname, pragma_mutex_table[i].varname) == 0) {
	    pragma_mutex_table[i].mode    = mode;
	    return 0;
	}
    }

    /* ޤϿƤʤ -> Ͽ */
    if (pragma_mutex_table_count >= PRAGMA_TABLE_SIZE) {
	trans_error("pragma: pragma table overflow");
	return 1;
    }
    pragma_mutex_table[pragma_mutex_table_count].varname = strdup(varname);
    pragma_mutex_table[pragma_mutex_table_count].mode    = mode;
    pragma_mutex_table_count++;
    return 0;
}

/* serach */
int
search_pragma_mutex(char *varname)
{
    int i;

    for (i = 0; i < pragma_mutex_table_count; i++) {
	if (strcmp(varname, pragma_mutex_table[i].varname) == 0)
	    return pragma_mutex_table[i].mode;
    }
    return PG_MUTEX_DEFAULT;
}

/* Ͽ */
static int
regist_boundary(char *varname, int mode, char *const_value)
{
    int i;

    /* ǤϿƤ뤫Ĵ٤ */
    for (i = 0; i < pragma_boundary_table_count; i++) {
	if (strcmp(varname, pragma_boundary_table[i].varname) == 0) {
	    pragma_boundary_table[i].mode = mode;
	    if (pragma_boundary_table[i].const_value != NULL)
		free(pragma_boundary_table[i].const_value);
	    pragma_boundary_table[i].const_value =
		const_value == NULL ? NULL : strdup(const_value);
	    return 0;
	}
    }

    /* ޤϿƤʤ -> Ͽ */
    if (pragma_boundary_table_count >= PRAGMA_TABLE_SIZE) {
	trans_error("#vios: boundary table overflow");
	return 1;
    }
    pragma_boundary_table[pragma_boundary_table_count].varname=strdup(varname);
    pragma_boundary_table[pragma_boundary_table_count].mode   = mode;
    pragma_boundary_table[pragma_boundary_table_count].const_value =
	const_value == NULL ? NULL : strdup(const_value);
    pragma_boundary_table_count++;

    return 0;
}

/* search */
int
search_pragma_boundary(char *varname, char **const_value_ret)
{
    int i;

    for (i = 0; i < pragma_boundary_table_count; i++) {
	if (strcmp(varname, pragma_boundary_table[i].varname) == 0) {
	    *const_value_ret = pragma_boundary_table[i].const_value;
	    return pragma_boundary_table[i].mode;
	}
    }
    return PG_BOUNDARY_DEFAULT;
}

/* Ͽ */
static int
regist_divtype(char *module_name, int type)
{
    int i;

    /* ǤϿƤ뤫Ĵ٤ */
    for (i = 0; i < pragma_divtype_table_count; i++) {
	if (strcmp(module_name, pragma_divtype_table[i].module_name) == 0) {
	    pragma_divtype_table[i].type = type;
	    return 0;
	}
    }

    /* ޤϿƤʤ -> Ͽ */
    if (pragma_divtype_table_count >= PRAGMA_TABLE_SIZE) {
	trans_error("#vios: divtype table overflow");
	return 1;
    }

    i = pragma_divtype_table_count;
    pragma_divtype_table[i].module_name = strdup(module_name);
    pragma_divtype_table[i].type        = type;
    pragma_divtype_table_count++;

    return 0;
}

/* search */
int
search_pragma_divtype(char *module_name)
{
    int i;

    for (i = 0; i < pragma_divtype_table_count; i++) {
	if (strcmp(module_name, pragma_divtype_table[i].module_name) == 0) {
	    return pragma_divtype_table[i].type;
	}
    }
    return PG_DIVTYPE_DEFAULT;
}

/** pragma  **/

static int /* s  d 1ñ쥳ԡ. ͤϥԡʸ */
get_word(char *s, char *d, int size)
{
    int n = 0;

    while (isalpha((int)*s) || *s == '_' || isdigit((int)*s) ||
	   *s == '.' || *s == '+' || *s == '-')
    {
	if (n < size - 1) {
	    *d++ = *s++;
	    n++;
	} else {
	    trans_error("pragma: line buffer overflow");
	    return -1;
	}
    }
    *d = '\0';

    return n;
}

static int
check_left(char *p)
    /* #vios ... Ϥäȡ;ʬʸʤĴ٤ */
{
    while (isspace((int)*p) && *p != '\n' && *p != '\0') p++;
    if (*p != '\n' && *p != '\0') {
	trans_error("parse error: too many strings after #vios ...");
	return 1;
    }
    return 0;
}

/** mutex **/

static int
parse_mutex_on(char *p, char *varname)
{
    int ret;

    ret = regist_mutex(varname, PG_MUTEX_ON);
    while (*p == ' ' || *p == '\t') p++;
    ret += check_left(p);
    return ret;
}

static int
parse_mutex_off(char *s, char *varname)
{
    char *p = s;
    char buf[LINEBUFSIZE];
    int  n;
    int  mode;
    int  ret;

    while (*p == ' ' || *p == '\t') p++;
    if (isalpha((int)*p) || *p == '_') {
	if ((n = get_word(p, buf, LINEBUFSIZE)) < 0) return 1;
	p += n;
	if (strcmp("max", buf) == 0) {
	    while (*p == ' ' || *p == '\t') p++;
	    if((n = get_word(p, buf, LINEBUFSIZE)) == 0){
		mode = PG_MUTEX_OFF_MAX;
	    }else if(n>0 && (strcmp("depend_on_first",buf) == 0)){
		p += n;
		mode = PG_MUTEX_OFF_DMAX;
	    }else {
		trans_error(
		    "#vios: mutex off error about `max'");
		return 1;	
	    }
	} else if (strcmp("min", buf) == 0) {
	    while (*p == ' ' || *p == '\t') p++;
	    if((n = get_word(p, buf, LINEBUFSIZE)) == 0){
		mode = PG_MUTEX_OFF_MIN;
	    }else if(n>0 && (strcmp("depend_on_first",buf) == 0)){
		p += n;
		mode = PG_MUTEX_OFF_DMIN;
	    }else {
		trans_error(
		    "#vios: mutex off error about `min'");
		return 1;	
	    }
	} else if (strcmp("add", buf) == 0) {
	    mode = PG_MUTEX_OFF_ADD;
	} else {
	    trans_error(
	     "#vios: mutex off requires `max' or `min' or `add' or nothing");
	    return 1;
	}
	ret = regist_mutex(varname, mode);
	while (*p == ' ' || *p == '\t') p++;
	ret += check_left(p);
    } else {
	if (*p == '\n' || *p == '\0') {
	    ret = regist_mutex(varname, PG_MUTEX_OFF);
	} else {
	    trans_error(
	     "#vios: mutex off requires `max' or `min' or `add' or nothing");
	    return 1;
	}
    }
    return ret;
}

static int
parse_mutex(char *s)
{
    char *p = s;
    char buf[LINEBUFSIZE];
    char varname[LINEBUFSIZE];
    int  n;

    while (*p == ' ' || *p == '\t') p++;
    if (isalpha((int)*p) || *p == '_') {
	/* ѿ̾ */
	if ((n = get_word(p, varname, LINEBUFSIZE)) < 0) return 1;
	p += n;
	while (*p == ' ' || *p == '\t') p++;

	/* "on" or "off" */
	if (isalpha((int)*p) || *p == '_') {
	    if ((n = get_word(p, buf, LINEBUFSIZE)) < 0) {
		return 1;
	    }
	    p += n;
	    if (strcmp("on", buf) == 0) {
		return parse_mutex_on(p, varname);
	    } else if (strcmp("off", buf) == 0) {
		return parse_mutex_off(p, varname);
	    }
	}
	trans_error("#vios: mutex requires `on' or `off' after var name");
	return 1;
    } else {
	trans_error("#vios: mutex requires var name");
	return 1;
    }
}

/** boundary **/

static int
check_number(char *s) /* s ɤ check */
{
    if (*s == '\0') return 1;
    return 0;
}

static int
parse_boundary_const(char *s, char *varname)
{
    char *p = s;
    char const_value[LINEBUFSIZE];
    int  n;
    int  ret;

    while (*p == ' ' || *p == '\t') p++;
    n = get_word(p, const_value, LINEBUFSIZE);
    p += n;
    if (check_number(const_value) != 0) {
	trans_error("#vios: const requires number");
	return 1;
    }

    ret = regist_boundary(varname, PG_BOUNDARY_CONST, const_value);
    ret += check_left(p);

    return ret;
}

static int
parse_boundary_complete(char *s, char *varname)
{
    char *p = s;
    char axis[LINEBUFSIZE];
    int  n;
    int  mode;
    int  ret;

    while (*p == ' ' || *p == '\t') p++;
    n = get_word(p, axis, LINEBUFSIZE);
    p += n;
    if (strcmp(axis, "") == 0) {
	mode = PG_BOUNDARY_COMPLETE;
    } else if (strcmp(axis, "x") == 0) {
	mode = PG_BOUNDARY_COMPLETE;
    } else if (strcmp(axis, "y") == 0) {
	mode = PG_BOUNDARY_COMPLETE;
    } else {
	trans_error("#vios: complete require `x' or `y' or nothing");
	return 1;
    }

    ret = regist_boundary(varname, mode, NULL);
    ret += check_left(p);

    return ret;
}

static int
parse_boundary(char *s)
{
    char *p = s;
    char buf[LINEBUFSIZE];
    char varname[LINEBUFSIZE];
    int  n;

    while (*p == ' ' || *p == '\t') p++;
    if (isalpha((int)*p) || *p == '_') {
	/* ѿ̾ */
	if ((n = get_word(p, varname, LINEBUFSIZE)) < 0) return 1;
	p += n;
	while (*p == ' ' || *p == '\t') p++;

	/* "nocheck" or "get" or "complete" or "const" */
	if (isalpha((int)*p) || *p == '_') {
	    if ((n = get_word(p, buf, LINEBUFSIZE)) < 0) {
		return 1;
	    }
	    p += n;
	    if (strcmp("nocheck", buf) == 0) {
		return regist_boundary(varname, PG_BOUNDARY_NOCHECK, NULL);
	    } else if (strcmp("get", buf) == 0) {
		return regist_boundary(varname, PG_BOUNDARY_GET, NULL);
	    } else if (strcmp("complete", buf) == 0) {
		return parse_boundary_complete(p, varname);
	    } else if (strcmp("const", buf) == 0) {
		return parse_boundary_const(p, varname);
	    } else {
		trans_error("#vios: boundary parse error");
		return 1;
	    }
	}
	trans_error("#vios boundary require image name");
	return 1;
    } else {
	trans_error("#vios boundary require image name");
	return 1;
    }
}

/** divtype **/

static int
parse_divtype(char *s)
{
    char *p = s;
    char buf[LINEBUFSIZE];
    char module_name[LINEBUFSIZE];
    int  n;

    while (*p == ' ' || *p == '\t') p++;
    if (isalpha((int)*p) || *p == '_') {
	/* ⥸塼̾ */
	if ((n = get_word(p, module_name, LINEBUFSIZE)) < 0) return 1;
	p += n;
	while (*p == ' ' || *p == '\t') p++;

	/* "box" or "row" or "column" */
	if (isalpha((int)*p) || *p == '_') {
	    if ((n = get_word(p, buf, LINEBUFSIZE)) < 0) {
		return 1;
	    }
	    p += n;
	    if (strcmp("box", buf) == 0) {
		return regist_divtype(module_name, PG_DIVTYPE_BOX);
	    } else if (strcmp("row", buf) == 0) {
		return regist_divtype(module_name, PG_DIVTYPE_ROW);
	    } else if (strcmp("column", buf) == 0) {
		return regist_divtype(module_name, PG_DIVTYPE_COLUMN);
	    } else {
		trans_error("#vios: divtype parse error");
		return 1;
	    }
	}
	trans_error("#vios divtype require module name");
	return 1;
    } else {
	trans_error("#vios divtype require module name");
	return 1;
    }
}

/** interface for lexical part **/

int
parse_pragma(char *s)
{
    char *p = s;
    char buf[LINEBUFSIZE];
    int  n;

    while (*p == ' ' || *p == '\t') p++;
    if (isalpha((int)*p) || *p == '_') {
	if ((n = get_word(p, buf, LINEBUFSIZE)) < 0) return 1;
	p += n;

	if (strcmp("mutex", buf) == 0) {
	    return parse_mutex(p);
	} else if (strcmp("boundary", buf) == 0) {
	    return parse_boundary(p);
	} else if (strcmp("divtype", buf) == 0) {
	    return parse_divtype(p);
	}
    }
    sprintf(buf, "unknown #vios `%s'", s);
    trans_warning(buf);
    return 0;
}
