/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  transdefs.h
*********************************************************************/

#ifndef TRANSDEFS_H
#define TRANSDEFS_H

#define TRANS_FILENAME_BUFSIZE 256

#include <stdio.h>  /* for FILE */
#include "tree.h"

/* declare */

typedef enum type_sp {	/* type-specifier */
    tp_none, tp_int, tp_float, tp_complex, tp_image,
    tp_int_image, tp_float_image, tp_complex_image, tp_char, tp_void
} type_sp_t;

extern char	trans_inputfile[];
extern int	trans_lineno;
extern int	trans_nerrors, trans_nwarnings;

/* VPE-Pη */
#define TP_PARAM	1	/* ⥸塼Ǥʤ   */
#define TP_IMAGE	2	/* ֥ȤǤʤ */
#define TP_INPUT	4	/* ϤϤ                   */

/* ʬñ */
#define dv_pixel	1
#define dv_row		2
#define dv_column	3
#define dv_box		4
#define dv_all		5

/* 黻 */
typedef enum op_code {
    op_func,		/* () (function call) */
    op_array,		/* []  */
    op_arrow,		/* ->  */
    op_dot,		/* .   */
    op_unop,		/* + - ! ~ * & */
    op_incop,		/* ++ -- */
    op_cast,		/* (type) */
    op_sizeof,		/* sizeof */
    op_mul,		/* *   */
    op_div,		/* /   */
    op_mod,		/* %   */
    op_addop,		/* + - */
    op_shift,		/* << >> */
    op_compop,		/* < <= > >=  */
    op_eqop,		/* == !=  */
    op_and,		/* &   */
    op_xor,		/* ^   */
    op_or,		/* |   */
    op_andand,		/* &&  */
    op_oror,		/* ||  */
    op_quest,		/* ? : */
    op_assign,		/* =   */
    op_assignop,	/* += -= ... */
    op_comma,		/* ,   */
    op_constant,	/* int  */
    op_fconstant,	/* float  */
    op_id,		/* ̻     */
    op_parent,		/* ( expression ) */
    op_string,		/* "..." */
    op_charcter		/* 'c'   */
} op_code_t;

/**** function prototypes ****/

/* transparse.y */
extern int	translate(char *);
extern void	trans_error(char *, ...);
extern void	trans_internal_error(char *, ...);
extern void	trans_warning(char *, ...);

/* translex.c */
extern int	translex_init(char *);
extern int	yylex(void);

/* declare.c */
extern int	declare_init(void);
extern int	declare_term(void);
extern int	regist_parameter(char *, Tree *, Tree *);
extern int	regist_module(char *, Tree *);
extern int	decl_levelup(void);
extern int	decl_leveldown(void);
extern int	regist_declaration(char *, Tree *);
extern Tree *	search_declarations(char *);
extern int	global_declare(Tree *);
extern int	local_declare(Tree *);
extern char *	get_input_image_name(int);
extern char *	get_output_image_name(int);
extern int	implicit_declare(char *);

/* gencode.c */
extern int	gen_init(void);
extern void	gen_programhead(void);
extern void	gen_modulehead(char *);
extern void	gen_moduleparam(Tree *);
extern void	gen_moduletail(char *);
extern int	gen_global_decl1(Tree *);
extern void	gen_global_decl2(Tree *);
extern void	gen_global_decl3(Tree *);
extern int	gen_module(char *mod_name, Tree *param, Tree *tr);
extern void	gen_begin_parallel(void);
extern void	gen_end_parallel(void);
extern void	gen_begin_sequential(void);
extern void	gen_end_sequential(void);
extern void	gen_local_declare(Tree *);
extern void	gen_LC(void);
extern void	gen_RC(void);
extern int	gen_exp(Tree *);
extern int	gen_foreachhead(char *);
extern int	gen_foreachtail(void);
extern int	get_unique_number(void);
extern int      gen_c_function(int type, char *func_name, Tree *arg, Tree *comp_st);
extern char	*type2string(int type);

/* genexp.c */
extern int	eval_constant_exp(Tree *);

/* genout.c */
extern int	gen_outf(char *form, ...);
extern void	gen_indent(void);
extern void	gen_indent_level_up(void);   /* 2Ĥgenmain.c */
extern void	gen_indent_level_down(void); /* Τ߸ƤФ   */
extern void	gen_LC(void);	/* ʳ   */
extern void	gen_RC(void);	/* äȤ */

/* genmain.c */
extern int	genmain_init(void);
extern int	genmain_term(void);
extern void	genmain_begin(void);
extern void	genmain_end(void);
extern int	genmain_declare(Tree *);
extern void	genmain_exp_stmt(Tree *);
extern void	genmain_ifhead(Tree *);
extern void	genmain_iftail(void);
extern void	genmain_ifelse(void);
extern void	genmain_ifelsetail(void);
extern void	genmain_forhead(Tree *, Tree *, Tree *);
extern void	genmain_fortail(void);

/* gbuf.c */
extern int	gbuf_init(void);
extern int	gbuf_term(void);
extern int	gen_read(Tree *);
extern int	gen_write(Tree *);
extern int	gbuf_declare(Tree *);
extern int	cache_gen_read_global_variable(void);
extern int	cache_gen_unify_global_variable(void);
extern int	cache_gen_write_global_variable(void);

/* 꽤ʬ */
extern int	gen_read2(Tree *);
extern int	gen_write2(Tree *);

#endif /* TRANSDEFS_H */
