/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  translex.c
    
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "transdefs.h"
#include "tree.h"
#include "macro.h"
#include "pragma.h"
#include "y.tab.h"

#define BUFSIZE 4096

static int check_keyword(char *);

char trans_inputfile[TRANS_FILENAME_BUFSIZE];
int  trans_lineno;

static int  unget_buf[256]; /* <- int ... for EOF */
static char *prg_ptr;
static int  *unget_ptr;

/** ν **/
int
translex_init(char *prg_buf)
{
    prg_ptr = prg_buf;
    unget_ptr = unget_buf;
    strcpy(trans_inputfile, "(no_name)");
    trans_lineno = 1;
    return 0;
}

/** Хåե1ʸɤ߹ ('\' + ԡ, Ƚդ) **/
static int
unget(int c)
{
    if (c == '\n') trans_lineno--;
    *unget_ptr++ = c;
    return 0;
}

static int
ungets(char *s)
{
    char *p;

    p = s + strlen(s);
    while (p > s)
	unget(*--p);
    return 0;
}

static int
get0(void)
{
    int c;

    if (unget_ptr > unget_buf)
	c = *--unget_ptr;
    else {
	c = *prg_ptr++;
	if (c == '\0') {
	    c = EOF;
	    prg_ptr--;
	}
    }
    if (c == '\n') trans_lineno++;
    return c;
}

static int
get(void)
{
    int c, c2;

    if ((c = get0()) == '\\') {
	if ((c2 = get0()) == '\n') {
	    return get();
	}
	unget(c2);
	return '\\';
    } else if (c == '/') {
	if ((c2 = get0()) == '*') {
	    /* ɤФ */
	    for (;;) {
		while ((c = get0()) != '*' && c != EOF) ;
		if (c == EOF) {
		    trans_error("comment not terminated\n");
		    break;
		}
		if ((c = get()) == '/') {
		    break;
		}
		unget(c);
	    }
	    return ' ';
	} else if (c2 == '/') { /* c++ Υ */
	    while ((c = get0()) != '\n' && c != EOF) ;
	    if (c == '\n') unget(c);
	    return ' ';
	} else {
	    unget(c2);
	    return '/';
	}
    }
    return c;
}

/** get() ѤƼ̻(ѻǻϤޤѿ) buf ɤ߹       **/
/** ǽ餬ѻǤ뤫ϥåƤʤ **/
static int
get_identifier(char *buf, int size)
{
    int i;
    int c;

    for (i = 0, c = get(); isalnum(c) || c == '_'; i++, c = get()) {
	if (i < size)
	    buf[i] = c;
    }
    unget(c);
    if (i >= size) {
	trans_error("translex: buffer overflow");
	buf[size - 1] = '\0';
	return 1;
    } else {
	buf[i] = '\0';
	return 0;
    }
}

/** get() Ѥʸ(ޤʸ) buf ɤ߹ **/
static int h2i(int ch);
static int
get_string(char *buf, int bufsize)
{
    int n;
    int c;
    int quote; /* ' or " */    /* '" */

    quote = get(); /* 1ʸ quote Ȥ */
    n = 0;
    for (;;) {
	c = get();
	if (c == quote)
	    break;
	if (c == '\\') {
	    c = get();
	    switch (c) {
	      case 'a':  c = '\a'; break;
	      case 'b':  c = '\b'; break;
	      case 'f':  c = '\f'; break;
	      case 'n':  c = '\n'; break;
	      case 'r':  c = '\r'; break;
	      case 't':  c = '\t'; break;
	      case 'v':  c = '\v'; break;
	      case '\\': c = '\\'; break;
	      case '?':  c = '?';  break;
	      case '\'': c = '\''; break;
	      case '\"': c = '\"'; break; /*"*/
	      case 'x':
		{
		    int ch;
		    c = 0;
		    for (ch = get(); isxdigit(ch); ch = get()) {
			c = c*16 + h2i(ch);
		    }
		    unget(ch);
		}
		break;
	      default:
		if (c >= '0' && c <= '7') {
		    int i, ch;
		    c -= '0';
		    for (i = 1; i < 3; i++) {
			ch = get();
			if (ch >= '0' && ch <= '7') {
			    c = c*8 + (ch - '0');
			} else {
			    unget(ch);
			    break;
			}
		    }
		} else {
		    char errmsg[256];
		    sprintf(errmsg, "unknown escape sequence '\\%c'", c);
		    trans_warning(errmsg);
		}
	    }
	}
	if (c == EOF) {
	    trans_error("EOF in string");
	    break;
	}
	if (n < bufsize)
	    buf[n++] = c;
    }
    if (n >= bufsize) {
	trans_error("translex: string buffer overflow\n");
	buf[bufsize - 1] = '\0';
	return 1;
    } else {
	buf[n] = '\0';
	return 0;
    }
}
static int
h2i(int ch)
{
    if (isdigit(ch)) {
	return ch - '0';
    } else {
	switch (ch) {
	  case 'a': case 'A': return 10;
	  case 'b': case 'B': return 11;
	  case 'c': case 'C': return 12;
	  case 'd': case 'D': return 13;
	  case 'e': case 'E': return 14;
	  case 'f': case 'F': return 15;
	}
    }
    return 0;
}

static int
get_number(char *buf, int size)
{
    int i, c;

    for (i = 0, c = get(); isdigit(c); i++, c = get()) {
	if (i < size)
	    buf[i] = c;
    }
    if (c != '.' && c != 'e' && c != 'E') {
	/*  */
	if ((c == 'x' || c == 'X') && buf[0] == '0' && i == 1) {
	    /* 16ʿ */
	    do {
		if (i < size)
		    buf[i] = c;
		i++;
	    } while (c = get(), isdigit(c));
	}
	/*  */
	if (c == 'L' || c == 'l') {
	    if (i < size) buf[i] = c;
	    i++;
	    c = get();
	    if (c == 'U' || c == 'u') {
		if (i < size) buf[i] = c;
		i++;
	    } else {
		unget(c);
	    }
	} else if (c == 'U' || c == 'u') {
	    if (i < size) buf[i] = c;
	    i++;
	    c = get();
	    if (c == 'L' || c == 'l') {
		if (i < size) buf[i] = c;
		i++;
	    } else {
		unget(c);
	    }
	} else {
	    unget(c);
	}

	if (i >= size) {
	    trans_error("translex: number string buffer overflow");
	    buf[size - 1] = '\0';
	} else {
	    buf[i] = '\0';
	}
	return INTEGER;
    }

    /* ư */
    if (c == '.') {
	if (i < size) buf[i] = c;
	i++;
	for (c = get(); isdigit(c); i++, c = get()) {
	    if (i < size)
		buf[i] = c;
	}
    }
    if (buf[0] == '.' && i == 1) {
	return DOT;	/* '.' */
    }
    if (c == 'e' || c == 'E') {
	/* ؿ */
	if (i < size) buf[i++] = c;
	c = get();
	if (c == '+' || c == '-') {
	    if (i < size) buf[i++] = c;
	    c = get();
	}
	while (isdigit(c)) {
	    if (i < size) buf[i++] = c;
	    c = get();
	}
    }
    if (c == 'F' || c == 'f' || c == 'L' || c == 'l') {
	if (i < size) buf[i] = c;
	i++;
    } else {
	unget(c);
    }
    if (i >= size) {
	trans_error("internal error: buffer overflow");
	buf[size - 1] = '\0';
    } else {
	buf[i] = '\0';
    }
    return FLOATING;
}

/** yylex  **/
int
yylex(void)
{
    int  c, c2, c3;
    int  i;
    char buf[BUFSIZE];

  try_again:
    /* ɤФ */
    while (c = get(), isspace(c)) ;

    if (c == EOF) return 0;
    if (isalpha(c) || c == '_') {		/* ͽ or ̻ */
	int     token;
	macro_t *macro_p;
	char    *replaced_string;

	unget(c);
	get_identifier(buf, BUFSIZE);
	if ((macro_p = search_macro(buf)) != NULL) {
	    replaced_string = replace_macro(macro_p);
	    ungets(replaced_string);
	    goto try_again;
	}
	if ((token = check_keyword(buf)) >= 0) {
	    return token;			/* ͽ */
	} else {
	    yylval.string = strdup(buf);	/* ̻ */
	    return IDENTIFIER;
	}
    } else if (isdigit(c) || c == '.') {	/*  or ¿ */
	int token;

	unget(c);
	token = get_number(buf, BUFSIZE);
	if (token == INTEGER || token == FLOATING) {
	    yylval.string = strdup(buf);
	    return token;
	} else {
	    return token;
	}
    } else if (c == ',') {
	return CM;
    } else if (c == '=') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '=';
	    return EQOP;
	} else {
	    unget(c2);
	    return ASSIGN;
	}
    } else if (c == '+') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '+';
	    return ASSIGNOP;
	} else if (c2 == '+') {
	    yylval.ival = '+';
	    return INCOP;
	} else {
	    unget(c2);
	    yylval.ival = '+';
	    return ADDOP;
	}
    } else if (c == '-') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '-';
	    return ASSIGNOP;
	} else if (c2 == '-') {
	    yylval.ival = '-';
	    return INCOP;
	} else if (c2 == '>') {
	    return ARROW;
	} else {
	    unget(c2);
	    yylval.ival = '-';
	    return ADDOP;
	}
    } else if (c == '*') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '*';
	    return ASSIGNOP;
	} else {
	    unget(c2);
	    return MULOP;
	}
    } else if (c == '/') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '/';
	    return ASSIGNOP;
	} else {
	    unget(c2);
	    return DIVOP;
	}
    } else if (c == '%') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '%';
	    return ASSIGNOP;
	} else {
	    unget(c2);
	    return MODOP;
	}
    } else if (c == '<') {
	if ((c2 = get()) == '<') {
	    if ((c3 = get()) == '=') {
		yylval.ival = '<';
		return ASSIGNOP;
	    } else {
		unget(c3);
		yylval.ival = '<';
		return SHIFTOP;
	    }
	} else if (c2 == '=') {
	    yylval.ival = '(';
	    return COMPOP;
	} else {
	    unget(c2);
	    yylval.ival = '<';
	    return COMPOP;
	}
    } else if (c == '>') {
	if ((c2 = get()) == '>') {
	    if ((c3 = get()) == '=') {
		yylval.ival = '>';
		return ASSIGNOP;
	    } else {
		unget(c3);
		yylval.ival = '>';
		return SHIFTOP;
	    }
	} else if (c2 == '=') {
	    yylval.ival = ')';
	    return COMPOP;
	} else {
	    unget(c2);
	    yylval.ival = '>';
	    return COMPOP;
	}
    } else if (c == '&') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '&';
	    return ASSIGNOP;
	} else if (c2 == '&') {
	    return ANDAND;
	} else {
	    unget(c2);
	    return ANDOP;
	}
    } else if (c == '|') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '|';
	    return ASSIGNOP;
	} else if (c2 == '|') {
	    return OROR;
	} else {
	    unget(c2);
	    return OROP;
	}
    } else if (c == '^') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '^';
	    return ASSIGNOP;
	} else {
	    unget(c2);
	    return XOROP;
	}
    } else if (c == '!') {
	if ((c2 = get()) == '=') {
	    yylval.ival = '!';
	    return EQOP;
	} else {
	    unget(c2);
	    yylval.ival = '!';
	    return UNOP;
	}
    } else if (c == '~') {
	yylval.ival = '~';
	return UNOP;
    } else if (c == '(') {
	return LP;
    } else if (c == ')') {
	return RP;
    } else if (c == '{') {
	return LC;
    } else if (c == '}') {
	return RC;
    } else if (c == '[') {
	return LB;
    } else if (c == ']') {
	return RB;
    } else if (c == ';') {
	return SM;
    } else if (c == ':') {
	return COLON;
    } else if (c == '?') {
	return QUEST;
    } else if (c == '"') {	/* ʸ */
	unget(c);
	get_string(buf, BUFSIZE);
	yylval.string = strdup(buf);
	return STRING;
    } else if (c == '\'') {	/* ʸ */
	unget(c);
	get_string(buf, BUFSIZE);
	yylval.ival = buf[0];
	return CHARACTER;
    } else if (c == '#') {	/* #... */
	while ((c = get()) == ' ' || c == '\t') ;
	if (!isalpha(c) && c != '_') {
	    while ((c = get()) != '\n' && c != EOF) ;
	    goto try_again;
	}
	unget(c);
	get_identifier(buf, BUFSIZE);
	if (strcmp(buf, "line") == 0) {
	    int lineno;
	    char filename[TRANS_FILENAME_BUFSIZE];

	    while ((c = get()) == ' ' || c == '\t') ;
	    unget(c);
	    if (isdigit(c)) {
		get_identifier(buf, BUFSIZE);
		lineno = atoi(buf) - 1;
		while ((c = get()) == ' ' || c == '\t') ;
		unget(c);
		if (c == '\"') { /*"*/
		    get_string(filename, TRANS_FILENAME_BUFSIZE);
		    trans_lineno = lineno;
		    strcpy(trans_inputfile, filename);
		} else {
		    trans_lineno = lineno;
		}
	    }
	} else if (strcmp(buf, "pragma") == 0 || strcmp(buf, "vios") == 0) {
	    while ((c = get()) == ' ' || c == '\t') ;
	    unget(c);
	    i = 0;
	    while ((c = get()) != '\n' && c != EOF) {
		buf[i++] = c;
	    }
	    buf[i] = '\0';
	    unget(c);
	    parse_pragma(buf);
	} else if (strcmp(buf, "define") == 0) {
	    macro_define(get, unget, get_identifier, get_string);
	    goto try_again;
	} else if (strcmp(buf, "undef") == 0) {
	    macro_undef(get, unget, get_identifier);
	    goto try_again;
	} else {
	    gen_outf("#%s", buf);
	    while ((c = get()) != '\n' && c != EOF) {
		if (c == '\\') {
		    if ((c = get()) != EOF) {
			gen_outf("\\%c", c);
		    }
		} else {
		    gen_outf("%c", c);
		}
	    }
	    if (c == '\n') gen_outf("\n");
	    unget(c);
	}
	while ((c = get()) != '\n' && c != EOF) ;
	goto try_again;
    } else {
	return c;
    }
}

/** ̻ҤʤΤͽʤΤĴ٤ **/

static struct keyword_table {
    char *keyword;
    int  token;
    int  lval;
} keyword_table[] = {
    {"break",		BREAK,		0},
    {"case",		CASE,		0},
    {"char",		TYPE,		tp_char},
    {"continue",	CONTINUE,	0},
    {"default",		DEFAULT,	0},
    {"do",		DO,		0},
    {"else",		ELSE,		0},
    {"extern",		EXTERN,		0},
    {"float",		TYPE,		tp_float},
    {"float_image",	TYPE,		tp_float_image},
    {"for",		FOR,		0},
    {"goto",		GOTO,		0},
    {"if",		IF,		0},
    {"int",		TYPE,		tp_int},
    {"int_image",	TYPE,		tp_int_image},
    {"return",		RETURN,		0},
    {"switch",		SWITCH,		0},
    {"while",		WHILE,		0},
    {"void",		TYPE,		tp_void},

    {"main",		MAIN,		0},
    {"module",		MODULE,		0},
    {"complex_image",	TYPE,		tp_complex_image},
    {"complex",		TYPE,		tp_complex},
    {"image",		TYPE,		tp_image},
    {"input",		INPUT,		0},
    {"output",		OUTPUT,		0},
    {"parameter",	PARAMETER,	0},
    {"var",		VAR,		0},
    {"on",		ON,		0},
    {"cache",		CACHE,		0},
    {"pixel",		PIXEL,		0},
    {"row",		ROW,		0},
    {"column",		COLUMN,		0},
    {"box",		BOX,		0},
    {"all",		ALL,		0},
    {"parallel",	PARALLEL,	0},
    {"foreach",		FOREACH,	0},

    {NULL,		0,		0},
};

static int
check_keyword(char *s)
{
    struct keyword_table *tp;

    for (tp = keyword_table; tp->keyword != NULL; tp++) {
	if (strcmp(tp->keyword, s) == 0) { /** ߤĤ **/
	    if (tp->token > 0) {
		yylval.ival = tp->lval;
	    }
	    return tp->token;
	}
    }
    return -1;
}
