/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  transmain.c
    main program for debug
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include "transdefs.h"
#include "trans.h"

#ifdef DEBUG
#include <signal.h>
#endif

#define VPEPBUFSIZE (1024*1024)

static void
usage(char *progname)
{
#ifdef YYDEBUG
    fprintf(stderr, "usage: %s [-d] sourcefile ...\n", progname);
#else
    fprintf(stderr, "usage: %s sourcefile ...\n", progname);
#endif
    exit(1);
}

static int
readfile(char *name, char *prg_buf, int buf_size)
{
    FILE *fp;
    char *name2;
    long size;
    int  len;

    if (strcmp(name, "-") == 0) {
	fp = stdin;
	name2 = "stdin";
    } else {
	if ((fp = fopen(name, "r")) == NULL) {
	    perror(name);
	    return 1;
	}
	name2 = name;
    }
    sprintf(prg_buf, "#line 1 \"%s\"\n", name2);
    len = strlen(prg_buf);
    size = fread(prg_buf + len, 1, buf_size - len, fp);
    if (strcmp(name, "-") != 0) fclose(fp);
    else clearerr(stdin);
    if (size < buf_size - len) {
	prg_buf[size + len] = '\0';
	return 0;
    } else {
	fprintf(stderr, "%s: file size is too large\n", name2);
	fprintf(stderr, "currently, buffer size is about %d bytes\n",
		buf_size);
	return 1;
    }
}

#ifdef DEBUG

#include <unistd.h>

static void
myitoa(char *ptr, int n)
{
    if (n > 9999) n = 9999;
    if (n > 999) {
	*ptr++ = (n / 1000) + '0';
	n = n % 1000;
    }
    if (n > 99) {
	*ptr++ = (n / 100) + '0';
	n = n % 100;
    }
    if (n > 9) {
	*ptr++ = (n / 10) + '0';
	n = n % 10;
    }
    *ptr++ = n + '0';
}

static void
sigfunc(int sig)
{
    extern int trans_lineno;
    char mesg[256] = "ֹ    ǥʥ    ȯޤ\n";
    int mesg_len = 41;

    myitoa(mesg+6, trans_lineno);
    myitoa(mesg+20, sig);
    write(2, mesg, mesg_len);
    abort();
    exit(trans_lineno);
}

#endif

FILE* outfp;

int
main(int argc, char *argv[])
{
    int 	j;
    int 	ret;
    static char	prg_buf[VPEPBUFSIZE];
#ifdef DEBUG
    int 	repeat_flag = 0;
#endif

    setlocale(LC_ALL, "");
#ifdef DEBUG
    signal(SIGSEGV, sigfunc);
    signal(SIGBUS, sigfunc);
#endif

    outfp = fopen("transout.prg", "w");
    if (outfp == NULL) {
	fprintf(stderr, "cannot open output file, use stdout\n");
	outfp = stdout;
    }

    ret = 0;

    for (j = 1; j < argc; j++) {

	if (argv[j][0] == '-') {
#ifdef YYDEBUG
	    if (strcmp(argv[j], "-d") == 0 || strcmp(argv[j], "-debug") == 0) {
		extern int yydebug;
		yydebug = 1;
		continue;
	    }
#endif
#ifdef DEBUG
	    if (strcmp(argv[j], "-repeat") == 0) {
		repeat_flag = 1;
		continue;
	    }
#endif
	    usage(argv[0]); /* & exit */
	}

	if (readfile(argv[j], prg_buf, VPEPBUFSIZE) == 0) {
	    ret += translate(prg_buf);
	}
#ifdef DEBUG
	if (repeat_flag) {
	    char buf[256];
	    fprintf(stderr, "1Ľޤ³ޤ[y(default) or n] ");
	    fflush(stderr);
	    fgets(buf, 256, stdin);
	    if (*buf == 'y' || *buf == '\n')
		j--;
	}
#endif
    }
    fclose(outfp);

    return ret;
}

int
module_set_param(char* name, int in_num, int out_num, int param_num,
	div_type divtype[], int boxsize[], int cache[], int rtype, int etime)
/*ARGSUSED*/
{
    int i;
    printf("name: %s\n", name);
    printf("%d %d %d\n", in_num, out_num, param_num);
    for (i = 0; i < in_num+out_num; i++) {
	printf("%d ", divtype[i]);
	if (divtype[i] == dt_box) {
	    printf("%d %d ", boxsize[i*2], boxsize[i*2+1]);
	}
	printf("cache %d\n", cache[i]);
    }
    printf("rdiv:  %d\n", rtype);
    printf("etime: %d\n", etime);
    return 0;
}

int
translator_output(char *data)
{
    fprintf(outfp, "%s", data);
    return 0;
}

int
translator_error(char *msg)
{
    fprintf(stderr, "%s", msg);
    return 0;
}

#ifdef DEBUG

#include <stdarg.h>

void
dpf(char *form, ...)
{
    va_list vp;
    char buf[4096], *p;
    int headflag;

    va_start(vp, form);
    vsprintf(buf, form, vp);
    va_end(vp);

    headflag = 1;
    for (p = buf; *p != '\0'; p++) {
	if (headflag) {
	    headflag = 0;
	    printf(">");
	}
	if (*p == '\n') {
	    printf("<");
	    headflag = 1;
	}
	putchar(*p);
    }
}

#endif
