/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  tree.c
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "transdefs.h"
#include "tree.h"

static Tree *
new_node(void)
{
    Tree *tp;

    if ((tp = (Tree *)malloc(sizeof(Tree))) == NULL) {
	trans_error("out of memory\n");
	return NULL;
    }
    tp->ch1 = tp->ch2 = tp->ch3 = NULL;
    tp->line_no = trans_lineno;
    return tp;
}

Tree *
newnode(void)
{
    Tree *tr;

    if ((tr = new_node()) == NULL) return NULL;
    tr->nodetype = nt_null;
    return tr;
}

/* : sval ϥԡΤޤ޻Ȥ */
Tree *
newnodes(char *sval)
{
    Tree *tr;

    if ((tr = new_node()) == NULL) return NULL;
    tr->node.string = sval;
    tr->nodetype = nt_string;
    return tr;
}

Tree *
newnodei(int ival)
{
    Tree *tr;

    if ((tr = new_node()) == NULL) return NULL;
    tr->node.ival = ival;
    tr->nodetype = nt_int;
    return tr;
}

Tree *
newnodeo(int op1, int op2)
{
    Tree *tr;

    if ((tr = new_node()) == NULL) return NULL;
    tr->node.op.op1 = op1;
    tr->node.op.op2 = op2;
    tr->nodetype = nt_op;
    return tr;
}

Tree *
newnodest(statement_t st)
{
    Tree *tr;

    if ((tr = new_node()) == NULL) return NULL;
    tr->node.st = st;
    tr->nodetype = nt_statement;
    return tr;
}

Tree *
treedup(Tree *tr)
{
    Tree *newtr;

    if (tr == NULL) return NULL;
    if ((newtr = new_node()) == NULL) return NULL;
    *newtr = *tr;
    if (newtr->nodetype == nt_string) {
	newtr->node.string = strdup(newtr->node.string);
	if (newtr->node.string == NULL) {
	    free_tree(newtr);
	    return NULL;
	}
    }
    newtr->ch1 = treedup(tr->ch1);
    newtr->ch2 = treedup(tr->ch2);
    newtr->ch3 = treedup(tr->ch3);
    return newtr;
}


/*
 * ڤ
 * ΤȤΡɤηnt_stringȤ
 */
void
free_tree(Tree *tp)
{
    if (tp == NULL) return;
    free_tree(tp->ch1);
    free_tree(tp->ch2);
    free_tree(tp->ch3);
    if (tp->nodetype == nt_string)
	free(tp->node.string);
    free(tp);
}

/** for debug only **/
static char *
op2string(int op1, int op2)
{
    static char buf[10];

    switch (op1) {
      case op_func:	 return "(func)";
      case op_array:	 return "[]";
      case op_arrow:	 return "->";
      case op_dot:	 return ".";
      case op_unop:	 sprintf(buf, "unop %c", op2); return buf;
      case op_incop:	 buf[0] = buf[1] = op2; buf[2] = '\0'; return buf;
      case op_cast:	 return "(cast)";
      case op_sizeof:	 return "sizeof";
      case op_mul:	 return "*(mul)";
      case op_div:	 return "/";
      case op_mod:	 return "%";
      case op_addop:	 buf[0] = op2; buf[1] = '\0'; return buf;
      case op_shift:	 buf[0] = buf[1] = op2; buf[2] = '\0'; return buf;
      case op_compop:	 switch (op2) {	case '>': return ">";
					case ')': return ">=";
					case '<': return "<";
					case '(': return "<=";
					default: return "uknown compop"; }
      case op_eqop:	 buf[0] = '='; buf[1] = op2; buf[2] = '\0'; return buf;
      case op_and:	 return "&";
      case op_xor:	 return "^";
      case op_or:	 return "|";
      case op_andand:	 return "&&";
      case op_oror:	 return "||";
      case op_quest:	 return "? :";
      case op_assign:	 return "=";
      case op_assignop:	 if (op2 == '<' || op2 == '>') {
			     buf[0] = buf[1] = op2;
			     buf[2] = '=';
			     buf[3] = '\0';
			 } else {
			     buf[0] = op2;
			     buf[1] = '=';
			     buf[2] = '\0';
			 }
			 return buf;
      case op_comma:	 return ",";
      case op_constant:	 return "constant";
      case op_fconstant: return "fconstant";
      case op_id:	 return "identifier";
      case op_parent:	 return "(exp)";
      case op_string:	 return "string";
      case op_charcter:	 return "char";
      default:		 return "unknown op";
    }
}

static char *
st2string(statement_t st)
{
    switch (st) {
      case st_parallel:    return "parallel";
      case st_sequential:  return "sequential";
      case st_comp:        return "comp";
      case st_decl_list:   return "decl_list";
      case st_decl:        return "decl";
      case st_stat_list:   return "stat_list";
      case st_if:          return "if";
      case st_switch:      return "switch";
      case st_for:         return "for";
      case st_while:       return "while";
      case st_do_while:    return "do_while";
      case st_foreach:     return "foreach";
      case st_goto:        return "goto";
      case st_continue:    return "continue";
      case st_break:       return "break";
      case st_return:      return "return";
      case st_labeled:     return "labeled";
      case st_case:        return "case";
      case st_default:     return "default";
      case st_exp:         return "exp";
      default:             return "unknwon";
    }
}

static void
print_string(FILE *fp, char *s)
{
    while (*s != '\0') {
	if (*s == '\n') {
	    fprintf(fp, "\\n");
	} else {
	    putc(*s, fp);
	}
	s++;
    }
}


static void
pr_tr(Tree *tr, int lv)
{
    int i;

    for (i = 0; i < lv*2; i++) fprintf(stderr, " ");
    if (tr == NULL) { fprintf(stderr, "<>\n"); return; }
    switch (tr->nodetype) {
      case nt_null:
	fprintf(stderr, "empty node\n");
	break;
      case nt_string:
	fprintf(stderr, "string:`");
	print_string(stderr, tr->node.string);
	fprintf(stderr, "'\n");
	break;
      case nt_int:
	fprintf(stderr, "int:%d\n", tr->node.ival);
	break;
      case nt_float:
	fprintf(stderr, "float:%g\n", tr->node.fval);
	break;
      case nt_type:
	fprintf(stderr, "type:%d-%d-%d-%d\n", tr->node.type.primary,
		tr->node.type.isimage, tr->node.type.isglobal,
		tr->node.type.isparam);
	break;
      case nt_op:
	fprintf(stderr, "op %s\n",
		op2string(tr->node.op.op1, tr->node.op.op2));
	break;
      case nt_statement:
	fprintf(stderr, "st:%s\n", st2string(tr->node.st));
	break;
    }
    lv++;
    if (tr->ch3 != NULL) {
	pr_tr(tr->ch1, lv);
	pr_tr(tr->ch2, lv);
	pr_tr(tr->ch3, lv);
    } else if (tr->ch2 != NULL) {
	pr_tr(tr->ch1, lv);
	pr_tr(tr->ch2, lv);
    } else if (tr->ch1 != NULL) {
	pr_tr(tr->ch1, lv);
    }
}

void
print_tree(Tree *tp)
{
    (void)print_tree;
    pr_tr(tp, 0);
    fflush(stderr);
}
