/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  image.c
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "image.h"
#include "tcl.h"
#include "tk.h"

/* ȤꤢΤߡviosPixelint˷ǤƤ롩*/

static int      create_pixmap_from_image(Tcl_Interp* interp, vpe_image *image);
static vpe_image *get_image(void *fileimage, int size);
static vpe_image *get_image_pnm(void *fileimage, int size);
     

/**** show_image: βɽ ****/

void
show_image(Tcl_Interp* interp, void *fileimage, int size)
{
    vpe_image *image;
    int       st;

     
    image = get_image(fileimage, size);
    if (image == NULL) {
      fprintf(stderr, "get_image failed\n");
      return;
    }
    st = create_pixmap_from_image(interp, image);
    free(image->image);
    free(image);
    if (st != 0) {
      fprintf(stderr, "create_pixmap_from_image failed\n");
      return;
    }
}

static int make_gray_image(vpe_image *image);
static int make_rgb_image(vpe_image *image);

/** VIOS β X  Pixmap  **/
/** ɽԤäƤ                 **/

static int
create_pixmap_from_image(Tcl_Interp* interp, vpe_image *image)
{
  Tk_PhotoImageBlock viosInfo;
  Tk_PhotoHandle     hndle;
  char buf[256];

  /* imageϰϤѴ */

    if (image->img_class == 1) {
	make_gray_image(image);
    } else {
	make_rgb_image(image);
    }
  
  /* Tcl/tkimageѴ */

  viosInfo.pixelPtr  = image->image;
  viosInfo.width     = image->width;
  viosInfo.height    = image->height;
  if(image->img_class == 1){ /* 졼 */
    viosInfo.pixelSize = (sizeof(int));
    viosInfo.pitch     = (sizeof(int)*image->width);
    viosInfo.offset[0] = 0;
    viosInfo.offset[1] = 0;
    viosInfo.offset[2] = 0;
  }else {                    /* 顼 */
    viosInfo.pixelSize = sizeof(int)*3;
    viosInfo.pitch     = (sizeof(int)*3*image->width);
    viosInfo.offset[0] = 0;
    viosInfo.offset[1] = sizeof(int);
    viosInfo.offset[2] = sizeof(int)*2;
  }
  /* Tcl¦ΥХѿdisp */  
  hndle = Tk_FindPhoto(interp, "disp");
  Tk_PhotoPutBlock(hndle, &viosInfo, 0, 0, 
		   viosInfo.width, viosInfo.height);

  /* ɽ */
  strcpy(buf,"ShowDisp");
  if(Tcl_Eval(interp, buf) == TCL_ERROR)
    fprintf(stderr, "Interper error: %s\n",Tcl_GetStringResult(interp));

  return 0;
}

static int
make_gray_image(vpe_image *image)
{
    int 	gray;
    int		x, y;
    int		max, min, diff;

    /* find max & min */
    max = min = ((int *)image->image)[0];
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    gray = ((int *)image->image)[image->width * y + x];
	    if (gray > 1000000) gray = 0;  /* Ȥꤢ */
	    if (gray < -1000000) gray = 0; /* Ȥꤢ */
	    if (max < gray) max = gray;
	    if (min > gray) min = gray;
	}
    }
    diff = max - min;
    if (diff < 1) diff = 1;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    gray = ((int *)image->image)[image->width * y + x];
	    if (gray >  1000000) gray = 0; /* Ȥꤢ */
	    if (gray < -1000000) gray = 0; /* Ȥꤢ */
	    gray = (gray - min) * 255 / diff;
	    if (gray < 0) gray = 0;
	    if (gray > 255) gray = 255;
	    ((int *)image->image)[image->width * y + x] = gray;
	}
    }
    return 0;
}

static int
make_rgb_image(vpe_image *image)
{
    int		x, y;
    int		max, min, diff;
    int 	rgb_i, value, rgb[3];
    int 	*img;

    img = (int *)image->image;

    /* find max & min */
    max = min = ((int *)image->image)[0];
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (rgb_i = 0; rgb_i < 3; rgb_i++) {
		value = img[(image->width * y + x) * 3 + rgb_i];
		if (max < value) max = value;
		if (min > value) min = value;
	    }
	}
    }
    diff = max - min;
    if (diff < 1) diff = 1;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (rgb_i = 0; rgb_i < 3; rgb_i++) {
		rgb[rgb_i] = img[(image->width * y + x)*3 + rgb_i];
		rgb[rgb_i] = (rgb[rgb_i] - min) * 0xffff / diff;
		if (rgb[rgb_i] <      0) rgb[rgb_i] = 0;
		if (rgb[rgb_i] > 0xffff) rgb[rgb_i] = 0xffff;
		img[(image->width * y + x)*3 + rgb_i] = rgb[rgb_i]; 
	    }
	}
    }
    return 0;
}

/**** fileimage  vpe_image  ****/

int
get_int(void *ptr)
{
    unsigned char *p = (unsigned char *)ptr;
    return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | (p[3]);
}

/* ե뤫 vpe_image βѴ */

vpe_image *
get_image(void *fileimage_param, int size)
{
    vpe_image *image;
    int width, height;
    int d, x, y, c;
    int n;
    int v, values;
    char *fileimage = (char *)fileimage_param;

    /* magic Υå */
    if (fileimage[0] == 'P') {
	return get_image_pnm(fileimage_param, size);
    } else if (strncmp(fileimage, "!OBJ", 4) != 0) {
	return NULL;
    }

    if ((image = (vpe_image *)malloc(sizeof(vpe_image))) == NULL) {
	fprintf(stderr, "get_image: malloc failed\n");
	return NULL;
    }

    /* إå */
    width  = get_int(fileimage + 4);
    height = get_int(fileimage + 8);
    d      = get_int(fileimage + 12);
    /* fileimage + 16, 64: name */
    x      = get_int(fileimage + 80);
    y      = get_int(fileimage + 84);
    c      = get_int(fileimage + 88);

    image->width     = width;
    image->height    = height;
    image->type      = d;
    image->img_class = c;

    if (d != 1) {
	fprintf(stderr, "d != 1\n");
	return NULL;
    }
    if (c == 1) {
	values = 1;
    } else if (c == 2) {
	values = 3;
    } else {
	fprintf(stderr, "c != 1 or 2\n");
	return NULL;
    }

    /*  */
    if ((image->image = malloc(width*height*values*sizeof(int))) == NULL) {
	fprintf(stderr, "get_image: malloc failed\n");
	return NULL;
    }
    n = 0;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (v = 0; v < values; v++) {
		((int *)(image->image))[n] = get_int(fileimage+(92 + n*4));
		n++;
	    }
	}
    }

    return image;
}

/**** pnm version ****/

static unsigned char *getc_ptr;
static unsigned char *getc_ptr_max;
static int read_number_from_image(int* ret);

static void
getc_from_image_init(void *fileimage, int size)
{
    getc_ptr = (unsigned char *)fileimage;
    getc_ptr_max = getc_ptr + size;
}

static int
getc_from_image(void)
{
    if (getc_ptr < getc_ptr_max) {
	return *getc_ptr++;
    } else {
	return EOF;
    }
}

static int
ungetc_from_image(int c)
{
    (void)c;
    getc_ptr--;
    return 0;
}

static vpe_image *
get_image_pnm(void *fileimage, int size)
{
    vpe_image *image;
    int c1, c2;
    int plain_or_raw;
    int width, height;
    int max_value;
    int x, y;
    int v, values;

    getc_from_image_init(fileimage, size);

    /* magic Υå */
    c1 = getc_from_image(); if (c1 != 'P') return NULL;
    c2 = getc_from_image();
    if (c2 == '2') {
	values = 1;       /* gray */
	plain_or_raw = 0; /* plain */
    } else if (c2 == '5') {
	values = 1;       /* gray */
	plain_or_raw = 1; /* raw */
    } else if (c2 == '3') {
	values = 3;       /* rgb */
	plain_or_raw = 0; /* plain */
    } else if (c2 == '6') {
	values = 3;       /* rgb */
	plain_or_raw = 1; /* raw */
    } else {
	fprintf(stderr, "get_image_pnm: not pgm or ppm file\n");
	return NULL;
    }
    
    /* إåɤ߹ */
    if (read_number_from_image(&width    ) != 0) return NULL;
    if (read_number_from_image(&height   ) != 0) return NULL;
    if (read_number_from_image(&max_value) != 0) return NULL;

    if (plain_or_raw == 1) {
	/* raw ΤȤ max_value μˤʤ⤷ʤ */
	/* ڡɤФƤ */
	int c;
	c = getc_from_image();
	if (!isspace(c))
	    ungetc_from_image(c);
    }
    /*  */

    if ((image = (vpe_image *)malloc(sizeof(vpe_image))) == NULL) {
	fprintf(stderr, "get_image_pnm: malloc failed\n");
	return NULL;
    }

    /* إå */
    image->width  = width;
    image->height = height;
    image->type = 1; /* 1: int */
    image->img_class = (values == 1) ? 1 : 2; /* 1: gray, 2: rgb */

    /*  */
    if ((image->image = malloc(width * height * values * sizeof(int))) == NULL)
    {
	fprintf(stderr, "get_image_pnm: malloc failed\n");
	free(image);
	return NULL;
    }

    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		int ret, value;
		if (plain_or_raw == 0) {
		    /* plain */
		    ret = read_number_from_image(&value);
		} else {
		    /* raw */
		    value = getc_from_image();
		    ret = (value == EOF);
		}
		if (ret) {
		    fprintf(stderr, "get_image_pnm: read error\n");
		    free(image->image); free(image);
		    return NULL;
		}
		((int *)(image->image))[(y*width + x) * values + v] = value;
	    }
	}
    }
    return image;
}

/* pnm file ɤ߹Ѵؿ */

static int get(void);
static void skip_space(void);

/* ȤɤФ1ʸ֤ */
static int
get()
{
    int c;

    for (;;) {
	c = getc_from_image();
	if (c != '#')
	    break;
	while ((c = getc_from_image()) != '\n' && c != EOF)
	    ;
    }
    return c;
}

static void
skip_space()
{
    int c;

    while (c = get(), isspace(c))
	;
    if (c != EOF)
	ungetc_from_image(c);
}

/* ɤ߹֤ */
/* ɤ᤿ 0 ɤʤä 1 ֤ */
static int
read_number_from_image(int* ret)
{
    int i, c;
    char buf[256];

    skip_space();
    c = get();
    i = 0;
    if (c == '-') {
	buf[i++] = c;
	c = get();
    } else {
	ungetc_from_image(c);
    }
    if (isdigit(c)) {
	for (; i < 256; i++) {
	    c = get();
	    if (!isdigit(c)) {
		ungetc_from_image(c);
		buf[i] = '\0';
		break;
	    }
	    buf[i] = c;
	}
	if (i >= 256) {
	    fprintf(stderr, "read_number_from_image: buffer overflow\n");
	    return 1;
	}
	*ret = atoi(buf);
	return 0;
    } else {
	return 1;
    }
}
