/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  pnm2vif.c
    pgm or ppm ե vif եѴ
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <sys/types.h>

typedef enum {
    dt_int = 1, dt_real = 2
} data_type_t;

typedef enum {
    dc_gray = 1, dc_rgb = 2
} data_class_t;

static void usage(char *prog_name);
static int  pnm2vif(FILE *infp, FILE *outfp, data_type_t dt, float scale);
static int  load_header(FILE *fp, int *width, int *height, int *plain_or_raw,
			data_class_t *dc);
static int  write_int(int value, FILE *fp);
static int  write_float(float value, FILE *fp);
static int  read_number(int *ret, FILE *fp);

int
main(int argc, char *argv[])
{
    FILE        *infp, *outfp;
    char        *infile = NULL;
    int         ret;
    data_type_t dt = dt_int;
    int         i;
    float       scale = 1.0;

    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-' && argv[i][1] != '\0') {
	    if (strcmp(argv[i], "-s") == 0 && i+1 < argc) {
		i++;
		scale = atof(argv[i]);
	    } else if (strcmp(argv[i], "-f") == 0) {
		dt = dt_real;
	    } else {
		usage(argv[0]);
	    }
	} else {
	    if (infile == NULL) {
		infile = argv[i];
	    } else {
		usage(argv[0]);
	    }
	}
    }

    if (infile == NULL || strcmp(infile, "-") == 0) {
	infp = stdin;
    } else {
	infp = fopen(infile, "rb");
	if (infp == NULL) {
	    perror(infile);
	    return 1;
	}
    }
    outfp = stdout;

    ret = pnm2vif(infp, outfp, dt, scale);

    if (infp != stdin) {
	fclose(infp);
    }

    return ret;
}

static void
usage(char *prog_name)
{
    fprintf(stderr, "usage: %s [options] [input-filename]\n", prog_name);
    fprintf(stderr, "  -f        : generate float format\n");
    fprintf(stderr, "  -s scale  : set float scale\n");
    exit(1);
}

int
pnm2vif(FILE *infp, FILE *outfp, data_type_t dt, float scale)
{
    int          width, height, plain_or_raw;
    data_class_t dc;
    int          x, y, value;
    int          ret;
    int          v, values;
    char         name[64];
    memset(name, 0, 64);

    if (load_header(infp, &width, &height, &plain_or_raw, &dc) != 0) {
	fprintf(stderr, "pnm file header load error\n");
	return 1;
    }

    /** output **/

    fputs("!OBJ", outfp);
    write_int(width, outfp);
    write_int(height, outfp);
    write_int(dt, outfp);       /* data type */
    fwrite(name, 1, 64, outfp); /* name */
    write_int(0, outfp);        /* hot x */
    write_int(0, outfp);        /* hot y */
    write_int(dc, outfp);       /* data class */

    if (dc == dc_gray) {
	values = 1;
    } else {
	values = 3;
    }

    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		if (plain_or_raw == 0) {
		    /* plain */
		    ret = read_number(&value, infp);
		} else {
		    /* raw */
		    value = getc(infp);
		    ret = (value == EOF);
		}
		if (ret) {
		    fprintf(stderr, "read error\n");
		    return 1;
		}
		if (dt == dt_int) {
		    write_int(value, outfp);
		} else {
		    write_float((float)value * scale, outfp);
		}
	    }
	}
    }

    return 0;
}

static int
load_header(FILE *fp, int *width, int *height, int *plain_or_raw,
	    data_class_t *dc)   /* plain_or_raw: 0: plain, 1: raw */
{
    int magic1, magic2;
    int max_gray_value;

    /* magic Υå */
    magic1 = getc(fp);
    if (magic1 != 'P') {
	return 1;
    }

    magic2 = getc(fp);
    if (magic2 == '2') {
	*plain_or_raw = 0;  /* plain */
	*dc = dc_gray;
    } else if (magic2 == '3') {
	*plain_or_raw = 0;  /* plain */
	*dc = dc_rgb;
    } else if (magic2 == '5') {
	*plain_or_raw = 1;  /* raw */
	*dc = dc_gray;
    } else if (magic2 == '6') {
	*plain_or_raw = 1;  /* raw */
	*dc = dc_rgb;
    } else {
	return 1;
    }
    
    /* إåɤ߹*/
    if (read_number(width,           fp) != 0) return 1;
    if (read_number(height,          fp) != 0) return 1;
    if (read_number(&max_gray_value, fp) != 0) return 1;

    if (*plain_or_raw == 1) { /* raw */
	/* raw ΤȤ max_gray_value μˤʤ⤷ʤ
	   ڡɤФƤ */
	int c;
	c = getc(fp);
	if (!isspace(c))
	    ungetc(c, fp);
    }
    return 0;
}

static int
write_int(int value, FILE *fp)
{
    putc((value >> 24) & 0xff, fp);
    putc((value >> 16) & 0xff, fp);
    putc((value >>  8) & 0xff, fp);
    putc((value      ) & 0xff, fp);
    return 0;
}

static int
write_float(float value, FILE *fp)
{
    unsigned char *p = (unsigned char *)&value;
    putc(p[3], fp);
    putc(p[2], fp);
    putc(p[1], fp);
    putc(p[0], fp);
    return 0;
}

/* pgm file ɤ */

/* ȤɤФ1ʸ֤ */
static int
get(FILE *fp)
{
    int c;

    for (;;) {
	c = getc(fp);
	if (c != '#')
	    break;
	while ((c = getc(fp)) != '\n' && c != EOF)
	    ;
    }
    return c;
}

static void skip_space(FILE *fp)
{
    int c;

    while (c = get(fp), isspace(c))
	;
    if (c != EOF)
	ungetc(c, fp);
}

static int
read_number(int *ret, FILE *fp)
{
    int num;
    int c;

    skip_space(fp);
    c = get(fp);
    ungetc(c, fp);
    if (isdigit(c)) {
	if (fscanf(fp, "%d", &num) == 1) {
	    *ret = num;
	    return 0;
	} else {
	    return 1;
	}
    } else {
	return 1;
    }
}
