/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  vif2pnm.c
    vif ե pgm or ppm եѴ
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void usage(char *prog_name);
static int  vif2pnm(FILE *infp, FILE *outfp, int mode, double scale);
static int  load_header(FILE *fp, int *width, int *height,
			int *data_type, int *data_class);
static int  read_int(int *ret, FILE *fp);
static int  read_float(float *ret, FILE *fp);

int
main(int argc, char *argv[])
{
    FILE *infp, *outfp;
    char *infile = NULL;
    int  ret;
    int  i;
    double scale = 1.0;
    int  mode = 0; /* 0: pnm, 1:text */

    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-' && argv[i][1] != '\0') {
	    if (strcmp(argv[i], "-s") == 0 && i+1 < argc) {
		i++;
		scale = atof(argv[i]);
	    } else if (strcmp(argv[i], "-t") == 0) {
		mode = 1;
	    } else {
		usage(argv[0]);
	    }
	} else {
	    if (infile == NULL) {
		infile = argv[i];
	    } else {
		usage(argv[0]);
	    }
	}
    }

    if (infile == NULL || strcmp(infile, "-") == 0) {
	infp = stdin;
    } else {
	infp = fopen(infile, "rb");
	if (infp == NULL) {
	    perror(infile);
	    return 1;
	}
    }
    outfp = stdout;

    ret = vif2pnm(infp, outfp, mode, scale);

    if (infp != stdin) {
	fclose(infp);
    }

    return ret;
}

static void
usage(char *prog_name)
{
    fprintf(stderr, "usage: %s [options] [input-filename]\n", prog_name);
    fprintf(stderr, "  -t        : generate text format\n");
    fprintf(stderr, "  -s scale  : set float scale\n");
    exit(1);
}

static int
vif2pnm(FILE *infp, FILE *outfp, int mode, double scale)
{
    int width, height, data_type, data_class;
    int x, y;
    int v, values;
    int *img;
    int i;
    int max, min, offset;

    if (load_header(infp, &width, &height, &data_type, &data_class) != 0) {
	fprintf(stderr, "header load error\n");
	return 1;
    }

    /* data_type:  1: int,  2: float */
    if (data_type != 1 && data_type != 2) {
	fprintf(stderr, "unknown data type (%d)\n", data_type);
	return 1;
    }

    if (data_class == 1) {
	/* gray */
	values = 1;
    } else if (data_class == 2) {
	/* rgb */
	values = 3;
    } else {
	fprintf(stderr, "unknown data class (%d)\n", data_class);
	return 1;
    }

    /* read & find max */

    if ((img = malloc(width * height * values * sizeof(int))) == NULL) {
	fprintf(stderr, "malloc failed\n");
	return 1;
    }

    i = 0;
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		int val;
		if (data_type == 1) {
		    if (read_int(&val, infp) != 0) {
			fprintf(stderr, "read error\n");
			return 1;
		    }
		    if (mode == 1) {
			fprintf(outfp, "%d\n", val);
		    }
		} else {
		    float fval;
		    if (read_float(&fval, infp) != 0) {
			fprintf(stderr, "read error\n");
			return 1;
		    }
		    fval *= scale;
		    if (mode == 1) {
			fprintf(outfp, "%g\n", fval);
		    }
		    val = (int)fval;
		}
		if (i == 0 || val < min) {
		    min = val;
		}
		if (i == 0 || val > max) {
		    max = val;
		}
		img[i++] = val;
	    }
	}
    }

    if (mode == 1) {
	return 0;
    }

    if (min < 0) {
	offset = -min;
    } else {
	offset = 0;
    }

    /* write */

    if (data_class == 1) {
	fprintf(outfp, "P2\n");
    } else {
	fprintf(outfp, "P3\n");
    }
    fprintf(outfp, "%d %d\n%d\n", width, height, max + offset);
    i = 0;
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		fprintf(outfp, "%d\n", img[i] + offset);
		i++;
	    }
	}
    }

    return 0;
}

static int
load_header(FILE *fp, int *width, int *height, int *data_type, int *data_class)
{
    char magic[5];
    char name[64];
    int  dummy;

    /* magic Υå */
    if (fread(magic, 1, 4, fp) != 4)
	return 1;
    magic[4] = '\0';
    if (strcmp(magic, "!OBJ") != 0)
	return 1;

    /* إåɤ߹*/
    if (read_int(width,      fp) != 0) return 1;
    if (read_int(height,     fp) != 0) return 1;
    if (read_int(data_type,  fp) != 0) return 1;
    if (fread(name, 1, 64, fp) != 64)  return 1;
    if (read_int(&dummy,     fp) != 0) return 1;
    if (read_int(&dummy,     fp) != 0) return 1;
    if (read_int(data_class, fp) != 0) return 1;

    return 0;
}

static int
read_int(int  *ret, FILE *fp)
{
    int c1, c2, c3, c4;
    
    c1 = getc(fp);
    c2 = getc(fp);
    c3 = getc(fp);
    c4 = getc(fp);
    if (c4 == EOF) return 1;
    *ret = (c1 << 24) | (c2 << 16) | (c3 << 8) | c4;
    return 0;
}

int
read_float(float *ret, FILE  *fp)
{
    int c1, c2, c3, c4;
    float val;
    unsigned char *p = (unsigned char *)&val;
    c1 = getc(fp);
    c2 = getc(fp);
    c3 = getc(fp);
    c4 = getc(fp);
    if (c4 == EOF) return 1;
    *p++ = c4;
    *p++ = c3;
    *p++ = c2;
    *p++ = c1;
    *ret = val;
    return 0;
}
