/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  com.c
    VPE ̿
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include "vpedefs.h"
#include "protocol.h"
#include "msgpopup.h"
#include "com.h"
#include "ipustat.h"
#include "image.h"

/**** connect_om:  host ǻꤵ줿ۥȾ OM ³ ****/
int
connect_om(char* host, int port)
{
    int 		om_fd;
    struct hostent*	server;
    struct sockaddr_in	addr;

    if ((server = gethostbyname(host)) == NULL) {
	msgpopup("cannot get IP address");
	return -1;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family      = AF_INET;
    memcpy(&addr.sin_addr, server->h_addr, server->h_length);
    addr.sin_port        = htons(port);

    if ((om_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	perror("socket");
	return -1;
    }
    if (connect(om_fd, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
	char buf[1024];
	sprintf(buf, "cannot connect to om\n%s", strerror(errno));
	msgpopup(buf);
	return -1;
    }

    net_write(om_fd, "VPE", 4);

    return om_fd;
}

/**** close_om: OM Ȥ³ ****/
int
close_om(int om_fd)
{
    sender(om_fd, p_vo_close, NULL);
    close(om_fd);
    return 0;
}

/**** ͥåȥɤ߽񤭴ؿ ****/

int
net_read(int fd, void* buf, int size)
{
    int 	wc;
    int 	rest = size;
    char*	rbuf = (char*)buf;

    while (rest > 0) {
	if ((wc = read(fd, rbuf, rest)) <= 0) {
	    return -1;
	}
	rbuf  += wc;
	rest  -= wc;
    }
    return 0;
}

int
net_write(int fd, void* buf, int size)
{
    int 	wc;
    int 	rest = size;
    char*	wbuf = (char*)buf;

    while (rest > 0) {
	if ((wc = write(fd, wbuf, rest)) < 0) {
	    perror("net_write: write failed");
	    return -1;
	}
	wbuf += wc;
	rest -= wc;
    }
    return 0;
}

/**** OM إåؿ ****/
void
sender(int om_fd, int message_type, char* message_body)
{
    vpe_om msg;
    unsigned char buf[8];
    int body_size;

    if (message_body != NULL) {
	body_size = strlen(message_body);
    } else {
	body_size = 0;
    }

    msg.type = message_type;
    msg.size = body_size;

    buf[0] = (unsigned char)((msg.type >> 24) & 0xff);
    buf[1] = (unsigned char)((msg.type >> 16) & 0xff);
    buf[2] = (unsigned char)((msg.type >>  8) & 0xff);
    buf[3] = (unsigned char)((msg.type      ) & 0xff);
    buf[4] = (unsigned char)((msg.size >> 24) & 0xff);
    buf[5] = (unsigned char)((msg.size >> 16) & 0xff);
    buf[6] = (unsigned char)((msg.size >>  8) & 0xff);
    buf[7] = (unsigned char)((msg.size      ) & 0xff);
    net_write(om_fd, buf, 8);

    if (body_size > 0) {
	net_write(om_fd, message_body, body_size);
    }
}


/**** om_read_cb: OM ǡ褿Ȥ˸ƤФؿ ****/
void
om_read_cb(XtPointer closure, int *source, XtInputId *id)
/*ARGSUSED*/
{
    unsigned char head[8];
    p_ov_t        type;
    int           size;
    unsigned char *buf;

    /* إåɤ߹ */
    if (net_read(om_fd, head, 8) != 0) {
	fprintf(stderr, "om_read_cb: net_read failed\n");
	close_om(om_fd);
	om_fd = -1;
	XtRemoveInput(*id);
	return;
    }
    type = (int)((head[0] << 24) | (head[1] << 16) | (head[2] << 8) | head[3]);
    size =       (head[4] << 24) | (head[5] << 16) | (head[6] << 8) | head[7];

    /* Ȥɤ߹ */
    if ((buf = (unsigned char *)malloc(size)) == NULL) {
	fprintf(stderr, "om_read_cb: malloc failed\n");
	exit(1);
    }
    if (net_read(om_fd, buf, size) != 0) {
	fprintf(stderr, "om_read_cb: net_read failed\n");
	exit(1);
    }

    /* פ˱ */
    switch (type) {
      case p_ov_ipu_create:
	{
	    int    ipu_no;
	    ipu_no = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
	    ipu_create(ipu_no, (char *)buf+4, size-4);
	}
	break;
      case p_ov_load_data:
	{
	    int    ipu_no;
	    int    iload;
	    double load;
	    ipu_no = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
	    iload  = (buf[4] << 24) | (buf[5] << 16) | (buf[6] << 8) | buf[7];
	    load = iload / 256.0;
	    ipu_load(ipu_no, load);
	}
	break;
      case p_ov_main_translate_stat:
      case p_ov_module_translate_stat:
      case p_ov_compile_stat:
	/* ѥη̤ɽ */
	{
	    int cp_stat;
	    char *trans_or_comp;
	    char stbuf[256];
	    if (type == p_ov_main_translate_stat) {
		trans_or_comp = "main translate";
	    } else if (type == p_ov_module_translate_stat) {
		trans_or_comp = "module translate";
	    } else {
		trans_or_comp = "compile";
	    }
	    cp_stat = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
	    if (cp_stat == 0) {
		sprintf(stbuf, "%s ok", trans_or_comp);
	    } else {
		sprintf(stbuf, "%s error", trans_or_comp, cp_stat);
	    }
	    if (size > 4) {
		/* åΤǤɽ */
		char *s;
		int  stlen = strlen(stbuf);
		if ((s = (char *)malloc(stlen+1 + size-4 + 1)) == NULL) {
		    fprintf(stderr, "malloc failed\n");
		    exit(1);
		}
		sprintf(s, "%s\n", stbuf);
		memcpy(s+stlen+1, buf+4, size-4);
		s[stlen+1 + size-4] = '\0';
		msgpopup(s);
		free(s);
	    } else {
		if (type == p_ov_compile_stat || cp_stat != 0) {
		    /* ȥ󥹥졼Ȥη̤ϥ顼ʤɽʤ */
		    msgpopup(stbuf);
		}
	    }
	    break;
	}
      case p_ov_module_done:
	/* ⥸塼μ¹Ԥä */
	{
	    int ipu_no, jobid, sec, usec;
	    ipu_no = (buf[0] << 24) | (buf[1] << 16) | (buf[2] << 8) | buf[3];
	    jobid  = (buf[4] << 24) | (buf[5] << 16) | (buf[6] << 8) | buf[7];
	    sec    = (buf[8] << 24) | (buf[9] << 16) | (buf[10]<< 8) | buf[11];
	    usec   = (buf[12]<< 24) | (buf[13]<< 16) | (buf[14]<< 8) | buf[15];
	    module_done_time(ipu_no, jobid, sec, usec);
	}
	break;
      case p_ov_image:
	{
	    show_image(buf, size);
	}
	break;
      default:
	printf("unknown data: type %d, size %d!\n", type, size);
	break;
    }
    free(buf);
}
