/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  filesel.c
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Shell.h>
#include "filesel.h"
#include "msgpopup.h"

static void popdown_filesel(void);
static void delete_fs_popup(Widget toplevel);
static void ok_action(Widget w, XEvent *event, String *params, Cardinal *n_params);
static void ok_cb(Widget w, XtPointer p1, XtPointer p2);

static Widget filesel_popup;
static Widget text;
static filesel_cb_stat (*filesel_callback)(char *filename);

void
create_fs_popup(Widget toplevel)
{
    Widget frame1;
    Widget frame2;
    Widget ok, cancel;
    char *translations = "\
Ctrl<Key>J:    fs_ok()\n\
Ctrl<Key>O:    fs_ok()\n\
Ctrl<Key>M:    fs_ok()\n\
<Key>Linefeed: fs_ok()\n\
<Key>Return:   fs_ok()\n\
";
    XtActionsRec action_table[] = {
	{"fs_ok", ok_action},
    };

    filesel_popup = XtVaCreatePopupShell("fileSelector",
			topLevelShellWidgetClass, toplevel,
			NULL);
    frame1 = XtVaCreateManagedWidget("frame1", formWidgetClass, filesel_popup,
				     NULL);

    text = XtVaCreateManagedWidget("input", asciiTextWidgetClass, frame1,
				   XtNeditType, XawtextEdit,
				   NULL);
    XtAppAddActions(XtWidgetToApplicationContext(toplevel),
		    action_table, XtNumber(action_table));
    XtOverrideTranslations(text, XtParseTranslationTable(translations));

    frame2 = XtVaCreateManagedWidget("frame2", boxWidgetClass, frame1,
				     NULL);
    ok = XtVaCreateManagedWidget("ok", commandWidgetClass, frame2,
				 NULL);
    cancel = XtVaCreateManagedWidget("cancel", commandWidgetClass, frame2,
				     NULL);
    XtAddCallback(ok, XtNcallback, ok_cb, (XtPointer)0);
    XtAddCallback(cancel, XtNcallback, ok_cb, (XtPointer)1);
}

void
popup_filesel(filesel_cb_stat (*callback)(char *filename))
{
    filesel_callback = callback;
    XtPopup(filesel_popup, XtGrabNone);
}

static void
popdown_filesel(void)
{
    XtPopdown(filesel_popup);
    (void)popdown_filesel;
}

static void
ok_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    char *filename;
    filesel_cb_stat stat;

    if ((int)p1 == 0) {
	/* OK */
	XtVaGetValues(text, XtNstring, &filename, NULL);
	if (*filename == '\0') {
	    /* ϤƤʤΤ */
	    return;
	}
    } else {
	/* cancel */
	filename = NULL;
    }
    stat = (*filesel_callback)(filename);
    if (stat == fcs_ok) {
	/*  */
	XtPopdown(filesel_popup);
    } else {
	/* 顼 */
	if (stat == fcs_err_errno) {
	    /* errno ߤƥ顼åɽ */
	    char *msg = strerror(errno);
	    if (msg == NULL) {
		msg = "ե륪ץ󥨥顼";
	    }
	    msgpopup(msg);
	} else if (stat == fcs_err_other) {
	    /* ¾Υ顼 */
	    msgpopup("cannot open file");
	}
    }
}

static void
ok_action(Widget w, XEvent *event, String *params, Cardinal *n_params)
/*ARGSUSED*/
{
    ok_cb(w, 0, 0);
}

static void
delete_fs_popup(Widget toplevel)
{
    Widget frame;
    Widget popup;
    popup = XtVaCreatePopupShell("fsPopup", topLevelShellWidgetClass,
				 toplevel,
				 NULL);
    frame = XtVaCreateManagedWidget("fs", boxWidgetClass, popup,
				    NULL);

    XtVaCreateManagedWidget("message", labelWidgetClass, frame, NULL);
    XtVaCreateManagedWidget("text", labelWidgetClass, frame, NULL);
    frame = XtVaCreateManagedWidget("fs", commandWidgetClass, frame,
				    NULL);
    (void)delete_fs_popup;
}
