/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  image.c
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/Command.h>
#include "image.h"

static int       create_pixmap_from_image(vpe_image *image, Pixmap *pm_ret,
					  Visual **v_ret, int *depth_ret);
static void      alloc_color(Display *display, Colormap cmap, XColor *col);
static vpe_image *get_image(void *fileimage, int size);
static void      create_image_popup(Pixmap pixmap, Visual *visual, int depth);
static void      image_close_cb(Widget w, XtPointer p1, XtPointer p2);
static vpe_image *get_image_pnm(void *fileimage, int size);

extern Widget toplevel;

/**** show_image: βɽ ****/

void
show_image(void *fileimage, int size)
{
    vpe_image *image;
    Pixmap    pixmap;
    int       st;
    Visual    *visual;
    int       depth;

    image = get_image(fileimage, size);
    if (image == NULL) {
	fprintf(stderr, "get_image failed\n");
	return;
    }
    st = create_pixmap_from_image(image, &pixmap, &visual, &depth);
    free(image->image);
    free(image);
    if (st != 0) {
	fprintf(stderr, "create_pixmap_from_image failed\n");
	return;
    }

    create_image_popup(pixmap, visual, depth);
}

/* 顼ɽѤΥӥ奢õ */
static Visual *
find_color_visual(Display *display, int screen, int *depth_ret)
{
    XVisualInfo  *vp;
    XVisualInfo  vtemp;
    int          nvs;
    Visual       *visual;
    int          i;
    int          depth;

    display = XtDisplay(toplevel);
    screen = DefaultScreen(display);

    vtemp.screen = screen;
    vp = XGetVisualInfo(display, VisualScreenMask, &vtemp, &nvs);
    visual = NULL;
    if (vp != NULL) {
	depth = 0;
	for (i = 0; i < nvs; i++) {
	    if (vp[i].class == TrueColor) {
		if (vp[i].depth > depth) {
		    depth = vp[i].depth;
		    visual = vp[i].visual;
		}
	    }
	}
	XFree(vp);
    }
    *depth_ret = depth;

    return visual;
}

static int make_gray_image(vpe_image *image, XImage *ximage);
static int make_rgb_image(vpe_image *image, XImage *ximage, Visual *visual);

/** VIOS β X  Pixmap  **/
static int
create_pixmap_from_image(vpe_image *image, Pixmap *pixmap_ret,
			 Visual **v_ret, int *depth_ret)
{
    int		depth, pad, bytes_per_line;
    XImage	*ximage;
    Display	*display;
    Visual	*visual;
    Pixmap	pixmap;
    static GC   pixmap_gc;

    display = XtDisplay(toplevel);
    if (image->img_class == 1) {
	/* gray */
	visual = DefaultVisual(display, DefaultScreen(display));
	depth = DefaultDepth(display, DefaultScreen(display));
    } else if (image->img_class == 2) {
	/* rgb */
	visual = find_color_visual(display, DefaultScreen(display), &depth);
	if (visual == NULL) {
	    fprintf(stderr, "cannot find visual\n");
	    return -1;
	}
    } else {
	fprintf(stderr, "bad image class (%d)\n", image->img_class);
	return -1;
    }
    *v_ret = visual;
    *depth_ret = depth;
    pad = 32;
    bytes_per_line = 0;
    if ((ximage = XCreateImage(display, visual, depth, ZPixmap, 0, NULL,
		image->width, image->height, pad, bytes_per_line)) == NULL)
    {
	fprintf(stderr, "cannot create ximage\n");
	return -1;
    }
    if ((ximage->data =
	 malloc(ximage->bytes_per_line * ximage->height)) == NULL)
    {
	fprintf(stderr, "malloc failed\n");
	XDestroyImage(ximage);
	return -1;
    }

    if (image->img_class == 1) {
	make_gray_image(image, ximage);
    } else {
	make_rgb_image(image, ximage, visual);
    }

    pixmap = XCreatePixmap(display, XtWindow(toplevel),
			   image->width, image->height, depth);
    if (pixmap_gc == 0) {
	pixmap_gc = XCreateGC(display, pixmap, 0, NULL);
    }
    XPutImage(display, pixmap, pixmap_gc, ximage, 0,0,0,0,
	      ximage->width, ximage->height);

    free(ximage->data);
    ximage->data = NULL;
    XDestroyImage(ximage);

    *pixmap_ret = pixmap;
    return 0;
}

static int
make_gray_image(vpe_image *image, XImage *ximage)
{
    int 	gray;
    Display	*display = XtDisplay(toplevel);
    Colormap	cmap = DefaultColormap(display, DefaultScreen(display));
    XColor	col;
    Pixel	alloced_pixel[256];
    int		x, y;
    int		max, min, diff;

    for (gray = 0; gray < 256; gray++) {
	col.red = col.green = col.blue = gray | (gray << 8);
	alloc_color(display, cmap, &col);
	alloced_pixel[gray] = col.pixel;
    }

    /* find max & min */
    max = min = ((int *)image->image)[0];
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    gray = ((int *)image->image)[image->width * y + x];
	    if (gray > 1000000) gray = 0; /* Ȥꤢ */
	    if (gray < -1000000) gray = 0; /* Ȥꤢ */
	    if (max < gray) max = gray;
	    if (min > gray) min = gray;
	}
    }
    diff = max - min;
    if (diff < 1) diff = 1;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    gray = ((int *)image->image)[image->width * y + x];
	    if (gray >  1000000) gray = 0; /* Ȥꤢ */
	    if (gray < -1000000) gray = 0; /* Ȥꤢ */
	    gray = (gray - min) * 255 / diff;
	    if (gray < 0) gray = 0;
	    if (gray > 255) gray = 255;
	    XPutPixel(ximage, x, y, alloced_pixel[gray]);
	}
    }
}

static void
calc_shift_bits(int *shift, int *bits, unsigned int mask)
{
    unsigned int work;
    int s, b;

    if (mask == 0) {
	*shift = 0;
	*bits = 1;
	return;
    }

    work = mask;
    for (s = 0; (work & 1) == 0; s++) {
	work >>= 1;
    }
    *shift = s;

    for (b = 0; (work & 1) == 1; b++) {
	work >>= 1;
    }
    *bits = b;
}

static Pixel
calc_pixel(int rgb[3], int shift[3], int bits[3])
{
    Pixel p = 0;
    int   i;
    unsigned int tmp;

    for (i = 0; i < 3; i++) {
	if (bits[i] > 16) {
	    tmp = rgb[i] << (bits[i] - 16);
	} else {
	    tmp = rgb[i] >> (16 - bits[i]);
	}
	p |= tmp << shift[i];
    }
    return p;
}

static int
make_rgb_image(vpe_image *image, XImage *ximage, Visual *visual)
{
    Display	*display;
    Colormap	cmap;
    XVisualInfo	*vp;
    XVisualInfo vtemp;
    int 	nvs;
    int 	shift[3], bits[3];
    int		x, y;
    int		max, min, diff;
    int 	rgb_i, value, rgb[3];
    int 	*img;
    Pixel	pixel;

    display = XtDisplay(toplevel);
    cmap = DefaultColormap(display, DefaultScreen(display));

    vtemp.visualid = XVisualIDFromVisual(visual);
    vp = XGetVisualInfo(display, VisualIDMask, &vtemp, &nvs);
    if (vp == NULL) {
	fprintf(stderr, "cannot get visual info\n");
	return -1;
    }
    calc_shift_bits(&shift[0], &bits[0], vp->red_mask);
    calc_shift_bits(&shift[1], &bits[1], vp->green_mask);
    calc_shift_bits(&shift[2], &bits[2], vp->blue_mask);
    XFree(vp);

    img = (int *)image->image;

    /* find max & min */
    max = min = ((int *)image->image)[0];
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (rgb_i = 0; rgb_i < 3; rgb_i++) {
		value = img[(image->width * y + x) * 3 + rgb_i];
		if (max < value) max = value;
		if (min > value) min = value;
	    }
	}
    }
    diff = max - min;
    if (diff < 1) diff = 1;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (rgb_i = 0; rgb_i < 3; rgb_i++) {
		rgb[rgb_i] = img[(image->width * y + x)*3 + rgb_i];
		rgb[rgb_i] = (rgb[rgb_i] - min) * 0xffff / diff;
		if (rgb[rgb_i] <      0) rgb[rgb_i] = 0;
		if (rgb[rgb_i] > 0xffff) rgb[rgb_i] = 0xffff;
	    }
	    pixel = calc_pixel(rgb, shift, bits);
	    XPutPixel(ximage, x, y, pixel);
	}
    }
}

/** οݤ **/
static void
alloc_color(Display *display, Colormap cmap, XColor *col)
{
    if (XAllocColor(display, cmap, col) != 0) {
	/*  */
	return;
    }
    /* ԡᤤ򤵤 */
    /* not implemented */
    return;
}

typedef struct popdown_arg {
    Widget popup;
    Pixmap pixmap;
} popdown_arg;

/**  Pixmap ĥݥåץåץɥäƥݥåץåפ **/
static void
create_image_popup(Pixmap pixmap, Visual *visual, int depth)
{
    Widget popup;
    Widget frame;
    Widget button;
    Widget image;
    Window root;
    int x, y;
    unsigned int width, height, border, pdepth;
    Display *display;
    popdown_arg *arg;

    display = XtDisplay(toplevel);

    XGetGeometry(display, pixmap, &root, &x, &y,
		 &width, &height, &border, &pdepth);

    if (visual == DefaultVisual(display, DefaultScreen(display)) &&
	depth  == DefaultDepth(display, DefaultScreen(display)))
    {
	popup = XtVaCreatePopupShell("imagePopup", topLevelShellWidgetClass,
				     toplevel,
				     NULL);
    } else {
	Colormap cmap;
	cmap = XCreateColormap(display, XtWindow(toplevel), visual, AllocNone);
	popup = XtVaCreatePopupShell("imagePopup", topLevelShellWidgetClass,
				     toplevel,
				     XtNvisual, visual,
				     XtNdepth, depth,
				     XtNcolormap, cmap,
				     NULL);
    }

    frame = XtVaCreateManagedWidget("frame", boxWidgetClass,
				    popup, NULL);
    image = XtVaCreateManagedWidget("image", labelWidgetClass,
				    frame,
				    XtNwidth, width,
				    XtNheight, height,
				    XtNbitmap, pixmap,
				    NULL);
    button = XtVaCreateManagedWidget("close", commandWidgetClass,
				     frame, NULL);
    arg = malloc(sizeof(popdown_arg));
    if (arg == NULL) {
	fprintf(stderr, "malloc failied\n");
	exit(1);
    }
    arg->popup = popup;
    arg->pixmap = pixmap;
    XtAddCallback(button, XtNcallback, image_close_cb, (XtPointer)arg);

    XtPopup(popup, XtGrabNone);
}

static void
image_close_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    popdown_arg *arg = (popdown_arg *)p1;
    XtPopdown(arg->popup);
    XtDestroyWidget(arg->popup);
    XFreePixmap(XtDisplay(w), arg->pixmap);
}

/**** fileimage  vpe_image  ****/

int get_int(void *ptr);

int
get_int(void *ptr)
{
    unsigned char *p = (unsigned char *)ptr;
    return (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | (p[3]);
}

/* ե뤫 vpe_image βѴ */
vpe_image *
get_image(void *fileimage_param, int size)
{
    vpe_image *image;
    int width, height;
    int d, x, y, c;
    int n;
    int v, values;
    char *fileimage = (char *)fileimage_param;

    /* magic Υå */
    if (fileimage[0] == 'P') {
	return get_image_pnm(fileimage_param, size);
    } else if (strncmp(fileimage, "!OBJ", 4) != 0) {
	return NULL;
    }

    if ((image = (vpe_image *)malloc(sizeof(vpe_image))) == NULL) {
	fprintf(stderr, "get_image: malloc failed\n");
	return NULL;
    }

    /* إå */
    width  = get_int(fileimage + 4);
    height = get_int(fileimage + 8);
    d      = get_int(fileimage + 12);
    /* fileimage + 16, 64: name */
    x      = get_int(fileimage + 80);
    y      = get_int(fileimage + 84);
    c      = get_int(fileimage + 88);

    image->width     = width;
    image->height    = height;
    image->type      = d;
    image->img_class = c;

    if (d != 1) {
	fprintf(stderr, "d != 1\n");
	return NULL;
    }
    if (c == 1) {
	values = 1;
    } else if (c == 2) {
	values = 3;
    } else {
	fprintf(stderr, "c != 1 or 2\n");
	return NULL;
    }

    /*  */
    if ((image->image = malloc(width*height*values*sizeof(int))) == NULL) {
	fprintf(stderr, "get_image: malloc failed\n");
	return NULL;
    }
    n = 0;
    for (y = 0; y < image->height; y++) {
	for (x = 0; x < image->width; x++) {
	    for (v = 0; v < values; v++) {
		((int *)(image->image))[n] = get_int(fileimage+(92 + n*4));
		n++;
	    }
	}
    }

    return image;
}

/**** pnm version ****/

static unsigned char *getc_ptr;
static unsigned char *getc_ptr_max;
static int read_number_from_image(int* ret);

static void
getc_from_image_init(void *fileimage, int size)
{
    getc_ptr = (unsigned char *)fileimage;
    getc_ptr_max = getc_ptr + size;
}

static int
getc_from_image(void)
{
    if (getc_ptr < getc_ptr_max) {
	return *getc_ptr++;
    } else {
	return EOF;
    }
}

static int
ungetc_from_image(int c)
{
    (void)c;
    getc_ptr--;
    return 0;
}

static vpe_image *
get_image_pnm(void *fileimage, int size)
{
    vpe_image *image;
    int c1, c2;
    int plain_or_raw;
    int width, height;
    int max_value;
    int x, y;
    int v, values;

    getc_from_image_init(fileimage, size);

    /* magic Υå */
    c1 = getc_from_image(); if (c1 != 'P') return NULL;
    c2 = getc_from_image();
    if (c2 == '2') {
	values = 1;       /* gray */
	plain_or_raw = 0; /* plain */
    } else if (c2 == '5') {
	values = 1;       /* gray */
	plain_or_raw = 1; /* raw */
    } else if (c2 == '3') {
	values = 3;       /* rgb */
	plain_or_raw = 0; /* plain */
    } else if (c2 == '6') {
	values = 3;       /* rgb */
	plain_or_raw = 1; /* raw */
    } else {
	fprintf(stderr, "get_image_pnm: not pgm or ppm file\n");
	return NULL;
    }
    
    /* إåɤ߹ */
    if (read_number_from_image(&width    ) != 0) return NULL;
    if (read_number_from_image(&height   ) != 0) return NULL;
    if (read_number_from_image(&max_value) != 0) return NULL;

    if (plain_or_raw == 1) {
	/* raw ΤȤ max_value μˤʤ⤷ʤ */
	/* ڡɤФƤ */
	int c;
	c = getc_from_image();
	if (!isspace(c))
	    ungetc_from_image(c);
    }
    /*  */

    if ((image = (vpe_image *)malloc(sizeof(vpe_image))) == NULL) {
	fprintf(stderr, "get_image_pnm: malloc failed\n");
	return NULL;
    }

    /* إå */
    image->width  = width;
    image->height = height;
    image->type = 1; /* 1: int */
    image->img_class = (values == 1) ? 1 : 2; /* 1: gray, 2: rgb */

    /*  */
    if ((image->image = malloc(width * height * values * sizeof(int))) == NULL)
    {
	fprintf(stderr, "get_image_pnm: malloc failed\n");
	free(image);
	return NULL;
    }

    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    for (v = 0; v < values; v++) {
		int ret, value;
		if (plain_or_raw == 0) {
		    /* plain */
		    ret = read_number_from_image(&value);
		} else {
		    /* raw */
		    value = getc_from_image();
		    ret = (value == EOF);
		}
		if (ret) {
		    fprintf(stderr, "get_image_pnm: read error\n");
		    free(image->image); free(image);
		    return NULL;
		}
		((int *)(image->image))[(y*width + x) * values + v] = value;
	    }
	}
    }
    return image;
}

/* pnm file ɤ߹Ѵؿ */

static int get(void);
static void skip_space(void);

/* ȤɤФ1ʸ֤ */
static int
get()
{
    int c;

    for (;;) {
	c = getc_from_image();
	if (c != '#')
	    break;
	while ((c = getc_from_image()) != '\n' && c != EOF)
	    ;
    }
    return c;
}

static void
skip_space()
{
    int c;

    while (c = get(), isspace(c))
	;
    if (c != EOF)
	ungetc_from_image(c);
}

/* ɤ߹֤ */
/* ɤ᤿ 0 ɤʤä 1 ֤ */
static int
read_number_from_image(int* ret)
{
    int i, c;
    char buf[256];

    skip_space();
    c = get();
    i = 0;
    if (c == '-') {
	buf[i++] = c;
	c = get();
    } else {
	ungetc_from_image(c);
    }
    if (isdigit(c)) {
	for (; i < 256; i++) {
	    c = get();
	    if (!isdigit(c)) {
		ungetc_from_image(c);
		buf[i] = '\0';
		break;
	    }
	    buf[i] = c;
	}
	if (i >= 256) {
	    fprintf(stderr, "read_number_from_image: buffer overflow\n");
	    return 1;
	}
	*ret = atoi(buf);
	return 0;
    } else {
	return 1;
    }
}
