/*********************************************************************
  Copyright (c) 1992-1999 Iwata Lab. All Rights Reserved.

  main.c
    VPE ᥤ
*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include "vpedefs.h"
#include "filesel.h"
#include "image.h"
#include "compile.h"
#include "com.h"
#include "protocol.h"
#include "msgpopup.h"
#include "port.h"

Widget		toplevel;
XtAppContext	app_context;
Widget		program_text_w;
Widget		quit_popup;

static Atom	wm_delete_window;

int		om_fd = -1;
int		modified_flag = 0;
char		*current_filename = NULL;

static void   create_menu(Widget parent);
static void   action_nop(void);
static Widget create_connect_om_popup(Widget parent);
static void   client_message_event(Widget w, XtPointer c, XEvent *e,Boolean *);
static void   file_cb(Widget w, XtPointer p1, XtPointer p2);
static filesel_cb_stat load_cb(char *input);
static filesel_cb_stat save_cb(char *input);

void display_cb(Widget w, XtPointer p1, XtPointer p2);
void quit(int status);
void quit_cb(Widget w, XtPointer p1, XtPointer p2);
void quit_ok_cb(Widget w, XtPointer p1, XtPointer p2);
void try_quit(void);
void quit_cancel_cb(Widget w, XtPointer p1, XtPointer p2);
void om_cb(Widget w, XtPointer p1, XtPointer p2);
void compile_cb(Widget w, XtPointer p1, XtPointer p2);
void run_cb(Widget w, XtPointer p1, XtPointer p2);

static char *fallback_resources[] = {
    "*title:			VPE",
    "*background:		#bebebe",
    "*Text*background:		white",
    "*Box.borderWidth:		0",
    "*fontSet:			-*-helvetica-medium-r-normal--14-*,*-fixed-medium-r-normal--16-*",
    "*yes.label:		yes",
    "*no.label:			no",
    "*ok.label:			ok",
    "*cancel.label:		cancel",
    "*dismiss.label:		dismiss",

    "*programText.width:	500",
    "*programText.height:	600",
    "*programText.fontSet:	*-fixed-medium-r-normal--16-*",
    "*progFrame.fromVert:	menu",
    "*progFrame.borderWidth: 	0",

    "*menu.orientation:		horizontal",
    "*menu.?.fontSet:		-*-helvetica-bold-r-normal--14-*,*-fixed-medium-r-normal--16-*",
    "*menu.file.label:		File",
    "*filePopup.load.label:	Load",
    "*filePopup.save.label:	Save",
    "*filePopup.display.label:	Display",
    "*filePopup.quit.label:	Quit",
    "*menu.connect.label:	Connect",
    "*connectPopup.om.label:	OM",
    "*menu.run.label:		Run",
    "*runPopup.compile.label:	Compile",
    "*runPopup.run.label:	Run",

    "*connectOmPopup.title:			connect to OM",
    "*connectOmPopup*frame.orientation:		vertical",
    "*connectOmPopup*frame1.orientation:	horizontal",
    "*connectOmPopup*frame2.orientation:	vertical",
    "*connectOmPopup*frame3.orientation:	vertical",
    "*connectOmPopup*frame4.orientation:	horizontal",
    "*connectOmPopup*frame2.host.label:		OM host",
    "*connectOmPopup*connect.label:		connect",
    "*connectOmPopup*disconnect.label:		disconnect",
    "*connectOmPopup*Text.translations: 	#override \
	Ctrl<Key>J:  nop() \n\
	Ctrl<Key>O:  nop() \n\
	Ctrl<Key>M:  nop() \n\
	<Key>Return: nop() \n\
",
    "*connectOmErrorPopup*message.label:	cannot connect to OM",
    "*connectOmErrorPopup*dismiss.label:	close",

    "*fileSelector.title:		file",
    "*fileSelector*top:			ChainTop",
    "*fileSelector*bottom:		ChainTop",
    "*fileSelector*input.width:		300",
    "*fileSelector*frame2.fromVert:	input",
    "*fileSelector*frame2.orientation:	horizontal",

    "*quitPopup*message.label:		are you sure?",
    "*quitPopup*frame2.fromVert:	message",
    "*quitPopup*ok.label:		yes",
    "*quitPopup*cancel.label:		no",
    "*quitPopup*cancel.fromHoriz:	ok",

    "*ipustat.title:			IPU status",
    "*ipustat*frame2.orientation:	horizontal",
    "*ipustat*frame2.hSpace:		0",
    "*ipustat*hostlabel.label:		hostname:",
    "*ipustat*loadlabel.label:		load:",
    "*ipustat*Box.vSpace:		2",
    "*ipustat*Box.vSpace:		2",

    "*confirmPopup*message.label:	file already exists, overwrite?",
    "*confirmPopup*frame2.orientation:	horizontal",

    "*msgPopup.title:			message",
    "*msgPopup*message.top:		ChainTop",
    "*msgPopup*message.bottom:		ChainBottom",
    "*msgPopup*dismiss.fromVert:	message",
    "*msgPopup*dismiss.top:		ChainBottom",
    "*msgPopup*dismiss.bottom:		ChainBottom",
    "*fileSelector*bottom:		ChainTop",

    NULL
};

static XtActionsRec actions[] = {
    {"nop",	(XtActionProc)action_nop},
};

/**** VPE ᥤؿ ****/
int
main(int argc, char *argv[])
{
    Widget top_frame;
    Widget frame2;
    int i;

    XtSetLanguageProc(NULL, NULL, NULL);
    toplevel = XtVaAppInitialize(&app_context, "VPE", NULL, 0,
				 &argc, argv, fallback_resources,
				 XtNmappedWhenManaged, False,
				 NULL);
    XtAppAddActions(app_context, actions, XtNumber(actions));

    top_frame = XtVaCreateManagedWidget("topFrame", formWidgetClass, toplevel,
					NULL);

    /** ˥塼 **/

    create_menu(top_frame);

    /** ץԽ **/

    frame2 = XtVaCreateManagedWidget("progFrame", formWidgetClass, top_frame,
				     NULL);
    program_text_w = XtVaCreateManagedWidget("programText",
			asciiTextWidgetClass, frame2,
			XtNeditType, XawtextEdit,
			XtNscrollVertical, XawtextScrollAlways,
			XtNscrollHorizontal, XawtextScrollAlways,
			NULL);

    XtRealizeWidget(toplevel);

    /* ɥޥ͡ delete б */
    wm_delete_window = XInternAtom(XtDisplay(toplevel),
				   "WM_DELETE_WINDOW", False);
    XSetWMProtocols(XtDisplay(toplevel), XtWindow(toplevel),
		    &wm_delete_window, 1);
    XtAddEventHandler(toplevel, 0, True, client_message_event, 0);

    XtMapWidget(toplevel);

    for (i = 1; i < argc; i++) {
	if (argv[i][0] == '-') {
	    /* option */
	} else {
	    /* load file */
	    load_cb(argv[1]);
	}
    }

    XtAppMainLoop(app_context);

    return 0;
}

/** ץ˥塼 **/
static Widget compile_bsb;
static Widget run_bsb;

static void create_quit_popup(Widget parent);
static void
create_menu(Widget parent)
{
    Widget menuframe;

    Widget file_popup;
    Widget load_bsb;
    Widget save_bsb;
    Widget display_bsb;
    Widget quit_bsb;

    Widget connect_popup;
    Widget om_bsb;
    Widget om_popup;

    Widget run_popup;
    Widget step_bsb;

    menuframe = XtVaCreateManagedWidget("menu", boxWidgetClass, parent,
					NULL);

    /** ե˥塼 **/

    XtVaCreateManagedWidget("file", menuButtonWidgetClass, menuframe,
			    XtNmenuName, "filePopup",
			    NULL);
    file_popup = XtVaCreatePopupShell("filePopup", simpleMenuWidgetClass,
				      toplevel,
				      NULL);
    load_bsb = XtVaCreateManagedWidget("load", smeBSBObjectClass, file_popup,
				       NULL);
    save_bsb = XtVaCreateManagedWidget("save", smeBSBObjectClass, file_popup,
				       NULL);
    display_bsb = XtVaCreateManagedWidget("display", smeBSBObjectClass,
					  file_popup,
					  NULL);
    XtVaCreateManagedWidget("line", smeLineObjectClass, file_popup, NULL);
    quit_bsb = XtVaCreateManagedWidget("quit", smeBSBObjectClass, file_popup,
				       NULL);
    XtAddCallback(load_bsb, XtNcallback, file_cb, (XtPointer)1);
    XtAddCallback(save_bsb, XtNcallback, file_cb, (XtPointer)2);
    XtAddCallback(display_bsb, XtNcallback, display_cb, 0);
    XtAddCallback(quit_bsb, XtNcallback, quit_cb, 0);

    create_fs_popup(toplevel);
    create_quit_popup(toplevel);

    /** ³˥塼 **/

    XtVaCreateManagedWidget("connect", menuButtonWidgetClass, menuframe,
			    XtNmenuName, "connectPopup",
			    NULL);
    connect_popup = XtVaCreatePopupShell("connectPopup", simpleMenuWidgetClass,
					 toplevel,
					 NULL);
    om_bsb = XtVaCreateManagedWidget("om", smeBSBObjectClass, connect_popup,
				     NULL);
    om_popup = create_connect_om_popup(toplevel);
    XtAddCallback(om_bsb, XtNcallback, om_cb, (XtPointer)om_popup);

    /** ¹ԥ˥塼 **/

    XtVaCreateManagedWidget("run", menuButtonWidgetClass, menuframe,
			    XtNmenuName, "runPopup",
			    NULL);
    run_popup = XtVaCreatePopupShell("runPopup",simpleMenuWidgetClass,toplevel,
				     NULL);
    compile_bsb = XtVaCreateManagedWidget("compile", smeBSBObjectClass,
					  run_popup,
					  XtNsensitive, False,
					  NULL);
    XtAddCallback(compile_bsb, XtNcallback, compile_cb, 0);
    run_bsb = XtVaCreateManagedWidget("run", smeBSBObjectClass, run_popup,
				      XtNsensitive, False,
				      NULL);
    XtAddCallback(run_bsb, XtNcallback, run_cb, 0);
}

/* λݥåץå */
static void
create_quit_popup(Widget parent)
{
    Widget quit_frame;
    Widget quit_frame2;
    Widget quit_ok;
    Widget quit_cancel;

    quit_popup = XtVaCreatePopupShell("quitPopup", topLevelShellWidgetClass,
				      parent, NULL);
    quit_frame = XtVaCreateManagedWidget("frame", formWidgetClass,
					 quit_popup, NULL);
    XtVaCreateManagedWidget("message", labelWidgetClass, quit_frame, NULL);
    quit_frame2 = XtVaCreateManagedWidget("frame2", formWidgetClass,
					  quit_frame, NULL);
    quit_ok = XtVaCreateManagedWidget("yes", commandWidgetClass, quit_frame2,
				      NULL);
    quit_cancel = XtVaCreateManagedWidget("no", commandWidgetClass,
					  quit_frame2,
					  NULL);
    XtAddCallback(quit_ok, XtNcallback, quit_ok_cb, 0);
    XtAddCallback(quit_cancel, XtNcallback, quit_cancel_cb, 0);
}

/** ץ˥塼 save  load Ф줿Ȥ˸ƤФ */
static void
file_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    if ((int)p1 == 1) {
	/* load */
	popup_filesel(load_cb);
    } else {
	/* save */
	popup_filesel(save_cb);
    }
}

static void
action_nop(void)
{
}

static void
set_current_filename(char *filename)
{
    if (current_filename != NULL) {
	free(current_filename);
    }
    current_filename = strdup(filename);
}

static filesel_cb_stat
load_cb(char *input)
{
    char *prog_buf = NULL;
    FILE *fp = NULL;
    struct stat st;
    char *filename;
    filesel_cb_stat ret;

    if (input == NULL) {
	/* canceled */
	return fcs_ok;
    }

    filename = input;
    if (stat(filename, &st) < 0) {
	/* ĤʤΤǥ顼 */
	ret   = fcs_err_errno;
	errno = ENOENT;
	goto end;
    }
    if ((fp = fopen(filename, "r")) == NULL) {
	ret = fcs_err_errno;
	goto end;
    }
    if ((prog_buf = (char *)malloc(st.st_size + 1)) == NULL) {
	ret = fcs_err_other;
	goto end;
    }
    if (fread(prog_buf, 1, st.st_size, fp) != st.st_size) {
	ret = fcs_err_other;
	goto end;
    }
    prog_buf[st.st_size] = '\0';
    XtVaSetValues(program_text_w, XtNstring, prog_buf, NULL);
    ret = fcs_ok;
    set_current_filename(filename);

  end:
    if (prog_buf != NULL) /*free(prog_buf)*/;
    if (fp != NULL) fclose(fp);

    return ret;
}

static int save_confirm(void);

static filesel_cb_stat
save_cb(char *input)
{
    int         conf;
    struct stat st;
    FILE        *fp;
    char        *prog_buf;

    if (input == NULL) {
	/* canceled */
	return fcs_ok;
    }

#if 0
    if (stat(input, &st) == 0) {
	conf = save_confirm();
	if (conf == 2) {
	    /* no save */
	    return fcs_err;
	}
    }
#endif

    if ((fp = fopen(input, "w")) == NULL) {
	return fcs_err_errno;
    }
    XtVaGetValues(program_text_w, XtNstring, &prog_buf, NULL);
    fwrite(prog_buf, strlen(prog_buf), 1, fp);
    /*XtFree(prog_buf)*/;
    fclose(fp);

    return fcs_ok;
}

/* ǧ */
static Widget save_confirm_popup = 0;
static int    save_confirm_flag;

static void save_confirm_cb(Widget w, XtPointer p1, XtPointer p2);

static int
save_confirm(void)
{
    if (save_confirm_popup == 0) {
	Widget popup, frame, frame2, ok, cancel;
	popup = XtVaCreatePopupShell("confirmPopup", topLevelShellWidgetClass,
				     toplevel, NULL);
	save_confirm_popup = popup;
	frame = XtVaCreateManagedWidget("frame", boxWidgetClass, popup, NULL);
	XtVaCreateManagedWidget("message", labelWidgetClass, frame, NULL);
	frame2 = XtVaCreateManagedWidget("frame2", boxWidgetClass,frame,NULL);
	ok = XtVaCreateManagedWidget("yes", commandWidgetClass, frame2, NULL);
	cancel = XtVaCreateManagedWidget("no", commandWidgetClass,
					 frame2, NULL);
	XtAddCallback(ok,     XtNcallback, save_confirm_cb, (XtPointer)1);
	XtAddCallback(cancel, XtNcallback, save_confirm_cb, (XtPointer)2);
    }
    save_confirm_flag = -1;
    XtPopup(save_confirm_popup, XtGrabExclusive);
    while (save_confirm_flag < 0) {
	XEvent ev;
	XtAppNextEvent(app_context, &ev);
	XtDispatchEvent(&ev);
    }
    return save_confirm_flag;
}

static void
save_confirm_cb(Widget w, XtPointer p1, XtPointer p2)
{
    save_confirm_flag = (int)p1;
    XtPopdown(save_confirm_popup);
}

/**** ɽ ****/

static filesel_cb_stat display_cb2(char *input);

void
display_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    popup_filesel(display_cb2);
}

static filesel_cb_stat
display_cb2(char *input)
{
    char *display_buf = NULL;
    FILE *fp = NULL;
    struct stat st;
    char *filename;
    int  ret = 1;

    if (input == NULL) {
	/* canceled */
	return 0;
    }

    filename = input;
    if (stat(filename, &st) < 0) {
	/* ĤʤΤǥ顼 */
	goto end;
    }
    if ((fp = fopen(filename, "rb")) == NULL) {
	goto end;
    }
    if ((display_buf = (char *)malloc(st.st_size)) == NULL) {
	goto end;
    }
    if (fread(display_buf, 1, st.st_size, fp) != st.st_size) {
	goto end;
    }
    show_image(display_buf, st.st_size);
    ret = 0;

  end:
    if (display_buf != NULL) /*free(display_buf)*/;
    if (fp != NULL) fclose(fp);

    return ret;
}

void
quit(int status)
{
    vpe_om		msg;
    unsigned char	buf[8];

    if (om_fd >= 0) {
	msg.type = p_vo_quit;
	msg.size = 0;

	buf[0] = (unsigned char)((msg.type >> 24) & 0xff);
	buf[1] = (unsigned char)((msg.type >> 16) & 0xff);
	buf[2] = (unsigned char)((msg.type >>  8) & 0xff);
	buf[3] = (unsigned char)((msg.type      ) & 0xff);
	buf[4] = (unsigned char)((msg.size >> 24) & 0xff);
	buf[5] = (unsigned char)((msg.size >> 16) & 0xff);
	buf[6] = (unsigned char)((msg.size >>  8) & 0xff);
	buf[7] = (unsigned char)((msg.size      ) & 0xff);
	net_write(om_fd, buf, 8);
    }
    exit(status);
}

void
try_quit(void)
{
    if (modified_flag) {
	XtPopup(quit_popup, XtGrabExclusive);
    } else {
	quit(0);
    }
}

void
quit_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    try_quit();
}

void
quit_ok_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    quit(0);
}

void
quit_cancel_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    XtPopdown(quit_popup);
}

static void
client_message_event(Widget w, XtPointer closure, XEvent *ev, Boolean *cont)
/*ARGSUSED*/
{
    if (ev->xclient.message_type ==
	XInternAtom(XtDisplay(w), "WM_PROTOCOLS", False))
    {
	if (ev->xclient.data.l[0] == wm_delete_window) {
	    try_quit();
	}
    }
}


/** ˥塼 OM Ф줿Ȥ˸ƤФ **/
void
om_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    XtPopup((Widget)p1, XtGrabExclusive);
}

void
compile_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    char *program;

    XtVaGetValues(program_text_w, XtNstring, &program, NULL);
    compile_module(program);
    /*XtFree(program);*/
}

void
run_cb(Widget w, XtPointer p1, XtPointer p2)
/*ARGSUSED*/
{
    char *program;

    XtVaGetValues(program_text_w, XtNstring, &program, NULL);
    run_module(program);
    /*XtFree(program);*/
}

/**** OM ³ ****/

static Widget om_host_w, om_port_w;
static XtInputId om_input_id;
static void om_connect_cb(Widget w, XtPointer closure, XtPointer call_data);
static void om_disconnect_cb(Widget w, XtPointer closure, XtPointer call_data);
static void om_cancel_cb(Widget w, XtPointer closure, XtPointer call_data);

/* ݥåץåפ */
static Widget
create_connect_om_popup(Widget parent)
{
    Widget popup;
    Widget frame;
    Widget frame1;
    Widget frame2;
    Widget frame3;
    Widget frame4;
    Widget connect;
    Widget disconnect;
    Widget cancel;

    popup = XtVaCreatePopupShell("connectOmPopup", topLevelShellWidgetClass,
				 parent, NULL);
    frame = XtVaCreateManagedWidget("frame", boxWidgetClass, popup, NULL);
    frame1 = XtVaCreateManagedWidget("frame1", boxWidgetClass, frame, NULL);
    frame2 = XtVaCreateManagedWidget("frame2", boxWidgetClass, frame1, NULL);
    XtVaCreateManagedWidget("host", labelWidgetClass, frame2, NULL);
    XtVaCreateManagedWidget("port", labelWidgetClass, frame2, NULL);
    frame3 = XtVaCreateManagedWidget("frame3", boxWidgetClass, frame1, NULL);
    om_host_w = XtVaCreateManagedWidget("host", asciiTextWidgetClass, frame3,
					XtNeditType, XawtextEdit,
					NULL);
    om_port_w = XtVaCreateManagedWidget("port", asciiTextWidgetClass, frame3,
					XtNeditType, XawtextEdit,
					NULL);
    frame4 = XtVaCreateManagedWidget("frame4", boxWidgetClass, frame, NULL);
    connect = XtVaCreateManagedWidget("connect", commandWidgetClass, frame4,
				      NULL);
    disconnect = XtVaCreateManagedWidget("disconnect", commandWidgetClass,
					 frame4, NULL);
    cancel = XtVaCreateManagedWidget("cancel", commandWidgetClass,
				     frame4, NULL);
    XtAddCallback(connect, XtNcallback, om_connect_cb, (XtPointer)popup);
    XtAddCallback(disconnect, XtNcallback, om_disconnect_cb, (XtPointer)popup);
    XtAddCallback(cancel, XtNcallback, om_cancel_cb, (XtPointer)popup);

    return popup;
}

static void
om_connect_cb(Widget w, XtPointer closure, XtPointer call_data)
/*ARGSUSED*/
{
    char *host_str, *port_str;
    char *host;
    int  int_port;
    Widget popup = (Widget)closure;

    XtVaGetValues(om_host_w, XtNstring, &host_str, NULL);
    XtVaGetValues(om_port_w, XtNstring, &port_str, NULL);

    if (*host_str == '\0') {
	host = "localhost";
    } else {
	host = host_str;
    }
    if (*port_str == '\0') {
	int_port = get_om_port();
    } else {
	int_port = atoi(port_str);
    }
    if ((om_fd = connect_om(host, int_port)) >= 0) {
	/*  */
	XtPopdown(popup);
	XtVaSetValues(compile_bsb, XtNsensitive, True, NULL);
	XtVaSetValues(run_bsb, XtNsensitive, True, NULL);
	om_input_id = XtAppAddInput(app_context, om_fd,
			(XtPointer)XtInputReadMask, om_read_cb, 0);
    }
}

static void
om_disconnect_cb(Widget w, XtPointer closure, XtPointer call_data)
/*ARGSUSED*/
{
    if (om_fd >= 0) {
	close_om(om_fd);
	om_fd = -1;
	XtRemoveInput(om_input_id);
	XtPopdown((Widget)closure);
    } else {
	msgpopup("not connected");
    }
}

static void
om_cancel_cb(Widget w, XtPointer closure, XtPointer call_data)
/*ARGSUSED*/
{
    XtPopdown((Widget)closure);
}
